//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Bwg.Logging;

namespace Bwg.Audio.Acm
{
    public unsafe class AcmDriver
    {
        #region private variables
        private Logger m_logger;
        private uint m_handle ;
        private uint m_driver_id;
        private AcmFourCC m_driver_type;
        private AcmFourCC m_driver_subtype;
        private ushort m_manufacturer_id;
        private ushort m_product_id;
        private uint m_acm_version;
        private uint m_support_flags;
        private uint m_format_tags;
        private uint m_filter_tags;
        private string m_short_name;
        private string m_long_name;
        private string m_copyright;
        private string m_license;
        private string m_features;

        private AcmFormat [] m_formats;
        private AcmFormatDetails [] m_format_details ;
        #endregion

        #region public properties

        public bool IsOpen
        {
            get
            {
                return m_handle != 0;
            }
        }

        public uint Handle
        {
            get
            {
                return m_handle ;
            }
        }

        public AcmFormatDetails[] FormatDescriptors
        {
            get
            {
                return m_format_details;
            }
        }

        public AcmFormat [] Formats
        {
            get
            {
                return m_formats ;
            }
        }

        public uint DriverID
        {
            get
            {
                return m_driver_id;
            }
        }

        public AcmFourCC DriverType
        {
            get
            {
                return m_driver_type;
            }
        }

        public AcmFourCC DriverSubType
        {
            get
            {
                return m_driver_subtype;
            }
        }

        public ushort ManufacturerId
        {
            get
            {
                return m_manufacturer_id;
            }
        }

        public ushort ProductId
        {
            get
            {
                return m_product_id;
            }
        }

        public uint SupportFlags
        {
            get
            {
                return m_support_flags;
            }
        }

        public bool IsAsync
        {
            get
            {
                return (m_support_flags & AcmInterface.ACMDRIVERDETAILS_SUPPORTF_ASYNC) != 0;
            }
        }

        public bool IsCodec
        {
            get
            {
                return (m_support_flags & AcmInterface.ACMDRIVERDETAILS_SUPPORTF_CODEC) != 0;
            }
        }

        public bool IsConverter
        {
            get
            {
                return (m_support_flags & AcmInterface.ACMDRIVERDETAILS_SUPPORTF_CONVERTER) != 0;
            }
        }

        public bool IsDisabled
        {
            get
            {
                return (m_support_flags & AcmInterface.ACMDRIVERDETAILS_SUPPORTF_DISABLED) != 0;
            }
        }

        public bool IsFilter
        {
            get
            {
                return (m_support_flags & AcmInterface.ACMDRIVERDETAILS_SUPPORTF_FILTER) != 0;
            }
        }

        public bool IsHardware
        {
            get
            {
                return (m_support_flags & AcmInterface.ACMDRIVERDETAILS_SUPPORTF_HARDWARE) != 0;
            }
        }

        public bool IsLocal
        {
            get
            {
                return (m_support_flags & AcmInterface.ACMDRIVERDETAILS_SUPPORTF_LOCAL) != 0;
            }
        }

        public uint FormatTagCount
        {
            get
            {
                return m_format_tags;
            }
        }

        public uint FilterTagCount
        {
            get
            {
                return m_filter_tags;
            }
        }

        public string ShortName
        {
            get
            {
                return m_short_name;
            }
        }

        public string LongName
        {
            get
            {
                return m_long_name;
            }
        }

        #endregion

        #region constructors
        public AcmDriver(Logger l, uint id)
        {
            m_logger = l;
            m_formats = new AcmFormat[0];
            m_format_details = new AcmFormatDetails[0];
            m_driver_id = id;
            m_handle = 0;
        }
        #endregion

        #region public methods
        public override string ToString()
        {
            return m_long_name;
        }

        public bool Open()
        {
            uint res;

            fixed (uint* ptr = &m_handle)
            {
                res = AcmInterface.acmDriverOpen(ptr, m_driver_id, 0);
                if (res != 0)
                    return false;
            }

            return true;
        }

        public bool Close()
        {
            uint result;

            result = AcmInterface.acmDriverClose(m_handle, 0);
            m_handle = 0;
            return result == 0;
        }

        public bool SupportsFormatTag(AcmInterface.WaveFormat tag)
        {
            foreach (AcmFormat f in m_formats)
            {
                if (f.Tag == tag)
                    return true;
            }
            return false;
        }

        public bool Initialize()
        {
            uint result ;

            AcmInterface.ACMDRIVERDETAILS details = new AcmInterface.ACMDRIVERDETAILS() ;
            details.cbStruct = (uint)Marshal.SizeOf(details);
            result = AcmInterface.acmDriverDetails(m_driver_id, &details, 0);
            if (result != 0)
                return false;

            m_driver_type = new AcmFourCC(details.fccType[0], details.fccType[1], details.fccType[2], details.fccType[3]);
            m_driver_subtype = new AcmFourCC(details.fccComp[0], details.fccComp[1], details.fccComp[2], details.fccComp[3]);
            m_manufacturer_id = details.wMid;
            m_product_id = details.wPid;
            m_acm_version = details.vdwACM;
            m_support_flags = details.fdwSupport;
            m_format_tags = details.cFormatTags;
            m_filter_tags = details.cFilterTags;
            m_short_name = Marshal.PtrToStringAnsi(new IntPtr(&details.szShortName[0]));
            m_long_name = Marshal.PtrToStringAnsi(new IntPtr(&details.szLongName[0]));
            m_copyright = Marshal.PtrToStringAnsi(new IntPtr(&details.szCopyright[0]));
            m_license = Marshal.PtrToStringAnsi(new IntPtr(&details.szLicensing[0]));
            m_features = Marshal.PtrToStringAnsi(new IntPtr(&details.szFeatures[0]));

            Debug.WriteLine("Initializing ACM driver '" + m_long_name + "'");

            if (!Open())
                return false ;

            AcmInterface.ACMFORMATTAGDETAILS tagdetails = new AcmInterface.ACMFORMATTAGDETAILS() ;
            tagdetails.cbStruct = (uint)Marshal.SizeOf(tagdetails);
            result = AcmInterface.acmFormatTagEnum(Handle, &tagdetails, new AcmInterface.AcmFormatTagEnumCallback(FormatTagCallback), 0, 0);
            if (result != 0)
            {
                Close();
                return false;
            }

            QueryAllFormats(Handle);
            return Close();
        }

        public bool QueryAllFormats(uint handle)
        {
            uint result;

            uint value ;
            result = AcmInterface.acmMetrics(handle, AcmInterface.ACM_METRIC_MAX_SIZE_FORMAT, &value) ;
            if (result != 0)
                value = 64 ;

            foreach (AcmFormat fmt in m_formats)
            {
                for (uint i = 0; i < fmt.FormatCount; i++)
                {
                    AcmInterface.ACMFORMATDETAILS fmtdetails = new AcmInterface.ACMFORMATDETAILS();
                    fmtdetails.cbStruct = (uint)Marshal.SizeOf(fmtdetails);
                    fmtdetails.dwFormatTag = (uint)fmt.Tag;
                    fmtdetails.dwFormatIndex = i;
                    fmtdetails.pwfx = (AcmInterface.WAVEFORMATEX*)Marshal.AllocHGlobal((int)value);
                    fmtdetails.cbwfx = value;

                    result = AcmInterface.acmFormatDetails(handle, &fmtdetails, AcmInterface.ACM_FORMATDETAILSF_INDEX);

                    string name = Marshal.PtrToStringAnsi(new IntPtr(&fmtdetails.szFormat[0]));
                    AcmFormatDetails fmtobj = new AcmFormatDetails(name, (AcmInterface.WaveFormat)fmtdetails.dwFormatTag, 
                                                                   new IntPtr(fmtdetails.pwfx));

                    string str ;
                    
                    str = "AcmDriver " + ShortName + " FormatTag " + fmt.Tag.ToString();
                    str += " - added format " + fmtobj.Name;
                    m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 3, str));

                    Array.Resize(ref m_format_details, m_format_details.GetLength(0) + 1);
                    m_format_details[m_format_details.GetLength(0) - 1] = fmtobj;
                }
            }

            return result == 0 ;
        }

        public byte[] GetFormat(AcmFormatDesc desc)
        {
            foreach (AcmFormatDetails details in m_format_details)
            {
                if (details.Tag == desc.Format)
                {
                    AcmInterface.WAVEFORMATEX* fmt_p = (AcmInterface.WAVEFORMATEX*)details.WaveFormatExPtr.ToPointer();
                    return ConvertToBytes(details.WaveFormatExPtr);
                }
            }
            return null;
        }

        #endregion

        #region private member functions
        private byte[] ConvertToBytes(IntPtr fmt)
        {
            AcmInterface.WAVEFORMATEX *fmt_p = (AcmInterface.WAVEFORMATEX *)fmt.ToPointer();
            int size = Marshal.SizeOf(*fmt_p) + fmt_p->cbSize;

            byte[] result = new byte[size];
            Marshal.Copy(fmt, result, 0, size);
            return result;
        }

        private bool FormatTagCallback(uint id, AcmInterface.ACMFORMATTAGDETAILS* details_p, uint instance, uint support)
        {
            string str;
            string name = Marshal.PtrToStringAnsi(new IntPtr(&details_p->szFormatTag[0]));
            AcmInterface.WaveFormat fmttag = (AcmInterface.WaveFormat)details_p->dwFormatTag;
            AcmFormat format = new AcmFormat(fmttag, name, details_p->cStandardFormats);

            str = "AcmDriver " + m_short_name + " - found format " + fmttag.ToString();
            str += ", name='" + name + "'";
            m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 2, str));

            Array.Resize(ref m_formats, m_formats.GetLength(0) + 1) ;
            m_formats[m_formats.GetLength(0) - 1] = format ;

            return true;
        }
        #endregion
    }
}
