//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Bwg.Audio.Acm ;

namespace Bwg.Audio
{
    unsafe class Mp3HeaderReader : AudioHeaderReader
    {
        #region private types
        private enum Id3Flags : byte
        {
            Unsynchronisation = 0x80,
            ExtendedHeader = 0x40,
            ExperimentalIndicator = 0x20,
            FooterPresent = 0x10,
        } ;
        #endregion

        #region private variables
        private byte m_major_version;
        private byte m_minor_version;
        private byte m_flags;
        private uint m_size;
        #endregion

        #region constructor
        public Mp3HeaderReader(AcmManager mgr) : base(mgr)
        {
            m_buffer_size = 417;
        }
        #endregion

        #region public methods
        public override void ReadHeader(string filename)
        {
            FileStream strm = new FileStream(filename, FileMode.Open, FileAccess.Read);
            ProbeFile(filename, strm);

            m_audio_data_position = strm.Position;

            AcmFormatDesc desc = new AcmFormatDesc(AcmInterface.WaveFormat.WAVE_FORMAT_MPEGLAYER3, 2);

            CreateFormat();

            m_file_type_string = "Mpeg Layer 3 File (mp3)";
        }
        #endregion

        #region private methods

        private void CreateFormat()
        {
            AcmInterface.MPEGLAYER3WAVEFORMAT fmt = new AcmInterface.MPEGLAYER3WAVEFORMAT();

            fmt.wfx.cbSize = AcmInterface.MPEGLAYER3_WFX_EXTRA_BYTES;
            fmt.wfx.wFormatTag = (ushort)AcmInterface.WaveFormat.WAVE_FORMAT_MPEGLAYER3;
            fmt.wfx.nAvgBytesPerSec = 160 * (1024 / 8);
            fmt.wfx.nChannels = 2;
            fmt.wfx.nSamplesPerSec = 44100;
            fmt.wfx.wBitsPerSample = 0;
            fmt.wfx.nBlockAlign = 1;
            fmt.wID = AcmInterface.MPEGLAYER3_ID_MPEG;
            fmt.fdwFlags = AcmInterface.MPEGLAYER3_FLAG_PADDING_OFF;
            fmt.nBlockSize = 417;
            fmt.nFramesPerBlock = 1;
            fmt.nCodecDelay = 1393;

            m_data_format = new byte[Marshal.SizeOf(fmt)];
            IntPtr ptr = new IntPtr(&fmt) ;
            Marshal.Copy(ptr, m_data_format, 0, m_data_format.GetLength(0));
        }

        private string ReadString(string filename, FileStream strm, int length)
        {
            string str = "";

            while (length-- > 0)
            {
                int b = strm.ReadByte();
                if (b == -1)
                    throw new InvalidFileFormat(filename, "unexpected end of file reading string");

                str += (char)b;
            }

            return str;
        }

        private byte Read8(string filename, FileStream strm)
        {
            int b = strm.ReadByte();
            if (b == -1)
                throw new InvalidFileFormat(filename, "unexpected end of file");

            return (byte)b;
        }

        private ushort Read16(string filename, FileStream strm)
        {
            byte b1 = Read8(filename, strm);
            byte b2 = Read8(filename, strm);

            return (ushort)((b1 << 8) | b2);
        }

        private uint Read32(string filename, FileStream strm)
        {
            byte b1 = Read8(filename, strm);
            byte b2 = Read8(filename, strm);
            byte b3 = Read8(filename, strm);
            byte b4 = Read8(filename, strm);

            return (uint)((b1 << 24) | (b2 << 16) | (b3 << 8) | b4);
        }

        private uint ReadSync32(string filename, FileStream strm)
        {
            byte b1 = Read8(filename, strm);
            byte b2 = Read8(filename, strm);
            byte b3 = Read8(filename, strm);
            byte b4 = Read8(filename, strm);

            Debug.Assert((b1 & 0x80) == 0);
            Debug.Assert((b2 & 0x80) == 0);
            Debug.Assert((b3 & 0x80) == 0);
            Debug.Assert((b4 & 0x80) == 0);

            return (uint)((b1 << 21) | (b2 << 14) | (b3 << 7) | b4);
        }

        private void ReadHeader(string filename, FileStream strm)
        {
            string tag;

            tag = ReadString(filename, strm, 3);
            if (tag != "ID3")
                throw new InvalidFileFormat(filename, "invalid MP3 file, expected header 'ID3'");

            m_major_version = Read8(filename, strm);
            if (m_major_version > 4)
                throw new InvalidFileFormat(filename, "invalid (or unsupported) version number '" + m_major_version.ToString() + "'");

            m_minor_version = Read8(filename, strm);

            m_flags = Read8(filename, strm);

            m_size = ReadSync32(filename, strm);
        }

        private void ReadExtendedHeader(string filename, FileStream strm)
        {
            uint hdsize = ReadSync32(filename, strm);

            // We ignore the extended header for now
            for (int i = 0; i < hdsize - 4; i++)
            {
                int b = strm.ReadByte();
                if (b == -1)
                    throw new InvalidFileFormat(filename, "unexpected end of file reading extended header");
            }
        }

        private bool ProbeFile(string filename, FileStream strm)
        {
            // Read the 10 byte header
            try
            {
                ReadHeader(filename, strm);
                if ((m_flags & (byte)Id3Flags.ExtendedHeader) != 0)
                    ReadExtendedHeader(filename, strm);

                long dpos = m_size + 10;
                if ((m_flags & (byte)Id3Flags.FooterPresent) != 0)
                    dpos += 10;

                // Seek to the data
                strm.Seek(dpos, SeekOrigin.Begin);
            }
            catch (InvalidFileFormat)
            {
                // 
                // Check to see if this is just raw MP3 data without an
                // ID3 headers
                //
                strm.Seek(0, SeekOrigin.Begin);
                while (true)
                {
                    int b = strm.ReadByte() ;
                    if (b == -1)
                        return false ;

                    if (b == 0xff)
                    {
                        b = strm.ReadByte() ;
                        if (b == 0xfb)
                        {
                            strm.Seek(-2, SeekOrigin.Current);
                            return true;
                        }
                        else
                        {
                            strm.Seek(-1, SeekOrigin.Current);
                        }
                    }
                }
            }

            return true;
        }
        #endregion
    }
}
