//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.IO ;
using System.Diagnostics;
using Bwg.Audio.Acm;

namespace Bwg.Audio
{
    /// <summary>
    /// Write the MP3 header.  
    /// Tags to consider:
    /// -  TIT2 - title
    /// -  TRCK - track number
    /// -  TYER - year
    /// -  TPUB - publisher
    /// -  TPE1 - artist
    /// -  TCON - content type
    /// -  TALB - album title
    /// </summary>
    class Mp3HeaderWriter : AudioHeaderWriter
    {
        #region private members
        /// <summary>
        /// The start position of the data in the file
        /// </summary>
        private long m_data_pos ;
        #endregion

        #region public properties
        public override string Extension
        {
            get
            {
                return "mp3";
            }
        }
        #endregion

        #region constructor
        public Mp3HeaderWriter(AcmFormatDetails fmt)
            : base(fmt)
        {
        }
        #endregion

        #region public methods
        /// <summary>
        /// Write the header for the given file type
        /// </summary>
        /// <param name="strm">the streawm to write to</param>
        /// <returns>true if sucessful, false otherwise</returns>
        public override bool WriteHeader(IDictionary<string, object> props, FileStream strm)
        {
            if (!LayoutMp3File())
                return false ;

            if (!WriteMp3Header(strm))
                return false;

            if (!WriteRecords(props, strm))
                return false ;

            // Now, pad the file to the location of the data
            byte [] zeros = new byte[2048] ;
            strm.Seek(0, SeekOrigin.End) ;
            while (strm.Position < m_data_pos)
            {
                long needed = m_data_pos - strm.Position ;
                if (needed > zeros.GetLength(0))
                    needed = zeros.GetLength(0) ;

                strm.Write(zeros, 0, (int)needed) ;
            }

            return true;
        }

        /// <summary>
        /// This method patches the previously written header based on the amount
        /// of data written by the converter
        /// </summary>
        /// <param name="strm">the stream to write to</param>
        /// <param name="totaldata">the total amount of audio data written, in bytes</param>
        /// <returns>true if sucessful, false otherwise</returns>
        public override bool PatchHeader(IDictionary<string, object> props, FileStream strm, int totaldata)
        {
            return true ;
        }
        #endregion

        #region private methods
        private bool LayoutMp3File()
        {
            //
            // Leave space in the headers for additional tags
            //
            m_data_pos = 1024 * 8;
            return true;
        }

        private bool WriteMp3Header(FileStream strm)
        {
            WriteString(strm, "ID3");                               // ID Field
            strm.WriteByte(3);                                      // Major Version
            strm.WriteByte(0);                                      // Minor Version
            strm.WriteByte(0);                                      // Flags
            Write32Sync(strm, (uint)(m_data_pos - 10));             // Tag size

            return true ;
        }

        private bool WriteRecords(IDictionary<string, object> props, FileStream strm)
        {
            if (props.ContainsKey(AcmWriter.PropNameSongTitle) && props[AcmWriter.PropNameSongTitle] is string)
            {
                if (!WriteStringRecord(strm, AcmWriter.PropNameSongTitle, (string)props[AcmWriter.PropNameSongTitle]))
                    return false ;
            }

            if (props.ContainsKey(AcmWriter.PropNameSongArtist) && props[AcmWriter.PropNameSongArtist] is string)
            {
                if (!WriteStringRecord(strm, AcmWriter.PropNameSongArtist, (string)props[AcmWriter.PropNameSongArtist]))
                    return false ;
            }

            if (props.ContainsKey(AcmWriter.PropNameAlbum) && props[AcmWriter.PropNameAlbum] is string)
            {
                if (!WriteStringRecord(strm, AcmWriter.PropNameAlbum, (string)props[AcmWriter.PropNameAlbum]))
                    return false ;
            }

            if (props.ContainsKey(AcmWriter.PropNameTrackNumber) && props[AcmWriter.PropNameTrackNumber] is string)
            {
                if (!WriteStringRecord(strm, AcmWriter.PropNameTrackNumber, (string)props[AcmWriter.PropNameTrackNumber]))
                    return false ;
            }
            return true;
        }

        private bool WriteStringRecord(FileStream strm, string tag, string value)
        {
            WriteString(strm, tag);                             // Write the tag for the record
            Write32Big(strm, (uint)value.Length + 1);           // String length + 1 for the encoding type byte
            Write16Big(strm, 0);                                // Flags
            strm.WriteByte(0);                                  // Encoding bytes
            WriteString(strm, value);                           // String bytes
            return true ;
        }

        private bool WriteIntegerRecord(FileStream strm, string tag, Int32 value)
        {
            Debug.Assert(false) ;
            return true ;
        }

        private void Write32Sync(FileStream strm, uint value)
        {
            byte b;

            b = (byte)((value >> 21) & 0x7f);
            strm.WriteByte(b);
            b = (byte)((value >> 14) & 0x7f);
            strm.WriteByte(b);
            b = (byte)((value >> 7) & 0x7f);
            strm.WriteByte(b);
            b = (byte)(value & 0x7f);
            strm.WriteByte(b);
        }

        #endregion
    }
}
