//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.IO;
using Bwg.Scsi;
using Bwg.Burner;
using Bwg.Logging;
using Bwg.Controls;

namespace Bwg.BurnIso
{
    public partial class BurnIsoForm : BwgBurnBaseForm
    {
        private Thread m_burner_thread;
        private Device m_device;
        private Drive m_drive;
        private string m_isofile;
        private System.Windows.Forms.Timer m_timer;
        private DateTime m_start;
        private Drive.BurnType m_burn_type;
        private bool m_multi_session;
        private SpeedInfo m_burn_speed;

        private delegate void TimerTickDelegate();
        private delegate void DoneDelegate(DiskOperationError status);
        private delegate void ProgressHandlerDelegate(object sender, ProgressChangedEventArgs args);

        /// <summary>
        /// Create a application form for the Burn ISO application
        /// </summary>
        public BurnIsoForm(string initial)
        {
            // System.Threading.Thread.CurrentThread.CurrentUICulture = new System.Globalization.CultureInfo("it-IT", false); 

            InitializeComponent();
            RepositionControlsBasedOnLanguage();

            LoggerWindow.SetupColumns();
            LoggerWindow.Sink = m_sink;

            m_logger.LogMessage(new UserMessage(UserMessage.Category.Info, 0, "ISO Burner Version " + GetVersion()));

            // Set the logger and populate the dropdown with the drives found
            DeviceSelector.Logger = m_logger;
            DeviceSelector.SelectedIndexChanged += new EventHandler(DeviceSelector_IndexChanged);
            DeviceSelector.InitializeDeviceList(false);

            m_timer = new System.Windows.Forms.Timer();

            m_burn_type = (Drive.BurnType)GetStartupInteger(PropertyNameBurnIsoImageBurnType, (int)Drive.BurnType.DontCare);
            m_multi_session = GetStartupBool(PropertyNameBurnIsoImageMultiSession, false);
            m_burn_speed = null;

            if (initial != string.Empty)
            {
                try
                {
                    FileInfo info = new FileInfo(initial);
                    IsoFileName.Text = info.FullName;
                }
                catch (Exception)
                {
                }
            }
        }

        private void RepositionControlsBasedOnLanguage()
        {
            MoveRightOf(DeviceSelector, CDDriveLabel, 4);
            MoveRightOfExpand(Status, DeviceSelector, 4, Status.Right - DeviceSelector.Left);
            MoveRightOfExpand(IsoFileName, IsoFileLabel, 4, IsoFileName.Right - IsoFileLabel.Left);
        }

        private void DeviceSelector_IndexChanged(object sender, EventArgs args)
        {
            if (m_drive != null)
            {
                m_drive.Dispose();
                m_drive = null;
            }

            if (m_device != null)
            {
                m_device.Dispose();
                m_device = null;
            }

            string dname = (string)DeviceSelector.SelectedItem;
            m_device = new Device(m_logger);
            if (!m_device.Open(dname[0]))
            {
                m_device = null;
                Status.Text = "Error opening drive";
                return;
            }

            m_drive = new Drive(m_device);
            DiskOperationError status = m_drive.Initialize();
            if (status != null)
            {
                Status.Text = status.Message;
                m_drive.Dispose();
                m_device.Dispose();
                m_device = null;
                m_drive = null;
                return;
            }

            Status.Text = m_drive.StatusString;
            toolTip1.SetToolTip(DeviceSelector, m_drive.StatusString);

            SpeedInfo[] speeds;
            if (m_drive.GetWriteSpeeds(out speeds) == null && speeds.GetLength(0) != 0)
                m_burn_speed = speeds[0];
            else
                m_burn_speed = null;
        }

        private void BrowseForFile_Click(object sender, EventArgs e)
        {
            OpenFileDialog dialog = new OpenFileDialog();
            dialog.Multiselect = false;
            dialog.Filter = "Disk Image Files (*.iso)|*.iso|All files (*.*)|*.*";
            dialog.RestoreDirectory = true;

            if (dialog.ShowDialog() == DialogResult.OK)
                IsoFileName.Text = dialog.FileName;
        }

        private void Done(DiskOperationError status)
        {
            m_timer.Stop();

            DeviceSelector.Enabled = true;
            Record.Enabled = true;
            IsoFileName.Enabled = true;
            BrowseForFile.Enabled = true;
            Settings.Enabled = true;

            string str = "";
            string caption = "Error!";
            MessageBoxIcon icon = MessageBoxIcon.Error;

            if (status == null)
            {
                str = "The burn operation completed sucessfully\r\n";
                str += "Target Speed: " + m_burn_speed + "\r\n";
                str += "Actual Speed: " + m_drive.BurnSpeed;
                caption = "Operation Complete";
                icon = MessageBoxIcon.Information;
            }
            else
            {
                str = "The burn operation failed - " + status.Message;
            }
            MessageBox.Show(str, caption, MessageBoxButtons.OK, icon);
        }

        //
        // This is basically the writer thread
        //
        private void RecordProc(object param)
        {
            DiskOperationError status = null;

            m_drive.ProgressChanged += new ProgressChangedEventHandler(ProgressHandler);

            // Erase if necessary
            if (param is bool && (bool)param == true)
                m_drive.Erase(Drive.EraseType.Minimal);

            FileInfo finfo = new FileInfo(m_isofile);
            if (finfo.Length / 2048 < m_drive.WritableCapacityInSectors)
            {
                // Create the disk image to burn
                DiskBurnImage dsk = new DiskBurnImage(m_logger);
                FileDataSource src = new FileDataSource(m_isofile);
                TrackBurnImage t = new TrackBurnImage(TrackBurnImage.TrackType.Data_Mode_1, src);
                dsk.AddTrack(t);

                // Burn the disk image onto the media in the drive
                status = m_drive.BurnDisk(m_burn_type, dsk, m_multi_session, m_burn_speed);

                // If we burned sucessfully, eject the disk
                if (status == null)
                    m_device.StartStopUnit(false, Device.PowerControl.NoChange, Device.StartState.EjectDisc);
            }
            else
            {
                string str;

                str = "the image selected '" + m_isofile + "' is larger than the writable space on the disk (" ;
                str += m_drive.WritableCapacityInSectors.ToString() + " sectors)" ;
                status = new BurnerError(str);
            }

            m_drive.ProgressChanged -= new ProgressChangedEventHandler(ProgressHandler);

            object[] args = new object[1];
            args[0] = status;
            Invoke(new DoneDelegate(Done), args);
        }

        private void Record_Click(object sender, EventArgs e)
        {
            bool erase = false;

            m_isofile = IsoFileName.Text;

            if (m_isofile.Length == 0)
            {
                MessageBox.Show("You did not select an ISO file.  Please select a valid ISO file to record to disk", "Error!",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            else if (!File.Exists(m_isofile))
            {
                string str = "The file '" + m_isofile + "' does not exist.  Please select a valid ISO file to burn";
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            DiskOperationError status = m_drive.Initialize();
            if (status != null)
            {
                MessageBox.Show(status.Message, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            if (!m_drive.IsBurnSupported)
            {
                MessageBox.Show("The media loaded in the drive is not supported!", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            if (m_drive.CurrentProfile == Drive.SupportedProfiles.CD_RW ||
                m_drive.CurrentProfile == Drive.SupportedProfiles.DVD_RW_RestrictedOverwrite ||
                m_drive.CurrentProfile == Drive.SupportedProfiles.DVD_RW_SequentialRecording)
            {
                DiscInformation info;
                if (m_device.ReadDiscInformation(out info) == Device.CommandStatus.Success)
                {
                    if (info.DiscStatus != DiscInformation.DiscStatusType.EmptyDisc)
                    {
                        DialogResult res;

                        res = MessageBox.Show("The disk in the drive is not erased.  Do you wish to erase the disk?", "Erase", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                        if (res == DialogResult.No)
                        {
                            MessageBox.Show("Insert a blank disk and press 'Burn Disc' again.");
                            return;
                        }
                        erase = true;
                    }
                }
            }

            Record.Enabled = false;
            DeviceSelector.Enabled = false;
            IsoFileName.Enabled = false;
            BrowseForFile.Enabled = false;
            Settings.Enabled = false;

            m_burner_thread = new Thread(new ParameterizedThreadStart(RecordProc));
            m_burner_thread.Start(erase);

            m_start = DateTime.Now;
            m_timer.Interval = 1000;
            m_timer.Tick += new EventHandler(TimerTickHandler);
            m_timer.Tag = this;
            m_timer.Start();
        }

        private void TimerTickHandler(object sender, EventArgs args)
        {
            TimerTickCallback();
        }

        private void TimerTickCallback()
        {
            if (InvokeRequired)
            {
                Invoke(new TimerTickDelegate(TimerTickCallback));
            }
            else
            {
                TimeSpan elapsed = DateTime.Now - m_start;
                ElapsedTime.Text = "Elapsed Time: " + elapsed.Minutes.ToString() + ":" + elapsed.Seconds.ToString("D2");

                if (m_drive.BufferLength != 0)
                {
                    BufferLevel.Maximum = (int)m_drive.BufferLength;
                    if (m_drive.BufferAvailable <= m_drive.BufferLength)
                        BufferLevel.Value = (int)(m_drive.BufferLength - m_drive.BufferAvailable);
                }
            }
        }

        private void ProgressHandler(object sender, ProgressChangedEventArgs args)
        {
            if (InvokeRequired)
            {
                object[] param = new object[2];
                param[0] = sender;
                param[1] = args;
                Invoke(new ProgressHandlerDelegate(ProgressHandler), param);
            }
            else
            {
                ProgressControl.Value = args.ProgressPercentage;
                if (args.UserState is ProgressText)
                {
                    ProgressText pt = (ProgressText)args.UserState ;
                    ProgressControl.Text = pt.LongName;
                    TopLevelControl.Text = pt.ShortName + " ( " + args.ProgressPercentage.ToString() + "% )";
                }
                else if (args.UserState is string)
                {
                    ProgressControl.Text = (string)args.UserState;
                }
            }
        }

        private void Settings_Click(object sender, EventArgs e)
        {
            bool warning, debug;

            if (m_logger.HasSink(UserMessage.Category.Debug))
                debug = true;
            else
                debug = false;

            if (m_logger.HasSink(UserMessage.Category.Warning))
                warning = true;
            else
                warning = false;

            int level = m_logger.GetLevel(UserMessage.Category.Debug);

            SpeedInfo[] speeds;
            DiskOperationError status = m_drive.GetWriteSpeeds(out speeds);
            if (status != null)
            {
                string str = "Cannot retreive speed data from drive - " + status.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            DriveSettings form = new DriveSettings(debug, level, warning, m_multi_session, m_burn_type, speeds, 
                                                   m_drive.IsTrackAtOnceSupported, m_drive.IsSessionAtOnceSupported,
                                                   m_drive.IsMultiSessionSupported);
            if (form.ShowDialog(this) == DialogResult.OK)
            {
                m_multi_session = form.LeaveSessionOpen;
                if (form.DebugMessagesFlag)
                {
                    m_logger.SetSink(UserMessage.Category.Debug, m_sink);
                    m_logger.SetLevel(UserMessage.Category.Debug, form.DebugLevel);
                }
                else
                    m_logger.RemoveSink(UserMessage.Category.Debug);

                if (form.WarningMessagesFlag)
                    m_logger.SetSink(UserMessage.Category.Warning, m_sink);
                else
                    m_logger.RemoveSink(UserMessage.Category.Warning);

                m_burn_type = form.BurnType;
                m_burn_speed = form.BurnSpeedValue;

                SetStartupBool(PropertyNameLoggerDebugMessages, m_logger.HasSink(UserMessage.Category.Debug));
                SetStartupInteger(PropertyNameLoggerDebugLevel, m_logger.GetLevel(UserMessage.Category.Debug));
                SetStartupBool(PropertyNameLoggerWarningMessages, m_logger.HasSink(UserMessage.Category.Warning));
                SetStartupInteger(PropertyNameBurnIsoImageBurnType, (int)m_burn_type);
                SetStartupBool(PropertyNameBurnIsoImageMultiSession, m_multi_session);
            }
        }
    }
}