//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using Bwg.Audio ;
using Bwg.Burner ;
using Bwg.Scsi ;

namespace Bwg.Burner
{
    /// <summary>
    /// This class represents the data source for a track of audio data.  This class acts as a
    /// go between for an audio reader and the data source required for burning to disk.
    /// </summary>
    public class AudioDataSource : DataSource
    {
        #region private data members
        private AudioReader m_reader;
        #endregion

        #region constructors

        /// <summary>
        /// Create an audio data source given an audio file reader.  The gap for this
        /// source is set to a fixed 2 seconds
        /// </summary>
        /// <param name="rdr">the reader for this audio source</param>
        public AudioDataSource(AudioReader rdr)
        {
            m_reader = rdr;
        }

        #endregion

        #region public properties

        /// <summary>
        /// Returns the audio reader associated with the audio data source
        /// </summary>
        public AudioReader Reader
        {
            get
            {
                return m_reader;
            }
        }

        /// <summary>
        /// Returns the name of the audio source, which is the name of the file
        /// </summary>
        public override string Name
        {
            get
            {
                return Path.GetFileName(m_reader.Name);
            }
        }

        /// <summary>
        /// Return the string representing the file type from the reader
        /// </summary>
        public string FileTypeString
        {
            get
            {
                return m_reader.FileTypeString;
            }
        }

        /// <summary>
        /// Return the size of the source in sectors.
        /// </summary>
        public override long SectorCount
        {
            get
            {
                // This is the number of samples (at 44100 per second, 16 bits, stereo)
                long bytecount = m_reader.ByteCount;

                // Find the number of sectors
                long sectors = bytecount / 2352;
                if ((bytecount % 2352) != 0)
                    sectors++;

                return sectors;
            }
        }

        #endregion

        #region public methods

        /// <summary>
        /// Send the data associated with the audio reader to the WriteBufferPool and therefore
        /// ultimately to the burner
        /// </summary>
        /// <param name="pool">the buffer pool that manages pages of data going to the CD/DVD burner</param>
        /// <param name="start_sector">the starting sector # for this data</param>
        /// <param name="sector_size">the size of each sector in bytes</param>
        /// <param name="lba">the logical block address to start the audio data</param>
        /// <returns>the total number of sector sent to the buffer pool (and device)</returns>
        public override long SendData(WriteBufferPool pool, long start_sector, int sector_size, long lba)
        {
            Debug.Assert(sector_size == 2352);

            int sectors = pool.PageSize / sector_size;
            WriteBuffer buf = null;
            int index = 0;
            long total = 0;
            long totalcount = 0;
            bool first = true;

            while (true)
            {
                if (buf == null)
                {
                    buf = pool.GetFreeBuffer();
                    index = 0;
                }

                //
                // Read 1176 ushorts from the file (or 588 samples)
                //
                IntPtr dest = new IntPtr((int)(buf.BufferPtr.ToInt32() + index));
                int cnt = m_reader.Read(dest, sector_size);
                totalcount += cnt;
                if (cnt == 0)
                {
                    Debug.Assert(totalcount == m_reader.ByteCount);

                    // We are done, send whatever is remaining to the device
                    buf.SectorSize = sector_size;
                    buf.DataSize = index / sector_size;

                    total += buf.DataSize;
                    pool.SendBufferToDevice(buf);
                    break;
                }

                // Increment the count
                index += cnt;
                if (index == sectors * sector_size)
                {
                    buf.SectorSize = sector_size;
                    buf.DataSize = index / sector_size;
                    if (first)
                    {
                        buf.LogicalBlockAddress = lba;
                        buf.SourceString = "AudioDataSource first sector";
                        first = false;
                    }
                    else
                    {
                        buf.SourceString = "AudioDataSource NOT first sector";
                        buf.LogicalBlockAddress = long.MaxValue;
                    }

                    total += buf.DataSize;
                    pool.SendBufferToDevice(buf);
                    buf = null;
                }
            }

            return total;
        }
        #endregion
    }
}
