using System;
using System.Collections.Generic;
using System.Text;
using Bwg.Scsi;
using Bwg.Logging;

namespace Bwg.Burner
{
    /// <summary>
    /// This class contains the required disk operations for a given type of media.  Specifically, this class
    /// supports the BluRay-RE (re-writable) media.
    /// 
    /// TODO: Handle setting write protect
    /// TODO: Handle emergency brake
    /// </summary>
    class BDREDiskOperations : DiskOperations
    {
        #region private member variables
        /// <summary>
        /// The write protect password for the disk
        /// </summary>
        private string m_password;

        /// <summary>
        /// The capacity of the drive
        /// </summary>
        private long m_capacity;
        #endregion

        #region constructor
        public BDREDiskOperations(Drive d)
            : base(d)
        {
            m_capacity = -1;
        }
        #endregion

        #region public properties

        public override bool IsFormatSupported
        {
            get
            {
                return true;
            }
        }

        public override long CapacityInSectors
        {
            get
            {
                //
                // Read the disk formatted capacities, if this fails, assume a normal DVD+RW
                // disk that is empty.
                //
                IList<CapacityDescriptor> caplist;
                if (MyDevice.ReadFormatCapacities(true, out caplist) != Device.CommandStatus.Success)
                    return Drive.D25GigBDRECapacity;

                //
                // Find the unformatted format descriptor, and return the given number of
                // blocks.
                //
                if (caplist[0].DescType == CapacityDescriptor.DescriptorType.Unformatted || caplist[0].DescType == CapacityDescriptor.DescriptorType.Formatted)
                    return caplist[0].NumberOfBlocks;


                return Drive.D25GigBDRECapacity;
            }
        }

        /// <summary>
        /// Returns the writable capacity in sectors.  This is the same as the capacity in sectors
        /// on RW media.
        /// </summary>
        public override long WritableCapacityInSectors
        {
            get
            {
                if (m_capacity == -1)
                    m_capacity = CapacityInSectors;

                return m_capacity;
            }
        }

        /// <summary>
        /// Returns the remaining capacity on the disk.
        /// </summary>
        public override long RemainingCapacityInWritableSectors
        {
            get
            {
                DiscInformation info;

                //
                // Since we don't support multi-session burning for DVD+RW yet, a disk is
                // either erased and the capacity is the entire disk, or the disk is used
                // and the capacity is zero.
                //
                if (MyDevice.ReadDiscInformation(out info) != Device.CommandStatus.Success)
                    return 0;

                if (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
                    return WritableCapacityInSectors;

                return 0;
            }
        }

        /// <summary>
        /// Returns true if write protect is supports (and it is for this media)
        /// </summary>
        public override bool IsWriteProtectSupported
        {
            get
            {
                return true;
            }
        }

        public override bool IsWriteProtected
        {
            get
            {
                return false;
            }
        }

        public override bool HasWriteProtectPassword
        {
            get
            {
                return false;
            }
        }

        public override string WriteProtectPassword
        {
            set
            {
                m_password = value;
            }
        }

        /// <summary>
        /// Returns true if burning is supported
        /// </summary>
        public override bool IsBurnSupported
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Returns true if multi-session is supported (false for now)
        /// </summary>
        public override bool IsMultiSessionSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Returns true if erase is supported, we do not erase DVD+RW media, we just write our
        /// image starting at sector 0.  Any old image is over-written and the drive will begin
        /// formatting automatically if necessary to have the available space.
        /// </summary>
        public override bool IsEraseSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Returns true if session at once burning is supported
        /// </summary>
        public override bool IsSessionAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDSessionAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        /// <summary>
        /// Returns true if track at once burning is supported
        /// </summary>
        public override bool IsTrackAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDTrackAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }


        #endregion

        #region public methods
        public override DiskOperationError Erase(Drive.EraseType et)
        {
            return null;
        }

        public override DiskOperationError BurnDisk(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            DiskOperationError status;
            Device.CommandStatus st;

            bytes = 0;

            if (multi_session)
                return new BurnerError("Multi-session burning of BD-RE is currently not supported");

            if (d.TrackCount != 1)
                return new BurnerError("multiple tracks targeted at a BD-RE media, BD-RE only supports a single track");

            if (d.ContainsAudio)
                return new BurnerError("cannot write audio tracks to a BD-RE media, BD-RE only supports data tracks");

            DiscInformation info;
            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            status = m_drive.SetSpeed(speed);
            if (status != null)
                return status;

            //
            // Set the target media, this sets the start sector for the disk as well as all of
            // the pre-track gap information
            //
            d.StartSector = 0;
            d.TargetMedia = DiskBurnImage.DiskTargetType.DVD;

            bool doformat = false;
            if (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
                doformat = true;

            if (doformat)
            {
                // status = FormatMoreRoom();
                // if (status != null)
                    // return status;
            }

            // Create a data generator for this stream of data.  For DVD is it easy.  There is always one
            // track that is a stream of 2048 byte sectors

            DataGenerator gen = new DataGenerator(MyDevice.GetLogger(), 64, MyDevice.MaximumTransferLength, 0, 0, 0);
            gen.AddSource("data", d[0], 0, d[0].DataToSend);

            status = m_drive.SendData(gen, d);
            bytes += (ulong)(d.LengthSectors * 2048);
            if (status != null)
                return status;

            status = m_drive.SynchronizeCache();
            if (status != null)
                return status;

            m_drive.ReadBufferLevels();

            return null;
        }
        #endregion
    }
}
