//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Bwg.Scsi;
using Bwg.Logging;

namespace Bwg.Burner
{
    /// <summary>
    /// This class is a base class for operations associated with both the
    /// CD/R and CD/RW media types.  Operations common to both the CD/R and the
    /// CD/RW media types are implemented here.
    /// 
    /// This class supports burndisk with the following features:
    /// 
    /// - Session At Once (also know as Disk At Once)
    /// - Track At once
    /// - Burn leaving disk open for additional sessions
    /// - Burn finalizing disk for maximum compatibility
    /// - Set burn speed
    /// 
    /// </summary>
    public abstract class CDR_RWDiskOperations : DiskOperations
    {
        #region public constants
        /// <summary>
        /// The minimum size for a track on a CD-R or CD-RW disk
        /// </summary>
        public static long MinimumTrackLengthInSectors = 4 * MinuteSecondFrame.FramesPerSecond;

        /// <summary>
        /// The default size of a pre-gap
        /// </summary>
        public static long DefaultPregapLengthInSectors = 2 * MinuteSecondFrame.FramesPerSecond;
        #endregion

        #region constructor

        /// <summary>
        /// Constructor for this immerdiate level, just calls the base class
        /// constructor with the drive object
        /// </summary>
        /// <param name="d"></param>
        public CDR_RWDiskOperations(Drive d) : base(d)
        {
        }
        #endregion

        #region public properties

        /// <summary>
        /// Returns false as format is currently not supported for CDR or CDRW
        /// </summary>
        public override bool IsFormatSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Returns true if write protect is supports (and it is for this media)
        /// </summary>
        public override bool IsWriteProtectSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// This property returns true if the loaded disk is write protected
        /// </summary>
        public override bool IsWriteProtected
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// This property returns true if the loaded disk has a write protect password
        /// </summary>
        public override bool HasWriteProtectPassword
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// This property provides the write protect password for a given media
        /// </summary>
        public override string WriteProtectPassword
        {
            set
            {
                throw new Exception("Cannot supply write protect password for this type of media");
            }
        }
        #endregion

        #region public methods

        /// <summary>
        /// This method burns a disk onto CDRW media.  The disk image is given by the
        /// object d and the burn type indicates a disk at once versus a track at once
        /// burn.
        /// </summary>
        /// <param name="bt">type of the burn (track vs session)</param>
        /// <param name="d">the image to burn to a disk</param>
        /// <param name="multi_session">if true, leave the disk open for multiple sessions</param>
        /// <param name="speed">the speed of the burn operation</param>
        /// <param name="bytes">output, the number of bytes burned</param>
        /// <returns>error object is an error occurs, otherwise null</returns>
        public override DiskOperationError BurnDisk(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            DiskOperationError status ;

            if (bt == Drive.BurnType.DontCare)
            {
                if (m_drive.IsSessionAtOnceSupported)
                    status = BurnDiskSession(d, multi_session, speed, out bytes);
                else
                    status = BurnDiskTrack(d, multi_session, speed, out bytes);
            }
            else if (bt == Drive.BurnType.Session)
                status = BurnDiskSession(d, multi_session, speed, out bytes);
            else
                status = BurnDiskTrack(d, multi_session, speed, out bytes);

            return status;
        }
        #endregion

        #region private methods
        private DiskOperationError BurnDiskTrack(DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            DiskOperationError status ;
            Device.CommandStatus st;

            bytes = 0;
            MyDevice.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, "burning disk using track at once"));

            DiscInformation info;
            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            bool proof;
            status = m_drive.GetBurnProofSetting(out proof);
            if (status != null)
                return status;

            if (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
            {
                //
                // Set the target media, this sets the start sector for the disk as well as all of
                // the pre-track gap information
                //
                d.StartTrack = 1;
            }
            else
            {
                d.StartTrack = info.FirstTrackInLastSession;
            }

            d.TargetMedia = DiskBurnImage.DiskTargetType.CD_TAO;

            //
            // Send the OPC information, note that this command is optional so if it fails, that is ok,
            //
            MyDevice.LogSenseState = false;
            st = MyDevice.SendOpcInformation(false, null);
            MyDevice.LogSenseState = true;

            status = m_drive.SetSpeed(speed);
            if (status != null)
                return status;

            //
            // Write each of the tracks
            //
            for (int i = 0; i < d.TrackCount; i++)
            {
                ulong count;
                TrackBurnImage t = d[i];
                status = WriteTrack(t, proof, multi_session, out count);
                bytes += count;
                if (status != null)
                    return status;
            }

            //
            // Close the current session
            //
            return m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
        }

        private DiskOperationError WriteTrack(TrackBurnImage t, bool proof, bool multi_session, out ulong count)
        {
            Device.CommandStatus st;
            ModeTable table;
            DiskOperationError status ;

            count = 0;

            st = MyDevice.ModeSense(Device.PageControl.Current, 5, out table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            //
            // Setup the mode page for write track at once
            //
            Debug.Assert(table.Pages[0] is WriteParameterModePage);
            WriteParameterModePage pg = (WriteParameterModePage)table.Pages[0];

            pg.ParametersSavable = false;
            pg.BurnProof = proof;
            pg.WriteType = WriteParameterModePage.WriteTypeType.TrackAtOnce;

            if (t.Type == TrackBurnImage.TrackType.Audio)
            {
                if (t.PreEmphasis)
                    pg.TrackMode = WriteParameterModePage.TrackModeType.TwoChannelAudioWithPreemphasis ;
                else
                    pg.TrackMode = WriteParameterModePage.TrackModeType.TwoChannelAudio ;

                pg.DataBlockType = WriteParameterModePage.DataBlockTypeType.RawData;
            }
            else
            {
                pg.TrackMode = WriteParameterModePage.TrackModeType.DataUninterrupted;
                pg.DataBlockType = WriteParameterModePage.DataBlockTypeType.DataMode1;
            }

            if (multi_session)
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NextSessionAllowed;
            else
                pg.MultiSession = WriteParameterModePage.MultiSessionType.CDNoNextSessionFFFF;

            pg.SessionFormat = WriteParameterModePage.SessionFormatType.CDDA_CDROM;

            st = MyDevice.ModeSelect(false, table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            //
            // Find the start address for the new track
            //
            TrackInformation info;
            st = MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, (uint)t.TrackNumber, out info);
            if (st != Device.CommandStatus.Success)
            {
                st = MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, Device.InvisibleTrack, out info);
                if (st != Device.CommandStatus.Success)
                    return CreateErrorObject(st);
            }

            //
            // Send the data to the drive
            //
            DataGenerator gen = new DataGenerator(MyDevice.GetLogger(), 64, MyDevice.MaximumTransferLength, 0, 0, 0);
            long length = t.DataToSend;
            if (length < MinimumTrackLengthInSectors)
                length = MinimumTrackLengthInSectors;
            gen.AddSource("track_data", t, info.TrackStartAddress, length);
            status = m_drive.SendData(gen, t);
            count = (ulong)(t.SectorSize * t.TotalLengthSectors);
            if (status != null)
                return status;

            //
            // Close the track
            //
            return m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseTrack, (ushort)t.TrackNumber);
        }

        private DiskOperationError BurnDiskSession(DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            DiskOperationError status;
            Device.CommandStatus st;
            byte firsttrack;
            long cdtext_start ;

            bytes = 0;

            MyDevice.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, "burning disk using session at once"));
            if (!m_drive.IsSessionAtOnceSupported)
                return new BurnerError("the drive '" + m_drive.DriveModel + "' does not support session at once burning");

            status = m_drive.MakeDriveReady();
            if (status != null)
                return status;

            DiscInformation info;
            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            if (info.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc)
                return new BurnerError("the media has been finalized without the option of new sessions");

            if (info.DiscStatus == DiscInformation.DiscStatusType.IncompleteDisc)
            {
                //
                // Find the next writable address for the next session as the starting point
                // for the write operation
                //
                TrackInformation tinfo;
                st = MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, Device.InvisibleTrack, out tinfo);
                if (st != Device.CommandStatus.Success)
                    return CreateErrorObject(st);

                if (!tinfo.NWA_V)
                    return new BurnerError("cannot retreive next writable address from disk");

                d.StartSector = tinfo.TrackStartAddress - 150 ;
                firsttrack = (byte)info.FirstTrackInLastSession;
                cdtext_start = long.MaxValue;
            }
            else
            {
                //
                // Set the target media, this sets the start sector for the disk as well as all of
                // the pre-track gap information
                //
                d.StartSector = -150 ;
                firsttrack = 1;
                cdtext_start = info.LastSessionLeadInStartAddress.LogicalBlockAddress;
            }

            //
            // Set the disk to session at once burning.  This just updates the image of the media
            // to be written to be compatible with session at once writing
            //
            d.TargetMedia = DiskBurnImage.DiskTargetType.CD_SAO;

            bool proof;
            status = m_drive.GetBurnProofSetting(out proof);
            if (status != null)
                return status;

            ModeTable table;
            st = MyDevice.ModeSense(Device.PageControl.Current, 5, out table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            Debug.Assert(table.Pages[0] is WriteParameterModePage);
            WriteParameterModePage pg = (WriteParameterModePage)table.Pages[0];

            pg.ParametersSavable = false;
            pg.BurnProof = proof;
            if (multi_session)
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NextSessionAllowed;
            else
                pg.MultiSession = WriteParameterModePage.MultiSessionType.CDNoNextSessionFFFF;

            pg.WriteType = WriteParameterModePage.WriteTypeType.SessionAtOnce;
            pg.DataBlockType = WriteParameterModePage.DataBlockTypeType.DataMode1;
            pg.TrackMode = WriteParameterModePage.TrackModeType.DataUninterrupted;
            pg.SessionFormat = WriteParameterModePage.SessionFormatType.CDDA_CDROM;

            st = MyDevice.ModeSelect(false, table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            MyDevice.LogSenseState = false;
            st = MyDevice.SendOpcInformation(false, null);
            MyDevice.LogSenseState = true;

            status = m_drive.SetSpeed(speed);
            if (status != null)
                return status;

            DataGenerator gen = new DataGenerator(MyLogger, 64, MyDevice.MaximumTransferLength, 0, 0, 0);
            byte[] sheet = d.CreateCueSheet(firsttrack, ref gen, cdtext_start);
            st = MyDevice.SendCueSheet(sheet);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            Debug.Assert(d.StartSector < Int32.MaxValue);
            int lba = (int)d.StartSector;
            status = m_drive.SendData(gen, d);

            //
            // We don't deal with audio/data mixed CDs yet
            //
            if (d.IsDiskPureAudio)
                bytes = (ulong)(d.LengthSectors * 2352);
            else
                bytes = (ulong)(d.LengthSectors * 2048);

            if (status != null)
            {
                m_drive.Device.CloseTrackSession(false, Device.CloseTrackSessionType.CloseSession, 0);
                return status;
            }

            //
            // We are going to test unit ready until the writing of data is
            // done,  do not log all of the sense messages that are tell us
            // that the drive is not ready
            //
            MyDevice.LogSenseState = false;
            while (true)
            {
                bool ready;

                st = MyDevice.TestUnitReady(out ready);
                if (ready)
                    break;
            }
            MyDevice.LogSenseState = true;

            status = m_drive.SynchronizeCache();
            if (status != null)
                return status;

            return m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
        }
        #endregion
    }
}
