//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using Bwg.Scsi;

namespace Bwg.Burner
{
    /// <summary>
    /// This is the set of operations for a CD-ROM media
    /// 
    /// This class supports almost nothing as there are no burner operations that
    /// can be performed on CD-ROM media.
    /// 
    /// </summary>
    public class CDRomDiskOperations : DiskOperations
    {
        #region constructor
        /// <summary>
        /// Create the object
        /// </summary>
        /// <param name="d"></param>
        public CDRomDiskOperations(Drive d) : base(d)
        {
        }
        #endregion

        #region public properties

        /// <summary>
        /// Return true if format operation is supported
        /// </summary>
        public override bool IsFormatSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Returns true if write protect is supports (and it is for this media)
        /// </summary>
        public override bool IsWriteProtectSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// This property returns true if the loaded disk is write protected
        /// </summary>
        public override bool IsWriteProtected
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// This property returns true if the loaded disk has a write protect password
        /// </summary>
        public override bool HasWriteProtectPassword
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// This property provides the write protect password for a given media
        /// </summary>
        public override string WriteProtectPassword
        {
            set
            {
                throw new Exception("Cannot supply write protect password for this type of media");
            }
        }

        /// <summary>
        /// This property returns the capacity of the CDROM in sectors.  This property is the total number of
        /// sectors on the media, writable or not
        /// </summary>
        public override long CapacityInSectors
        {
            get
            {
                DiscInformation info;
                Device.CommandStatus st;

                st = MyDevice.ReadDiscInformation(out info);
                if (st != Device.CommandStatus.Success)
                    return 0;

                return info.LastPossibleLeadoutStartAddress.SectorCount;
            }
        }

        /// <summary>
        /// This property returns the writable capacity in sectors, 0 for CD-ROM media
        /// </summary>
        public override long WritableCapacityInSectors
        {
            get
            {
                return 0;
            }
        }

        /// <summary>
        /// This property returns the writable capacity in sectors remaining on the disk, 0 for
        /// CD-ROM media.
        /// </summary>
        public override long RemainingCapacityInWritableSectors
        {
            get
            {
                return 0;
            }
        }

        /// <summary>
        /// Always returns false as burning is not supported on CD-ROM media
        /// </summary>
        public override bool IsBurnSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Always returns false as erasing is not supported on CD-ROM media
        /// </summary>
        public override bool IsEraseSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Always returns false as burning is not supported on CD-ROM media
        /// </summary>
        public override bool IsMultiSessionSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Always returns false as burning is not supported on CD-ROM media
        /// </summary>
        public override bool IsSessionAtOnceSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Always returns false as burning is not supported on CD-ROM media
        /// </summary>
        public override bool IsTrackAtOnceSupported
        {
            get
            {
                return false;
            }
        }
        #endregion

        #region public methods

        /// <summary>
        /// This is the burn disk operation for a CD/ROM disk.  It basically is a place holder that generates
        /// an error if you get here.
        /// </summary>
        /// <param name="bt"></param>
        /// <param name="d"></param>
        /// <param name="multi_session">if true, leave the disk open for future sessions, not supported here</param>\
        /// <param name="speed">the speed of the burn operation</param>
        /// <param name="bytes">output, the number of bytes burnred</param>
        /// <returns></returns>
        public override DiskOperationError BurnDisk(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            bytes = 0;
            return NotSupported("burndisk");
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="et"></param>
        /// <returns></returns>
        public override DiskOperationError Erase(Drive.EraseType et)
        {
            return NotSupported("erase");
        }
        #endregion
    }
}
