using System;
using System.Collections.Generic;
using System.Text;

namespace Bwg.Burner
{
    /// <summary>
    /// This class represents string data as the data associated with a CD text field
    /// entry.
    /// </summary>
    [Serializable()]
    public class CDTextDataString : CDTextData
    {
        #region private member variables
        /// <summary>
        /// The text of the CD text data field
        /// </summary>
        string m_data ;

        /// <summary>
        /// The encoding to use for the text, null means ASCII
        /// </summary>
        Encoding m_encoding;

        #endregion

        #region constructor
        /// <summary>
        /// Create a data entry value based on a string
        /// </summary>
        /// <param name="str">string value</param>
        /// <param name="enc">encoding to be used, null means ASCII</param>
        public CDTextDataString(string str, Encoding enc)
        {
            m_data = str;
            m_encoding = enc;
        }

        /// <summary>
        /// Create a data entry containing an empty string and no encoding
        /// </summary>
        public CDTextDataString()
        {
            m_data = string.Empty;
            m_encoding = null;
        }
        #endregion

        #region public properties
        /// <summary>
        /// The text that is the value of the data associated with a cd text field.
        /// </summary>
        public string Text
        {
            get
            {
                return m_data;
            }
        }

        /// <summary>
        /// The length of the wrapped string in characters
        /// </summary>
        public int Length
        {
            get
            {
                return m_data.Length;
            }
        }

        /// <summary>
        /// This property returns the array of bytes that represents this string field
        /// </summary>
        public override byte[] Data
        {
            get
            {
                Encoding enc = m_encoding;
                if (enc == null)
                    enc = Encoding.ASCII;

                byte [] data = enc.GetBytes(m_data);
                byte[] retdata = new byte[data.GetLength(0) + 1];
                Array.Copy(data, retdata, data.GetLength(0));

                return retdata;
            }
        }

        /// <summary>
        /// This property returns the size of the data in bytes
        /// </summary>
        public override int DataLength
        {
            get
            {
                Encoding enc = m_encoding;
                if (enc == null)
                    enc = Encoding.ASCII;

                return enc.GetByteCount(m_data);
            }
        }
        #endregion

        #region Public Methods

        /// <summary>
        /// Add a string to the existing string that is the data for this element.
        /// </summary>
        /// <param name="str">the string to add to the existing string</param>
        public void AddString(string str)
        {
            m_data += str;
        }

        /// <summary>
        /// Returns a string representation of the data value
        /// </summary>
        /// <returns>returns the string</returns>
        public override string ToString()
        {
            return m_data;
        }

        /// <summary>
        /// Add a set of data bytes to the string.
        /// </summary>
        /// <param name="pos">the position in the string</param>
        /// <param name="data">the data bytes to add</param>
        public override void AddData(byte pos, byte[] data)
        {
            Encoding enc = m_encoding;
            if (enc == null)
                enc = Encoding.ASCII;

            int eos = -1;
            for (int i = 0; i < data.GetLength(0); i++)
            {
                if (data[i] == 0)
                {
                    eos = i;
                    break;
                }
            }

            if (eos == -1)
                m_data += enc.GetString(data);
            else
                m_data += enc.GetString(data, 0, eos + 1);
        }
        #endregion
    }
}
