using System;
using System.Collections.Generic;
using System.Text;

namespace Bwg.Burner
{
    class CDTextEncoder
    {
        #region private member variables
        private IList<CDTextField> m_disk;
        private IList<CDTextField> [] m_tracks;
        #endregion

        #region constructor
        public CDTextEncoder(IList<CDTextField> disk, int tracks)
        {
            m_disk = disk;
            m_tracks = new IList<CDTextField>[tracks];
        }
        #endregion

        #region public members
        public void SetTrackCDText(int track, IList<CDTextField> list)
        {
            m_tracks[track] = list;
        }

        public IList<CDTextBlock> Encode()
        {
            IList<CDTextBlock> blocks = null ;
            CDTextField.LanguageType[] langs = CollectLanguages();
            if (langs.GetLength(0) > 7)
            {
                string str;

                str = "There are too many languages specified with the various CD Text fields ";
                str += "assigned to the disk and its various tracks.  Detected languages are:";
                foreach (CDTextField.LanguageType lang in langs)
                    str += " " + lang.ToString();
                throw new Exception(str);
            }

            blocks = new List<CDTextBlock>() ;
            for (int i = 0; i < langs.GetLength(0); i++)
                PackOneLanguage(ref blocks, langs[i]);

            return blocks ;
        }
        #endregion

        #region private member functions

        /// <summary>
        /// Find a field given the list, the language, and the type.
        /// </summary>
        /// <param name="list">the list of fields to search</param>
        /// <param name="lang">the language we want</param>
        /// <param name="type">the type we want</param>
        /// <returns>the field if it exists, or null</returns>
        private CDTextField FindField(IList<CDTextField> list, CDTextField.LanguageType lang, CDTextField.TextFieldType type)
        {
            foreach (CDTextField field in list)
            {
                if (field.Language == lang && field.Type == type)
                    return field;
            }
            return null;
        }

        /// <summary>
        /// For a given language, process all of the tags for that langauge.  The tags are 
        /// processes in a specific order based on their type.
        /// </summary>
        /// <param name="blocks">the list of blocks that result from packing</param>
        /// <param name="lang">the language to process</param>
        private void PackOneLanguage(ref IList<CDTextBlock> blocks, CDTextField.LanguageType lang)
        {
            PackOneField(ref blocks, lang, CDTextField.TextFieldType.Title);
            PackOneField(ref blocks, lang, CDTextField.TextFieldType.Performer);
            PackOneField(ref blocks, lang, CDTextField.TextFieldType.Songwriter);
            PackOneField(ref blocks, lang, CDTextField.TextFieldType.Composer);
            PackOneField(ref blocks, lang, CDTextField.TextFieldType.Arranger);
            PackOneField(ref blocks, lang, CDTextField.TextFieldType.Message);
            PackOneField(ref blocks, lang, CDTextField.TextFieldType.DiskId);
            PackOneField(ref blocks, lang, CDTextField.TextFieldType.Genre);
            PackOneField(ref blocks, lang, CDTextField.TextFieldType.Isrc);
        }

        private void EncodeField(ref IList<CDTextBlock> blocks, int trackno, CDTextField field)
        {
        }

        private void PackOneField(ref IList<CDTextBlock> blocks, CDTextField.LanguageType lang, CDTextField.TextFieldType type)
        {
            CDTextField diskfield = FindField(m_disk, lang, type);
            if (diskfield != null)
                EncodeField(ref blocks, 0, diskfield);

            if (CDTextField.AppliesToTracks(type))
            {
                int count = 0;
                foreach (IList<CDTextField> list in m_tracks)
                {
                    if (FindField(list, lang, type) != null)
                        count++;
                }

                if (count > 0)
                {
                    //
                    // A special case for ISRC fields.  If any ISRC fields exist for tracks, we create one
                    // for the disk automatically as well
                    //
                    if (type == CDTextField.TextFieldType.Isrc && diskfield == null)
                    {
                        diskfield = new CDTextField(lang, type, string.Empty, null);
                        EncodeField(ref blocks, 0, diskfield);
                    }

                    int trackno = 1 ;
                    foreach (IList<CDTextField> list in m_tracks)
                    {
                        CDTextField trackfield = FindField(list, lang, type);
                        if (trackfield == null)
                            trackfield = diskfield;

                        if (trackfield != null)
                            EncodeField(ref blocks, trackno, trackfield);

                        trackno++;
                    }
                }
            }
        }

        /// <summary>
        /// Search the disk fields as well as the track fields and extract the languages
        /// in the set of all these fields
        /// </summary>
        /// <returns></returns>
        private CDTextField.LanguageType[] CollectLanguages()
        {
            CDTextField.LanguageType[] langs = new CDTextField.LanguageType[0];

            CollectLanguages(m_disk, ref langs) ;
            foreach (IList<CDTextField> list in m_tracks)
                CollectLanguages(list, ref langs);

            return langs;
        }

        /// <summary>
        /// Collect the languages references in the single list of CD text fields
        /// </summary>
        /// <param name="list">the list to process</param>
        /// <param name="langs">the array of languages found so far</param>
        private void CollectLanguages(IList<CDTextField> list, ref CDTextField.LanguageType[] langs)
        {
            foreach (CDTextField field in list)
            {
                bool found = false;

                foreach (CDTextField.LanguageType lang in langs)
                {
                    if (lang == field.Language)
                        found = true;
                }

                if (!found)
                {
                    int n = langs.GetLength(0);
                    Array.Resize(ref langs, n + 1);
                    langs[n] = field.Language;
                }
            }
        }
        #endregion
    }
}
