using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics ;

namespace Bwg.Burner
{
    /// <summary>
    /// This class decodes or encodes CD text information.  A CD Text packet is a 18 byte packet
    /// that is stored on the disk in the R - W subchannels.  This 18 byte packet is represented in this code
    /// as a CDTextBlock data structure which contains an internal 18 byte buffer.  Three bits of these
    /// eighteen bytes is an index that determines the language for the packet.  The index is an index into
    /// the table of languages that is stored in the sizeinfo packets in the cdtext area.  The value stored in
    /// the size info table is determined via the values given below as the LanguageType.
    /// 
    /// Packet (18 bytes)
    ///     Byte 0: Packet Type (see CDTextField.TextFieldType)
    ///     Byte 1: Track Number (unsigned byte)
    ///     Byte 2: Sequence Number (one sequence per language index)
    ///     Byte 3:
    ///         Bit 7     : Not sure what this bit does
    ///         Bit 6 - 4 : Three bit languange index
    ///         Bit 3 - 0 : Data Position
    ///     Byte 4: Payload byte 0
    ///     Byte 5: Payload byte 1
    ///     Byte 6: Payload byte 2
    ///     Byte 7: Payload byte 3
    ///     Byte 8: Payload byte 4
    ///     Byte 9: Payload byte 5
    ///     Byte 10: Payload byte 6
    ///     Byte 11: Payload byte 7
    ///     Byte 12: Payload byte 8
    ///     Byte 13: Payload byte 9
    ///     Byte 14: Payload byte 10
    ///     Byte 15: Payload byte 11
    ///     Byte 16: CRC Byte 0
    ///     Byte 17: CRC Byte 1
    /// </summary>
    public class CDTextEncoderDecoder
    {
        #region public types
        /// <summary>
        /// The language for this CD text field entry.
        /// </summary>
        public enum LanguageType : byte
        {
            /// <summary>
            /// Self evident
            /// </summary>
            CHINESE = 0x75,

            /// <summary>
            /// Self evident
            /// </summary>
            CZECH = 0x06,

            /// <summary>
            /// Self evident
            /// </summary>
            DANISH = 0x07,

            /// <summary>
            /// Self evident
            /// </summary>
            DUTCH = 0x1d,

            /// <summary>
            /// Self evident
            /// </summary>
            ENGLISH = 0x09,

            /// <summary>
            /// Self evident
            /// </summary>
            FINNISH = 0x27,

            /// <summary>
            /// Self evident
            /// </summary>
            FRENCH = 0x0f,

            /// <summary>
            /// Self evident
            /// </summary>
            GERMAN = 0x08,

            /// <summary>
            /// Self evident
            /// </summary>
            GREEK = 0x70,

            /// <summary>
            /// Self evident
            /// </summary>
            HUNGARIAN = 0x1b,

            /// <summary>
            /// Self evident
            /// </summary>
            ITALIAN = 0x15,

            /// <summary>
            /// Self evident
            /// </summary>
            JAPANESE = 0x69,

            /// <summary>
            /// Self evident
            /// </summary>
            KOREAN = 0x65,

            /// <summary>
            /// Self evident
            /// </summary>
            NORWEGIAN = 0x1e,

            /// <summary>
            /// Self evident
            /// </summary>
            POLISH = 0x20,

            /// <summary>
            /// Self evident
            /// </summary>
            PORTUGUESE = 0x21,

            /// <summary>
            /// Self evident
            /// </summary>
            RUSSIAN = 0x56,

            /// <summary>
            /// Self evident
            /// </summary>
            SLOVENE = 0x26,

            /// <summary>
            /// Self evident
            /// </summary>
            SPANISH = 0x0a,

            /// <summary>
            /// Self evident
            /// </summary>
            SWEDISH = 0x28
        } ;
        #endregion

        #region public constants
        /// <summary>
        /// The maximum number of language blocks allows for CD Text.
        /// </summary>
        public const byte MaximumLanguageBlocks = 8;
        #endregion

        #region private member variables

        private CDTextBlock m_last_field;

        private IList<CDTextField> m_disk;
        private IList<CDTextField> [] m_tracks;

        private LanguageType[] m_languages;
        private byte [] m_seqno;
        private byte [] m_packcnt;

        #endregion

        #region constructor
        /// <summary>
        /// Create a CDText encoder given the list of global text fields and the
        /// number of tracks on the disk
        /// </summary>
        /// <param name="disk">the cdtext fields assocaited with the disk (global)</param>
        /// <param name="tracks">the number of tracks on the disk</param>
        public CDTextEncoderDecoder(IList<CDTextField> disk, int tracks)
        {
            m_disk = disk;
            m_tracks = new IList<CDTextField>[tracks];
            m_languages = new LanguageType[0];
            m_seqno = new byte[8];
            m_packcnt = new byte[16];
        }

        /// <summary>
        /// Create a CDText decoder based on the cdtext information read from a disk and stored
        /// in a byte array.
        /// </summary>
        /// <param name="data">the array of cdtext data</param>
        public CDTextEncoderDecoder(byte[] data)
        {
            m_languages = new LanguageType[0];
            m_seqno = new byte[8];
            m_packcnt = new byte[16];

            m_disk = new List<CDTextField>();
            m_tracks = new IList<CDTextField>[0];
            DecodeCDText(data, 0, data.GetLength(0));
        }

        /// <summary>
        /// Create a CDText decoder based on the cdtext information read from a disk and stored in a
        /// byte array.
        /// </summary>
        /// <param name="data">the array of cdtext data</param>
        /// <param name="offset">the offset to the start of the data</param>
        public CDTextEncoderDecoder(byte[] data, int offset)
        {
            m_languages = new LanguageType[0];
            m_seqno = new byte[8];
            m_packcnt = new byte[16];

            m_disk = new List<CDTextField>();
            m_tracks = new IList<CDTextField>[0];
            DecodeCDText(data, offset, data.GetLength(0));
        }
        #endregion

        #region public properties
        /// <summary>
        /// This method returns true if this object contains no cd text for either any
        /// tracks of the main disc.
        /// </summary>
        public bool IsEmpty
        {
            get
            {
                if (m_disk.Count > 0)
                    return false;

                foreach (IList<CDTextField> textlist in m_tracks)
                    if (textlist.Count > 0)
                        return false;

                return true;
            }
        }

        /// <summary>
        /// Returns the number of languages stored in the CD text
        /// </summary>
        public int LanguageCount
        {
            get
            {
                return m_languages.GetLength(0);
            }
        }
        #endregion

        #region public member function

        /// <summary>
        /// This method removes CDText of a given type and language index (langidx) for 
        /// from a track
        /// </summary>
        /// <param name="trackno"></param>
        /// <param name="langidx"></param>
        /// <param name="type"></param>
        public void RemoveCDText(int trackno, int langidx, CDTextField.TextFieldType type)
        {
            IList<CDTextField> list = GetCDTextList(trackno);

            for (int i = 0; i < list.Count; i++)
            {
                CDTextField field = list[i];
                if (field.LanguageIndex == langidx && field.Type == type)
                {
                    list.RemoveAt(i);
                    return;
                }
            }
        }

        /// <summary>
        /// Return the list of CD text fields given the track #, where track # 0
        /// means global.
        /// </summary>
        /// <param name="index"></param>
        /// <returns></returns>
        public IList<CDTextField> GetCDTextList(int index)
        {
            if (index == 0)
                return m_disk;

            if (index <= m_tracks.GetLength(0))
                return m_tracks[index - 1];

            return null;
        }

        /// <summary>
        /// Return the language associated with the language index given
        /// </summary>
        /// <param name="index"></param>
        /// <returns></returns>
        public LanguageType GetLanguage(int index)
        {
            return m_languages[index];
        }

        /// <summary>
        /// Add a langauge to the list of langauges supported by the CD text encoder/decoder
        /// </summary>
        /// <param name="l"></param>
        public void AddLanguage(LanguageType l)
        {
            int len = m_languages.GetLength(0) ;

            Debug.Assert(len < 8) ;
            Array.Resize(ref m_languages, len + 1) ;
            m_languages[len] = l ;
        }

        /// <summary>
        /// This method sets the list of cdtext fields assocaited with a given
        /// track.
        /// </summary>
        /// <param name="track">the number of the track to set</param>
        /// <param name="list">the list of text field entries for this track</param>
        public void SetTrackCDText(int track, IList<CDTextField> list)
        {
            m_tracks[track - 1] = list;
        }

        /// <summary>
        /// Process each language in turn and encode the text field associated with the language
        /// </summary>
        /// <returns></returns>
        internal IList<CDTextBlock> Encode()
        {
            IList<CDTextBlock> blocks = new List<CDTextBlock>() ;
            for (byte i = 0; i < MaximumLanguageBlocks ; i++)
                PackOneLanguage(ref blocks, i);

            CreateSizeInfoBlock(ref blocks);

            return blocks ;
        }
        #endregion

        #region private member functions

        private void CreateSizeInfoBlock(ref IList<CDTextBlock> blocks)
        {
            byte[] data = new byte[36];
            int offset = 0;

            data[offset++] = 0;                                 // Character code
            data[offset++] = 1;                                 // First track
            data[offset++] = (byte)m_tracks.GetLength(0);       // Last track
            data[offset++] = 0;                                 // Copyright

            // Increment the packcount for the three we are going to add as
            // the sizeinfo pack
            m_packcnt[15] += 3;

            // Add the pack count information
            for (int i = 0; i < m_packcnt.GetLength(0); i++)
                data[offset++] = m_packcnt[i];

            // Add the last sequence # information
            for (int i = 0; i < m_seqno.GetLength(0); i++)
            {
                if (m_seqno[i] != 0)
                {
                    //
                    // Our sequence # array is always the seq # for the next
                    // packet, not the current packets, so we subtract one for
                    // that.  But when we encode the sizeinfo data, we will add
                    // three more packets, so this becomes seqno + 3 - 1 or + 2.
                    //
                    data[offset++] = (byte)(m_seqno[i] + 2);
                }
            }

            // Add the language information
            for (int i = 0; i < m_languages.GetLength(0); i++)
                data[offset + 7 - i] = (byte)m_languages[i];

            CDTextField field = new CDTextField(0, CDTextField.TextFieldType.SizeInfo);
            field.Value.AddData(0, data);
            EncodeField(ref blocks, 0, field, true);
        }

        /// <summary>
        /// Find a field given the list, the language, and the type.
        /// </summary>
        /// <param name="list">the list of fields to search</param>
        /// <param name="langidx">the language index we want</param>
        /// <param name="type">the type we want</param>
        /// <returns>the field if it exists, or null</returns>
        private CDTextField FindField(IList<CDTextField> list, byte langidx, CDTextField.TextFieldType type)
        {
            foreach (CDTextField field in list)
            {
                if (field.LanguageIndex == langidx && field.Type == type)
                    return field;
            }
            return null;
        }

        /// <summary>
        /// For a given language, process all of the tags for that langauge.  The tags are 
        /// processes in a specific order based on their type.
        /// </summary>
        /// <param name="blocks">the list of blocks that result from packing</param>
        /// <param name="langidx">the language index to process</param>
        private void PackOneLanguage(ref IList<CDTextBlock> blocks, byte langidx)
        {
            PackOneField(ref blocks, langidx, CDTextField.TextFieldType.Title);
            PackOneField(ref blocks, langidx, CDTextField.TextFieldType.Performer);
            PackOneField(ref blocks, langidx, CDTextField.TextFieldType.Songwriter);
            PackOneField(ref blocks, langidx, CDTextField.TextFieldType.Composer);
            PackOneField(ref blocks, langidx, CDTextField.TextFieldType.Arranger);
            PackOneField(ref blocks, langidx, CDTextField.TextFieldType.Message);
            PackOneField(ref blocks, langidx, CDTextField.TextFieldType.DiskId);
            PackOneField(ref blocks, langidx, CDTextField.TextFieldType.Genre);
            PackOneField(ref blocks, langidx, CDTextField.TextFieldType.Isrc);
        }

        private void EncodeField(ref IList<CDTextBlock> blocks, int trackno, CDTextField field, bool incrtrack)
        {
            byte[] data = field.Value.Data;
            int len = data.GetLength(0);
            int off = 0;

            while (off < len)
            {
                if (m_last_field == null || m_last_field.Type != field.Type || m_last_field.LanguageIndex != field.LanguageIndex)
                {
                    int blkoff = (off < 15) ? off : 15;
                    if (field.Type == CDTextField.TextFieldType.SizeInfo)
                        blkoff = 0;

                    m_last_field = new CDTextBlock(field.Type, field.LanguageIndex, (byte)trackno, m_seqno[field.LanguageIndex], blkoff);
                    m_packcnt[(byte)(field.Type) - 0x80]++;
                    m_seqno[field.LanguageIndex]++;
                    if (incrtrack)
                        trackno++;
                    blocks.Add(m_last_field);
                }

                //
                // We can back characters into the last field used
                //
                while (off < len && m_last_field.BytesRemaining > 0)
                    m_last_field.AddDataByte(data[off++]);

                if (m_last_field.BytesRemaining == 0)
                    m_last_field = null;
            }
        }

        private void PackOneField(ref IList<CDTextBlock> blocks, byte langidx, CDTextField.TextFieldType type)
        {
            m_last_field = null;

            CDTextField diskfield = FindField(m_disk, langidx, type);
            if (diskfield != null)
                EncodeField(ref blocks, 0, diskfield, false);

            if (CDTextField.AppliesToTracks(type))
            {
                int count = 0;
                foreach (IList<CDTextField> list in m_tracks)
                {
                    if (FindField(list, langidx, type) != null)
                        count++;
                }

                if (count > 0)
                {
                    //
                    // A special case for ISRC fields.  If any ISRC fields exist for tracks, we create one
                    // for the disk automatically as well
                    //
                    if (type == CDTextField.TextFieldType.Isrc && diskfield == null)
                    {
                        diskfield = new CDTextField(langidx, type, new CDTextDataString());
                        EncodeField(ref blocks, 0, diskfield, false);
                    }

                    int trackno = 1 ;
                    foreach (IList<CDTextField> list in m_tracks)
                    {
                        CDTextField trackfield = FindField(list, langidx, type);
                        if (trackfield == null)
                            trackfield = diskfield;

                        if (trackfield != null)
                            EncodeField(ref blocks, trackno, trackfield, false);

                        trackno++;
                    }
                }
            }
        }

        private CDTextField FindOrCreateField(CDTextField.TextFieldType packtype, byte trackno, byte langidx)
        {
            CDTextField f = null;
            if (trackno == 0)
            {
                foreach (CDTextField field in m_disk)
                {
                    if (field.Type == packtype && field.LanguageIndex == langidx)
                    {
                        f = field;
                        break;
                    }
                }

                if (f == null)
                {
                    f = new CDTextField(langidx, packtype);
                    m_disk.Add(f);
                }
            }
            else
            {
                IList<CDTextField> list ;

                if (trackno >= m_tracks.GetLength(0))
                    Array.Resize(ref m_tracks, trackno);

                list = m_tracks[trackno - 1];
                if (list == null)
                    list = m_tracks[trackno - 1] = new List<CDTextField>();

                foreach (CDTextField field in list)
                {
                    if (field.Type == packtype && field.LanguageIndex == langidx)
                    {
                        f = field;
                        break;
                    }
                }

                if (f == null)
                {
                    f = new CDTextField(langidx, packtype);
                    list.Add(f);
                }
            }

            return f;
        }

        private void DecodeCDText(byte[] data, int start, int length)
        {
            byte[] block = new byte[12];
            int offset = start;
            CDTextField field;

            while (offset + 18 < start + length)
            {
                CDTextField.TextFieldType packtype = (CDTextField.TextFieldType)data[offset++];
                byte trackno = data[offset++];
                byte seqno = data[offset++];
                byte blockchar = data[offset++];
                byte pos = (byte)(blockchar & 0x0f);
                byte langidx =(byte) ((blockchar >> 4) & 0x07);

                if (packtype == CDTextField.TextFieldType.SizeInfo)
                    trackno = 0;

                field = FindOrCreateField(packtype, trackno, langidx);
                trackno++;
                if (field.Value is CDTextDataString)
                {
                    CDTextDataString datval = (CDTextDataString)field.Value;
                    int n = 0;
                    string str = string.Empty;
                    while (n < 12)
                    {

                        // Assume ASCII encoding
                        if (data[offset + n] == 0)
                        {
                            datval.AddString(str);
                            str = string.Empty;
                            if (n != 11 && data[offset + n + 1] != 0)
                            {
                                field = FindOrCreateField(packtype, trackno, langidx);
                                datval = (CDTextDataString)field.Value;
                            }
                        }
                        else
                        {
                            str += (char)data[offset + n];
                        }

                        n++ ;
                        if (n == 12 && str.Length > 0)
                            datval.AddString(str);
                    }
                }
                else
                {
                    Array.Copy(data, offset, block, 0, 12);
                    field.Value.AddData(pos, block);
                }
                offset += 14;
            }

            //
            // Now, look for size info records and process them
            //
            field = FindField(m_disk, 0, CDTextField.TextFieldType.SizeInfo);
            if (field != null)
            {
                CDTextDataBinary bdata = (CDTextDataBinary)field.Value;

                for (byte i = 0; i < MaximumLanguageBlocks; i++)
                {
                    if (bdata[i + 28] != 0)
                        AddLanguage((LanguageType)bdata[i + 28]);
                }

                RemoveCDText(0, 0, CDTextField.TextFieldType.SizeInfo);
            }
        }
        #endregion
    }
}
