using System;
using System.Collections.Generic;
using System.Text;

namespace Bwg.Burner
{
    /// <summary>
    /// This class represents a single field of CD text attached to either a disc
    /// or a track
    /// </summary>
    [Serializable()]
    public class CDTextField
    {
        #region public types

        /// <summary>
        /// The type of this CD Text field
        /// </summary>
        public enum TextFieldType : byte
        {
            /// <summary>
            /// Self evident
            /// </summary>
            Title = 0x80,

            /// <summary>
            /// Self evident
            /// </summary>
            Performer = 0x81,

            /// <summary>
            /// Self evident
            /// </summary>
            Songwriter = 0x82,

            /// <summary>
            /// Self evident
            /// </summary>
            Composer = 0x83,

            /// <summary>
            /// Self evident
            /// </summary>
            Arranger = 0x84,

            /// <summary>
            /// Self evident
            /// </summary>
            Message = 0x85,

            /// <summary>
            /// Self evident
            /// </summary>
            DiskId = 0x86,

            /// <summary>
            /// Self evident
            /// </summary>
            Genre = 0x87,

            /// <summary>
            /// ???
            /// </summary>
            TocInfo1 = 0x88,
            
            /// <summary>
            /// ???
            /// </summary>
            TocInfo2 = 0x89,

            /// <summary>
            /// Reserved, not used
            /// </summary>
            Reserved1 = 0x8a,

            /// <summary>
            /// Reserved, not used
            /// </summary>
            Reserved2 = 0x8b,

            /// <summary>
            /// Reserved, not used
            /// </summary>
            Reserved3 = 0x8c,

            /// <summary>
            /// Reserved, not used
            /// </summary>
            Reserved4 = 0x8d,

            /// <summary>
            /// Self evident
            /// </summary>
            Isrc = 0x8e,

            /// <summary>
            /// Self evident
            /// </summary>
            SizeInfo = 0x8f
        } ;


        #endregion

        #region private member variables
        /// <summary>
        /// The type of this CD text field
        /// </summary>
        TextFieldType m_type;

        /// <summary>
        /// The data associated with this CD text field
        /// </summary>
        CDTextData m_data;

        /// <summary>
        /// The language associated with this field item
        /// </summary>
        byte m_lang_index;

        #endregion

        #region constructor
        /// <summary>
        /// Create a new CD text field object to attach to a disc or a track
        /// </summary>
        /// <param name="blocknr">the language index for this type</param>
        /// <param name="type">the type of the field</param>
        /// <param name="value">the value of the field</param>
        public CDTextField(byte blocknr, TextFieldType type, CDTextDataString value)
        {
            m_type = type;
            m_data = value;
            m_lang_index = blocknr;
        }

        /// <summary>
        /// Createa a new text field with a default data value
        /// </summary>
        /// <param name="blocknr">the languange index for this cd text data field</param>
        /// <param name="type">the type of this cd text data field</param>
        public CDTextField(byte blocknr, TextFieldType type)
        {
            m_type = type;
            m_lang_index = blocknr;

            switch (m_type)
            {
                case TextFieldType.Arranger:
                case TextFieldType.Composer:
                case TextFieldType.Message:
                case TextFieldType.Performer:
                case TextFieldType.Songwriter:
                case TextFieldType.Title:
                    m_data = new CDTextDataString();
                    break;

                case TextFieldType.DiskId:
                case TextFieldType.Isrc:
                case TextFieldType.SizeInfo:
                case TextFieldType.TocInfo1:
                case TextFieldType.TocInfo2:
                    m_data = new CDTextDataBinary();
                    break ;

                case TextFieldType.Genre:
                    m_data = new CDTextDataGenre();
                    break ;
            }
        }
        #endregion

        #region public properties
        /// <summary>
        /// This property returns the type of the cd text field entry
        /// </summary>
        public TextFieldType Type
        {
            get
            {
                return m_type;
            }
        }

        /// <summary>
        /// This property returns the value of the cd text field entry
        /// </summary>
        public CDTextData Value
        {
            get
            {
                return m_data;
            }
        }

        /// <summary>
        /// This property returns the language of the ce text field entry
        /// </summary>
        public byte LanguageIndex
        {
            get
            {
                return m_lang_index;
            }
        }
        #endregion

        #region public methods

        /// <summary>
        /// Returns a string representing this text field
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return m_type.ToString() + "/" + m_lang_index.ToString() + " = " + m_data.ToString();
        }

        /// <summary>
        /// This static method returns true if the CD text field type can apply to
        /// tracks as well as to the entire disc
        /// </summary>
        /// <param name="type">the type in question</param>
        /// <returns>true if the type applies to tracks, false otherwise</returns>
        static public bool AppliesToTracks(TextFieldType type)
        {
            bool ret = false;

            if (type == TextFieldType.Title || type == TextFieldType.Performer ||
                type == TextFieldType.Songwriter || type == TextFieldType.Composer ||
                type == TextFieldType.Arranger || type == TextFieldType.Message ||
                type == TextFieldType.Isrc)
            {
                ret = true;
            }

            return ret;
        }
        #endregion
    }
}
