//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Bwg.Scsi;
using Bwg.Logging;

namespace Bwg.Burner
{
    class DVDMinusRDiskOperations : DVDBaseDiskOperations
    {
        /// This class supports burndisk with the following features:
        /// 
        /// - Session At Once (also know as Disk At Once)
        /// - Track At once
        /// - Burn leaving disk open for additional sessions
        /// - Burn finalizing disk for maximum compatibility
        /// - Set burn speed
        ///
        public DVDMinusRDiskOperations(Drive d)
            : base(d)
        {
        }

        private long DefaultCapacity
        {
            get
            {
                long ret = 0;

                switch (m_drive.CurrentProfile)
                {
                    case Drive.SupportedProfiles.DVD_R_DualLayer_LayerJumpRecording:
                    case Drive.SupportedProfiles.DVD_R_DualLayer_SequentialRecording:
                        ret = Drive.D47GigDVDMinusRCapacity;
                        break;

                    case Drive.SupportedProfiles.DVD_R_SequentialRecording:
                        ret = Drive.D47GigDVDMinusRCapacity;
                        break;

                    default:
                        Debug.Assert(false);
                        break;
                }

                return ret;
            }
        }

        /// <summary>
        /// The total capacity of this disk in sectors
        /// </summary>
        public override long CapacityInSectors
        {
            get
            {
                IList<CapacityDescriptor> caplist;
                if (MyDevice.ReadFormatCapacities(true, out caplist) != Device.CommandStatus.Success)
                    return DefaultCapacity;

                if (caplist[0].DescType == CapacityDescriptor.DescriptorType.Unformatted)
                    return caplist[0].NumberOfBlocks;

                for (int i = 1; i < caplist.Count; i++)
                {
                    if (caplist[i].FormatType == 0x26)
                        return caplist[i].NumberOfBlocks;
                }

                return DefaultCapacity;
            }
        }

        /// <summary>
        /// This property returns the writable capacity in sectors for the DVD-R media.  This
        /// is the total capacity regardless of what is available.  Since all of the media is 
        /// writable initially the return value is the same as the overall capacity
        /// </summary>
        public override long WritableCapacityInSectors
        {
            get
            {
                return CapacityInSectors;
            }
        }

        /// <summary>
        /// The remaining empty capacity for this disk in sectors
        /// </summary>
        public override long RemainingCapacityInWritableSectors
        {
            get
            {
                DiscInformation info;

                if (MyDevice.ReadDiscInformation(out info) != Device.CommandStatus.Success)
                    return 0;

                if (info.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc ||
                    info.DiscStatus == DiscInformation.DiscStatusType.Other)
                    return 0;

                // To find the remaining capacity, find the next writable address of the
                // invisible track
                TrackInformation tinfo;
                if (MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, Device.InvisibleTrack, out tinfo) != Device.CommandStatus.Success)
                    return 0;

                return WritableCapacityInSectors - tinfo.NextWritableAddress;
            }
        }

        /// <summary>
        /// This property returns true if burn is supported on this media
        /// </summary>
        public override bool IsBurnSupported
        {
            get
            {
                DiscInformation info;

                if (MyDevice.ReadDiscInformation(out info) == Device.CommandStatus.Success &&
                    (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc ||
                    info.DiscStatus == DiscInformation.DiscStatusType.IncompleteDisc))
                    return true;

                return false;
            }
        }

        public override bool IsEraseSupported
        {
            get
            {
                return false;
            }
        }

        public override bool IsMultiSessionSupported
        {
            get
            {
                return true;
            }
        }

        public override bool IsTrackAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDTrackAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        public override bool IsSessionAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDSessionAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        public override DiskOperationError Erase(Drive.EraseType et)
        {
            return NotSupported("erase");
        }

        /// <summary>
        /// This method burns a disk onto CDRW media.  The disk image is given by the
        /// object d and the burn type indicates a disk at once versus a track at once
        /// burn.
        /// </summary>
        /// <param name="bt">type of the burn (track vs session)</param>
        /// <param name="d">the image to burn to a disk</param>
        /// <param name="multi_session">if true, leave the disk open for multiple sessions</param>
        /// <param name="speed">the speed of the burn operation</param>
        /// <param name="bytes">output, the number of bytes burned</param>
        /// <returns>error object is an error occurs, otherwise null</returns>
        public override DiskOperationError BurnDisk(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            Device.CommandStatus st;
            DiskOperationError status ;

            bytes = 0;

            if (d.ContainsAudio)
                return new BurnerError("cannot write audio tracks to a DVD-R/DVD-RW media, DVD-R/DVD-RW only supports data tracks");

            if (d.TrackCount != 1)
                return new BurnerError("cannot write multi-track content to a DVD-R/DVD-RW media");

            DiscInformation info;
            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            if (info.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc)
                return new BurnerError("the media has been finalized without the option of added new sessions");

            if (bt == Drive.BurnType.DontCare)
            {
                if (m_drive.CurrentProfile == Drive.SupportedProfiles.DVD_R_DualLayer_LayerJumpRecording)
                    status = BurnDiskLayerJump(d, multi_session, speed, out bytes);
                else if (m_drive.IsSessionAtOnceSupported)
                    status = BurnDiskSession(d, multi_session, speed, out bytes);
                else
                    status = BurnDiskTrack(d, multi_session, speed, out bytes);
            }
            else if (bt == Drive.BurnType.Session)
                status = BurnDiskSession(d, multi_session, speed, out bytes);
            else
                status = BurnDiskTrack(d, multi_session, speed, out bytes);

            return status;
        }

        private DiskOperationError BurnDiskLayerJump(DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            bytes = 0 ;
            return new BurnerError("Layer Jump DL burning not yet supported.");
        }

        private DiskOperationError BurnDiskTrack(DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            Device.CommandStatus st;
            DiskOperationError status;
            ModeTable table;

            bytes = 0;

            bool proof;
            status = m_drive.GetBurnProofSetting(out proof);
            if (status != null)
                return status;

            st = MyDevice.ModeSense(Device.PageControl.Current, 5, out table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            //
            // Setup the mode page for write track at once
            //
            Debug.Assert(table.Pages[0] is WriteParameterModePage);
            WriteParameterModePage pg = (WriteParameterModePage)table.Pages[0];

            pg.ParametersSavable = false;
            pg.BurnProof = proof;
            pg.TrackMode = WriteParameterModePage.TrackModeType.DataIncremental;
            pg.SessionFormat = WriteParameterModePage.SessionFormatType.CDDA_CDROM;

            if (multi_session)
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NextSessionAllowed;
            else
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NoNextSession;

            pg.WriteType = WriteParameterModePage.WriteTypeType.TrackAtOnce;

            st = MyDevice.ModeSelect(false, table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            MyDevice.LogSenseState = false;
            st = MyDevice.SendOpcInformation(false, null);
            MyDevice.LogSenseState = true;

            status = m_drive.SetSpeed(speed);
            if (status != null)
                return status;

            int lba;
            int trackno;
            status = NextWritableAddress(out lba, out trackno);
            if (status != null)
                return status;

            d.StartSector = lba;
            d.TargetMedia = DiskBurnImage.DiskTargetType.DVD;

            DataGenerator gen = new DataGenerator(MyDevice.GetLogger(), 64, MyDevice.MaximumTransferLength, 0, 0, 0);
            gen.AddSource("data", d[0], 0, d[0].DataToSend);

            status = m_drive.SendData(gen, d);
            bytes = (ulong)(d.LengthSectors * 2048);
            if (status != null)
            {
                m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
                return status;
            }

            status = m_drive.SynchronizeCache();
            if (status != null)
                return status;

            m_drive.WaitForDeviceReady(false);

            // status = m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseTrack, (ushort)trackno);
            // if (status != null)
                // return status;

            status = m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
            if (status != null)
                return status;

            return null;
        }

        private DiskOperationError BurnDiskSession(DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            Device.CommandStatus st;
            DiskOperationError status;
            ModeTable table;

            bytes = 0;

            bool proof;
            status = m_drive.GetBurnProofSetting(out proof);
            if (status != null)
                return status;

            st = MyDevice.ModeSense(Device.PageControl.Current, 5, out table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            //
            // Setup the mode page for write session at once
            //
            Debug.Assert(table.Pages[0] is WriteParameterModePage);
            WriteParameterModePage pg = (WriteParameterModePage)table.Pages[0];

            pg.ParametersSavable = false;
            pg.BurnProof = proof;
            if (multi_session)
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NextSessionAllowed;
            else
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NoNextSession;

            pg.WriteType = WriteParameterModePage.WriteTypeType.SessionAtOnce;
            pg.DataBlockType = WriteParameterModePage.DataBlockTypeType.DataMode1;
            pg.TrackMode = WriteParameterModePage.TrackModeType.DataIncremental;
            pg.SessionFormat = WriteParameterModePage.SessionFormatType.CDDA_CDROM;

            st = MyDevice.ModeSelect(false, table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            MyDevice.LogSenseState = false;
            st = MyDevice.SendOpcInformation(false, null);
            MyDevice.LogSenseState = true;

            status = m_drive.SetSpeed(speed);
            if (status != null)
                return status;

            int lba;
            int trackno;
            status = NextWritableAddress(out lba, out trackno);
            if (status != null)
                return status;

            d.StartSector = lba;
            d.TargetMedia = DiskBurnImage.DiskTargetType.DVD;

            st = MyDevice.ReserveTrack((uint)d.LengthSectors);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            DataGenerator gen = new DataGenerator(MyDevice.GetLogger(), 64, MyDevice.MaximumTransferLength, 0, 0, 0);
            gen.AddSource("data", d[0], 0, d[0].DataToSend);

            status = m_drive.SendData(gen, d);
            bytes = (ulong)(d.LengthSectors * 2048);
            if (status != null)
            {
                m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
                return status;
            }

            status = m_drive.SynchronizeCache();
            if (status != null)
                return status;

            m_drive.WaitForDeviceReady(false);

            status = m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
            if (status != null)
                return status;

            return null;
        }

        private DiskOperationError NextWritableAddress(out int lba, out int trackno)
        {
            Device.CommandStatus st;
            DiscInformation dinfo;
            lba = 0;
            trackno = 1;

            st = MyDevice.ReadDiscInformation(out dinfo);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            if (dinfo.DiscStatus != DiscInformation.DiscStatusType.EmptyDisc)
            {
                if (dinfo.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc)
                    return new BurnerError("trying to append a session to a finalized disk");

                // Now, find the next writable address
                //
                // Find the start address for the new track
                //
                TrackInformation info;
                st = MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, Device.InvisibleTrack, out info);
                if (st != Device.CommandStatus.Success)
                    return CreateErrorObject(st);

                lba = info.NextWritableAddress;
                trackno = info.TrackNumber;
            }

            return null;
        }
    }
}
