//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Bwg.Logging;
using Bwg.Scsi;

namespace Bwg.Burner
{
    //
    // TODO: Multi-Sessions support
    //
    class DVDMinusRWDiskOperations : DVDBaseDiskOperations
    {
        public DVDMinusRWDiskOperations(Drive d) : base(d)
        {
        }

        public override long CapacityInSectors
        {
            get
            {
                IList<CapacityDescriptor> caplist;
                if (MyDevice.ReadFormatCapacities(true, out caplist) != Device.CommandStatus.Success)
                    return Drive.D47GigDVDMinusRCapacity;

                if (caplist[0].DescType == CapacityDescriptor.DescriptorType.Unformatted)
                    return caplist[0].NumberOfBlocks;

                for (int i = 1; i < caplist.Count; i++)
                {
                    if (caplist[i].DescType == CapacityDescriptor.DescriptorType.Reserved && 
                        caplist[i].FormatType == 0x00)
                        return caplist[i].NumberOfBlocks;
                }

                return Drive.D47GigDVDMinusRCapacity;
            }
        }

        public override long WritableCapacityInSectors
        {
            get
            {
                return CapacityInSectors;
            }
        }

        public override long RemainingCapacityInWritableSectors
        {
            get
            {
                DiscInformation info;

                //
                // Since we don't support multi-session burning for DVD-RW yet, a disk is
                // either erased and the capacity is the entire disk, or the disk is used
                // and the capacity is zero.
                //
                // TODO: Update this when we figure out how to support multi-session DVD-RW disks
                //

                if (MyDevice.ReadDiscInformation(out info) != Device.CommandStatus.Success)
                    return 0;

                if (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
                    return WritableCapacityInSectors;

                return 0;
            }
        }

        public override bool IsBurnSupported
        {
            get
            {
                return true;
            }
        }

        public override bool IsMultiSessionSupported
        {
            get
            {
                return true;
            }
        }

        public override bool IsEraseSupported
        {
            get
            {
                return true;
            }
        }

        public override bool IsTrackAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDTrackAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        public override bool IsSessionAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDSessionAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        /// <summary>
        /// This method erases a DVD-RW disk that is currently in the drive.
        /// </summary>
        /// <param name="et">the type of erase to do (full or quick)</param>
        /// <returns>status indicating the success of the erase operation</returns>
        public override DiskOperationError Erase(Drive.EraseType et)
        {
            DiskOperationError status;

            if (MyLogger != null)
            {
                string args = et.ToString();
                MyLogger.LogMessage(new UserMessage(UserMessage.Category.Debug, 0, "calling Bwg.Burner.Drive.Erase(" + args + ")"));
            }

            Device.CommandStatus st;
            status = m_drive.MakeDriveReady();
            if (status != null)
                return status;

            DiscInformation info;
            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            // This should really never happen, but just in case
            if (!info.Erasable)
                return new BurnerError("the media in the drive is not erasable");

            //
            // It is OK for OPC to fail on blank media
            //
            MyDevice.LogSenseState = false;
            st = MyDevice.SendOpcInformation(false, null);
            MyDevice.LogSenseState = true;

            int cnt = 0;
            while (true)
            {
                if (et == Drive.EraseType.Full)
                    st = MyDevice.Blank(true, Device.BlankType.FullDisc, 0);
                else
                    st = MyDevice.Blank(true, Device.BlankType.MinimalDisc, 0);

                if (st == Device.CommandStatus.Success)
                    break;

                if (st == Device.CommandStatus.DeviceFailed && MyDevice.GetSenseAsc() == 4 && MyDevice.GetSenseAscq() == 8 && cnt++ < 8)
                {
                    m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
                    continue;
                }

                return CreateErrorObject(st);
            }

            st = m_drive.WaitForDeviceReady(true);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            st = MyDevice.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            return null;
        }

        public override DiskOperationError BurnDisk(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            return BurnDiskRestrictedOverwrite(bt, d, multi_session, speed, out bytes);
        }

        private DiskOperationError BurnDiskSequential(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            Device.CommandStatus st;
            DiskOperationError status;
            ModeTable table;
            bool streaming = false ;

            bytes = 0;

            bool proof;
            status = m_drive.GetBurnProofSetting(out proof);
            if (status != null)
                return status;

            st = MyDevice.ModeSense(Device.PageControl.Current, 5, out table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            if (m_drive.GetFeatureState(Feature.FeatureType.IncrementalStreamWritable) == Drive.FeatureState.PresentAndCurrent)
            {
                Feature f ;
                m_drive.GetFeature(Feature.FeatureType.IncrementalStreamWritable, out f);
                streaming = true;
            }

            //
            // Setup the mode page for write track at once
            //
            Debug.Assert(table.Pages[0] is WriteParameterModePage);
            WriteParameterModePage pg = (WriteParameterModePage)table.Pages[0];

            pg.ParametersSavable = false;
            pg.BurnProof = proof;
            pg.TrackMode = WriteParameterModePage.TrackModeType.DataIncremental;
            pg.SessionFormat = WriteParameterModePage.SessionFormatType.CDDA_CDROM;
            pg.DataBlockType = WriteParameterModePage.DataBlockTypeType.DataMode1;

            if (streaming)
            {
                pg.WriteType = WriteParameterModePage.WriteTypeType.PacketIncremental;
                pg.PacketSize = 16;
                pg.FixedPacket = true;
            }
            else
            {
                pg.WriteType = WriteParameterModePage.WriteTypeType.SessionAtOnce;
                pg.PacketSize = 16;
                pg.FixedPacket = false;
            }

            if (multi_session)
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NextSessionAllowed;
            else
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NoNextSession;

            st = MyDevice.ModeSelect(false, table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            MyDevice.LogSenseState = false;
            st = MyDevice.SendOpcInformation(false, null);
            MyDevice.LogSenseState = true;

            status = m_drive.SetSpeed(speed);
            if (status != null)
                return status;

            int lba;
            int trackno;
            status = NextWritableAddress(out lba, out trackno);
            if (status != null)
                return status;

            d.StartSector = lba;
            d.TargetMedia = DiskBurnImage.DiskTargetType.DVD;

            DataGenerator gen = new DataGenerator(MyDevice.GetLogger(), 64, MyDevice.MaximumTransferLength, 16, 0, 2048);
            gen.AddSource("data", d[0], 0, d[0].DataToSend);

            status = m_drive.SendData(gen, d);
            bytes = (ulong)(d.LengthSectors * 2048);
            if (status != null)
            {
                m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
                return status;
            }

            status = m_drive.SynchronizeCache();
            if (status != null)
                return status;

            m_drive.WaitForDeviceReady(false);

            status = m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseTrack, (ushort)trackno);
            if (status != null)
                return status;

            status = m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
            if (status != null)
                return status;

            return null;
        }
        
        /// <summary>
        /// This method burns a disk onto CDRW media.  The disk image is given by the
        /// object d and the burn type indicates a disk at once versus a track at once
        /// burn.
        /// </summary>
        /// <param name="bt">type of the burn (track vs session)</param>
        /// <param name="d">the image to burn to a disk</param>
        /// <param name="multi_session">if true, leave the disk open for multiple sessions</param>
        /// <param name="speed">the speed of the burn operation</param>
        /// <param name="bytes">output, the number of bytes burned</param>
        /// <returns>error object is an error occurs, otherwise null</returns>
        public DiskOperationError BurnDiskRestrictedOverwrite(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            DiskOperationError status;
            Device.CommandStatus st;
            ModeTable table;
            DiscInformation info;
            int total_length;

            bytes = 0;

            if (d.ContainsAudio)
                return new BurnerError("cannot write audio tracks to a DVD-R/DVD-RW media, DVD-R/DVD-RW only supports data tracks");

            if (d.TrackCount != 1)
                return new BurnerError("cannot write multi-track content to a DVD-R/DVD-RW media");

            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            bool proof;
            status = m_drive.GetBurnProofSetting(out proof);
            if (status != null)
                return status;

            st = MyDevice.ModeSense(Device.PageControl.Current, 5, out table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            //
            // Setup the mode page for write track at once
            //
            Debug.Assert(table.Pages[0] is WriteParameterModePage);
            WriteParameterModePage pg = (WriteParameterModePage)table.Pages[0];

            pg.ParametersSavable = false;
            pg.BurnProof = proof;
            pg.TrackMode = WriteParameterModePage.TrackModeType.DataIncremental;
            pg.SessionFormat = WriteParameterModePage.SessionFormatType.CDDA_CDROM;
            pg.DataBlockType = WriteParameterModePage.DataBlockTypeType.DataMode1;
            pg.WriteType = WriteParameterModePage.WriteTypeType.SessionAtOnce;
            pg.PacketSize = 16;

            if (multi_session)
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NextSessionAllowed;
            else
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NoNextSession;

            st = MyDevice.ModeSelect(false, table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            MyDevice.LogSenseState = false;
            st = MyDevice.SendOpcInformation(false, null);
            MyDevice.LogSenseState = true;

            status = m_drive.SetSpeed(speed);
            if (status != null)
                return status;

            total_length = (int)d.LengthSectors;
            if ((total_length % 16) != 0)
                total_length = ((total_length / 16) + 1) * 16;

            if (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
            {
                //
                // This is the first session on an empty disc, QuickFormat() the space needed
                // for the session
                //
                DateTime start = DateTime.Now;
                status = QuickFormat((uint)total_length);
                if (status != null)
                    return status;
                TimeSpan duration = DateTime.Now - start;
                string str = "quick formatted " + total_length.ToString() + " sectors in " + duration.TotalSeconds.ToString() + " seconds";
                MyLogger.LogMessage(new UserMessage(UserMessage.Category.Info, 0, str));

                d.StartSector = 0;
                d.TargetMedia = DiskBurnImage.DiskTargetType.DVD;
            }
            else
            {
                IList<CapacityDescriptor> list ;
                MyDevice.ReadFormatCapacities(true, out list);

                DateTime start = DateTime.Now;
                status = QuickAddSession((uint)total_length);
                if (status != null)
                    return status;
                TimeSpan duration = DateTime.Now - start;
                string str = "quick formatted " + total_length.ToString() + " sectors in " + duration.TotalSeconds.ToString() + " seconds";
                MyLogger.LogMessage(new UserMessage(UserMessage.Category.Info, 0, str));

                TrackInformation tinfo;
                st = MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, info.LastTrackInLastSession, out tinfo);
                if (st != Device.CommandStatus.Success)
                    return CreateErrorObject(st);

                d.StartSector = tinfo.NextWritableAddress;
                d.TargetMedia = DiskBurnImage.DiskTargetType.DVD;
            }

            pg.ParametersSavable = false;
            pg.BurnProof = proof;
            pg.TrackMode = WriteParameterModePage.TrackModeType.DataIncremental;
            pg.SessionFormat = WriteParameterModePage.SessionFormatType.CDDA_CDROM;

            if (multi_session)
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NextSessionAllowed;
            else
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NoNextSession;

            pg.WriteType = WriteParameterModePage.WriteTypeType.TrackAtOnce;

            st = MyDevice.ModeSelect(false, table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            DataGenerator gen = new DataGenerator(MyDevice.GetLogger(), 64, BufferSize, 16, 0, 2048);
            gen.AddSource("data", d[0], 0, d[0].DataToSend);

            status = m_drive.SendData(gen, d);
            bytes = (ulong)(total_length * 2048);
            if (status != null)
            {
                MyLogger.LogMessage(new UserMessage(UserMessage.Category.Info, 0, "error occurred during burn, closing session"));
                m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
                return status;
            }

            status = m_drive.SynchronizeCache();
            if (status != null)
            {
                m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
                return status;
            }

            m_drive.ReadBufferLevels();

            return m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
        }

        private DiskOperationError NextWritableAddress(out int lba, out int trackno)
        {
            Device.CommandStatus st;
            DiscInformation dinfo;
            lba = 0;
            trackno = 1;

            st = MyDevice.ReadDiscInformation(out dinfo);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            if (dinfo.DiscStatus != DiscInformation.DiscStatusType.EmptyDisc)
            {
                if (dinfo.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc)
                    return new BurnerError("trying to append a session to a finalized disk");

                // Now, find the next writable address
                //
                // Find the start address for the new track
                //
                TrackInformation info;
                st = MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, dinfo.LastTrackInLastSession, out info);
                if (st != Device.CommandStatus.Success)
                    return CreateErrorObject(st);

                lba = info.NextWritableAddress;
                trackno = info.TrackNumber ;
            }

            return null;
        }

        private int BufferSize
        {
            get
            {
                int sectors = MyDevice.MaximumTransferLength / 2048;

                sectors = sectors - (sectors % 16);

                return sectors * 2048;
            }
        }

        private DiskOperationError QuickFormat(uint size)
        {
            Device.CommandStatus st;
            string str;

            // We need to format the unit, up to the sector needed
            if (MyDevice.GetLogger() != null)
            {
                str = "doing DVD-RW quick format, " + size.ToString() + " sectors";
                MyLogger.LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));
            }

            m_drive.CurrentOperation = new ProgressText("Performing Quick Format", "Formatting");
            // m_drive.SetTimedOperation((int)(.000751 * size + 24));

            FormatParameterList plist = new FormatParameterList();
            plist.IMMED = 1;
            plist.FOV = 1;
            plist.DCRT = 1;
            plist.FormatType = FormatParameterList.UnitFormatType.DVD_RW_QuickFormat;
            plist.NumberOfBlocks = size;
            plist.Parameter = 16;
            st = MyDevice.FormatUnit(plist);

            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            m_drive.WaitForDeviceReady(true);

            return null;
        }

        private DiskOperationError QuickAddSession(uint size)
        {
            Device.CommandStatus st;
            string str;

            // We need to format the unit, up to the sector needed
            if (MyDevice.GetLogger() != null)
            {
                str = "doing DVD-RW quick add session, " + size.ToString() + " sectors";
                MyLogger.LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));
            }

            m_drive.CurrentOperation = new ProgressText("Performing Quick Add Session", "Formatting");
            // m_drive.SetTimedOperation((int)(.000751 * size + 24));

            FormatParameterList plist = new FormatParameterList();
            plist.IMMED = 0;
            plist.FOV = 0;
            plist.DCRT = 0;
            plist.FormatType = FormatParameterList.UnitFormatType.DVD_RW_QuickAddSession ;
            plist.NumberOfBlocks = size;
            plist.Parameter = 16;
            st = MyDevice.FormatUnit(plist);

            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            m_drive.WaitForDeviceReady(true);

            return null;
        }

        private DiskOperationError QuickGrow(uint size)
        {
            Device.CommandStatus st;

            FormatParameterList plist = new FormatParameterList();
            plist.IMMED = 1;
            plist.FOV = 1;
            plist.DCRT = 1;
            plist.FormatType = FormatParameterList.UnitFormatType.DVD_RW_QuickGrowLastSession;
            plist.NumberOfBlocks = size;
            plist.Parameter = 16;
            st = MyDevice.FormatUnit(plist);

            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            m_drive.WaitForDeviceReady(false);

            return null;
        }
    }
}
