//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Bwg.Scsi;
using Bwg.Burner;
using Bwg.Logging;

namespace Bwg.Burner
{
    /// <summary>
    /// Disk operations for a DVD+R disk
    /// </summary>
    public class DVDPlusRDiskOperations : DVDBaseDiskOperations
    {
        /// <summary>
        /// 
        /// </summary>
        /// <param name="d"></param>
        public DVDPlusRDiskOperations(Drive d) : base(d)
        {
        }

        /// <summary>
        /// The default capacity of the disk based on the type of mendia loaded
        /// </summary>
        private long DefaultCapacity
        {
            get
            {
                long ret = 0;

                switch(m_drive.CurrentProfile)
                {
                    case Drive.SupportedProfiles.DVD_PLUS_R:
                        ret = Drive.D47GigDVDPlusRCapacity ;
                        break ;

                    case Drive.SupportedProfiles.DVD_PLUS_R_DUAL_LAYER:
                        ret = Drive.D47GigDVDPlusRDLCapacity ;
                        break ;

                    default:
                        Debug.Assert(false) ;
                        break ;
                }
                return ret ;
            }
        }

        /// <summary>
        /// This property returns the capacity of this media in sectors
        /// </summary>
        public override long CapacityInSectors
        {
            get
            {
                IList<CapacityDescriptor> caplist;
                if (MyDevice.ReadFormatCapacities(true, out caplist) != Device.CommandStatus.Success)
                    return DefaultCapacity;

                if (caplist[0].DescType == CapacityDescriptor.DescriptorType.Unformatted)
                    return caplist[0].NumberOfBlocks;

                for (int i = 0; i < caplist.Count; i++)
                {
                    if (caplist[i].FormatType == 0x26 || caplist[i].FormatType == 0)
                        return caplist[i].NumberOfBlocks;
                }

                return DefaultCapacity;
            }
        }

        /// <summary>
        /// This property returns the writable capacity in sectors for the DVD+R media.  This
        /// is the total capacity regardless of what is available.  Since all of the media is 
        /// writable initially the return value is the same as the overall capacity
        /// </summary>
        public override long WritableCapacityInSectors
        {
            get
            {
                return CapacityInSectors;
            }
        }

        /// <summary>
        /// This is the capacity of the disk that is remaining.  If the disk has been finalized, the
        /// return value is zero, otherwise, additional sessions can be added and the return value is
        /// the space remaning.
        /// </summary>
        public override long RemainingCapacityInWritableSectors
        {
            get
            {
                DiscInformation info;

                if (MyDevice.ReadDiscInformation(out info) != Device.CommandStatus.Success)
                    return 0;

                if (info.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc ||
                    info.DiscStatus == DiscInformation.DiscStatusType.Other)
                    return 0;

                // To find the remaining capacity, find the next writable address of the
                // invisible track
                TrackInformation tinfo;
                if (MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, Device.InvisibleTrack, out tinfo) != Device.CommandStatus.Success)
                    return 0;

                return WritableCapacityInSectors - tinfo.NextWritableAddress;
            }
        }

        /// <summary>
        /// Returns true if data can be burned onto the media
        /// </summary>
        public override bool IsBurnSupported
        {
            get
            {
                DiscInformation info;

                if (MyDevice.ReadDiscInformation(out info) == Device.CommandStatus.Success &&
                    (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc ||
                    info.DiscStatus == DiscInformation.DiscStatusType.IncompleteDisc))
                    return true;

                return false ;
            }
        }

        /// <summary>
        /// Returns false always, as DVD+R media is not erasable
        /// </summary>
        public override bool IsEraseSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Returns true if burn is supported.
        /// </summary>
        public override bool IsMultiSessionSupported
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Track at once is always supported as long as data can be burned to the disk.
        /// </summary>
        public override bool IsTrackAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDTrackAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        /// <summary>
        /// Sessions at once is supported for this media as long as the drive can be burned, and the feature exist in the
        /// current drive.
        /// </summary>
        public override bool IsSessionAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDSessionAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        /// <summary>
        /// This method burns the contents of the disk object to the DVD+R media that is in the drive
        /// </summary>
        /// <param name="bt">type of the burn (disk at once or session at once), does not apply here</param>
        /// <param name="d">the disk contents to burn to the media</param>
        /// <param name="multi_session">if true, leave media open for future sessions</param>
        /// <param name="speed">the speed of the burn</param>
        /// <param name="bytes">output, the number of bytes burned</param>
        /// <returns></returns>
        public override DiskOperationError BurnDisk(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            string str;
            Device.CommandStatus st ;
            DiskOperationError status;
            ModeTable table;

            bytes = 0;

            if (d.ContainsAudio)
                return new BurnerError("cannot write audio tracks to a DVD+R media, DVD+R only supports data tracks");

            if (m_drive.CurrentProfile == Drive.SupportedProfiles.DVD_PLUS_R_DUAL_LAYER && multi_session)
                return new BurnerError("multi-session is not support with DL media (yet)");

            bool proof;
            status = m_drive.GetBurnProofSetting(out proof);
            if (status != null)
                return status;

            st = MyDevice.ModeSense(Device.PageControl.Current, 5, out table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            //
            // Setup the mode page for write track at once
            //
            Debug.Assert(table.Pages[0] is WriteParameterModePage);
            WriteParameterModePage pg = (WriteParameterModePage)table.Pages[0];

            pg.ParametersSavable = false;
            pg.BurnProof = proof;
            pg.WriteType = WriteParameterModePage.WriteTypeType.TrackAtOnce;

            pg.TrackMode = WriteParameterModePage.TrackModeType.DataIncremental;
            pg.DataBlockType = WriteParameterModePage.DataBlockTypeType.DataMode1;

            if (multi_session)
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NextSessionAllowed;
            else
                pg.MultiSession = WriteParameterModePage.MultiSessionType.NoNextSession;

            pg.SessionFormat = WriteParameterModePage.SessionFormatType.CDDA_CDROM;

            st = MyDevice.ModeSelect(false, table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            DiscInformation info;
            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            if (info.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc)
                return new BurnerError("the media has been finalized, no data can be appended to the media");

            status = m_drive.SetSpeed(speed);
            if (status != null)
                return status;

            //
            // Set the target media, this sets the start sector for the disk as well as all of
            // the pre-track gap information
            //
            d.StartSector = 0;
            d.TargetMedia = DiskBurnImage.DiskTargetType.DVD;

            if (m_drive.CurrentProfile == Drive.SupportedProfiles.DVD_PLUS_R_DUAL_LAYER)
            {
                status = SplitDLDisk(d.DataToSend);
                if (status != null)
                    return status;
            }


            // Find out how many sessions already exist on the media
            uint offset = info.SessionCount;

            if (offset == 1 && info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
            {
                str = "writing data to empty DVD+R disk";
                MyDevice.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));
            }
            else if (offset > 1 && info.DiscStatus == DiscInformation.DiscStatusType.IncompleteDisc)
            {
                str = "appending data to non-empty DVD+R disk, " + offset.ToString() + " existing sessions";
                MyDevice.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));
            }
            else
                return new BurnerError("bad disk format, status " + info.DiscStatus.ToString() + ", offset " + offset.ToString());


            Device.CloseTrackSessionType closetype = Device.CloseTrackSessionType.FinalizeDiscCompatible;
            if (multi_session)
                closetype = Device.CloseTrackSessionType.CloseSession ;

            for (int i = 0; i < d.TrackCount; i++)
            {
                TrackBurnImage t = d[i];
                int eff_track = (int)(t.TrackNumber + offset);

                // Now write the tracks, starting at the track # just after the existing sessions
                TrackInformation tinfo;
                st = MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, (uint)eff_track, out tinfo);
                if (st != Device.CommandStatus.Success)
                {
                    st = MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, Device.InvisibleTrack, out tinfo);
                    if (st != Device.CommandStatus.Success)
                        return CreateErrorObject(st);
                }

                int lba = tinfo.NextWritableAddress;
                str = "Track (session) " + eff_track.ToString() + " next writable address " + lba.ToString();
                MyDevice.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));

                DataGenerator gen = new DataGenerator(MyDevice.GetLogger(), 64, MyDevice.MaximumTransferLength, 0, 0, 0);
                gen.AddSource("track_data", t, lba, t.DataToSend);
                status = m_drive.SendData(gen, t);
                bytes += (ulong)(t.TotalLengthSectors * t.SectorSize);
                if (status != null)
                {
                    m_drive.SynchronizeCache();
                    m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseTrack, (ushort)eff_track);
                    m_drive.CloseTrackSession(closetype, 0);

                    return status;
                }

                status = m_drive.SynchronizeCache();
                if (status != null)
                    return status;

                m_drive.WaitForDeviceReady(false);
                status = m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseTrack, (ushort)eff_track);
                if (status != null)
                    return status;

                if (i == d.TrackCount - 1)
                {
                    //
                    // This is the last track, on this track we honor the multi-session 
                    // requeste passed into this method.
                    //
                    status = m_drive.CloseTrackSession(closetype, 0);
                    if (status != null)
                        return status;
                }
                else
                {
                    //
                    // Since each track is recorded as a session, all sessions but the last one must be
                    // closed with the possibility of a next session.
                    //
                    status = m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
                    if (status != null)
                        return status;
                }
            }

            return null;
        }

        /// <summary>
        /// This method erases the DVD +R media in the drive.  This is not supported for this media type
        /// and if called an error will be returned.
        /// </summary>
        /// <param name="et">the erase type</param>
        /// <returns></returns>
        public override DiskOperationError Erase(Drive.EraseType et)
        {
            return NotSupported("erase");
        }

        private uint SplitLocation(long size_in_sectors)
        {
            Debug.Assert(size_in_sectors <= UInt32.MaxValue);
            uint size = (uint)size_in_sectors;
            uint eccblocks;
            uint eccboundary;

            //
            // Round the size to the nearest ECC block (16 sectors)
            //
            size = ((size + 15) / 16) * 16;

            //
            // Calculate the total # of ECC blocks, add one if odd size or ecc blocks
            //
            eccblocks = size / 16;
            if ((eccblocks % 2) == 1)
                eccblocks++;

            //
            // Find the ECC block boundary for the split
            //
            eccboundary = eccblocks / 2;

            //
            // Convert this back to sectors
            //
            return eccboundary * 16;
        }

        private DiskOperationError SplitDLDisk(long size_in_sectors)
        {
            bool isfixed ;
            uint location ;
            Device.CommandStatus st;
            uint split = SplitLocation(size_in_sectors);

            st = MyDevice.ReadDvdLayer0Size(out isfixed, out location);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            if (isfixed)
                return new BurnerError("InternalError - the DVD+R DL split boundary has already been set") ;

            if (size_in_sectors < location)
                return new BurnerError("the burn data uses less than this DL disk, burn not allowed");

            st = MyDevice.SendDvdLayerBoundaryInformation(split);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            return null; 
        }
    }
}
