//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

//
// TODO List for this burner
//
// o  Support multi-session DVD+RW disks
//

using System;
using System.Collections.Generic;
using System.Text;
using Bwg.Scsi;
using Bwg.Logging;

namespace Bwg.Burner
{
    class DVDPlusRWDiskOperations : DVDBaseDiskOperations
    {
        public const byte FormatTypeFullFormat = 0x26;

        public DVDPlusRWDiskOperations(Drive d) : base(d)
        {
        }

        public override long CapacityInSectors
        {
            get
            {
                //
                // Read the disk formatted capacities, if this fails, assume a normal DVD+RW
                // disk that is empty.
                //
                IList<CapacityDescriptor> caplist;
                if (MyDevice.ReadFormatCapacities(true, out caplist) != Device.CommandStatus.Success)
                    return Drive.D47GigDVDPlusRCapacity;

                //
                // Find the unformatted format descriptor, and return the given number of
                // blocks.
                //
                if (caplist[0].DescType == CapacityDescriptor.DescriptorType.Unformatted)
                    return caplist[0].NumberOfBlocks;

                //
                // Finally search for a FormatType that is of type 0x26 (DVD+RW)
                // and return that number
                //
                for (int i = 1; i < caplist.Count; i++)
                {
                    if (caplist[i].FormatType == FormatTypeFullFormat)
                        return caplist[i].NumberOfBlocks;
                }

                //
                // Finally, nothing of interest was found.  Maybe this is not a DVD+RW or a
                // drive thatis not well behaved.  Return the full capacity of the DVD+RW and
                // let other code deal with the issues if they exist with this drive.
                //
                return Drive.D47GigDVDPlusRCapacity;
            }
        }

        /// <summary>
        /// Returns the writable capacity in sectors.  This is the same as the capacity in sectors
        /// on RW media.
        /// </summary>
        public override long WritableCapacityInSectors
        {
            get
            {
                return CapacityInSectors;
            }
        }

        /// <summary>
        /// Returns the remaining capacity on the disk.
        /// </summary>
        public override long RemainingCapacityInWritableSectors
        {
            get
            {
                DiscInformation info;

                //
                // Since we don't support multi-session burning for DVD+RW yet, a disk is
                // either erased and the capacity is the entire disk, or the disk is used
                // and the capacity is zero.
                //
                // TODO: Update this when we figure out how to support multi-session DVD+RW disks
                //
                if (MyDevice.ReadDiscInformation(out info) != Device.CommandStatus.Success)
                    return 0;

                if (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
                    return WritableCapacityInSectors;

                return 0;
            }
        }

        /// <summary>
        /// Returns true if burning is supported
        /// </summary>
        public override bool IsBurnSupported
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Returns true if multi-session is supported (false for now)
        /// </summary>
        public override bool IsMultiSessionSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Returns true if erase is supported, we do not erase DVD+RW media, we just write our
        /// image starting at sector 0.  Any old image is over-written and the drive will begin
        /// formatting automatically if necessary to have the available space.
        /// </summary>
        public override bool IsEraseSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Returns true if session at once burning is supported
        /// </summary>
        public override bool IsSessionAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDSessionAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        /// <summary>
        /// Returns true if track at once burning is supported
        /// </summary>
        public override bool IsTrackAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDTrackAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        /// <summary>
        /// Burn the disk image to the disk
        /// </summary>
        /// <param name="bt">the type of burn (session versus track)</param>
        /// <param name="d">the disk image to burn</param>
        /// <param name="multi_session">if true, leave open for multi-session burns in the future</param>
        /// <param name="speed">the speed to use to burn</param>
        /// <param name="bytes">return value for the number of bytes sent to the drive during the burn</param>
        /// <returns>null if sucessful, otherwise an error object is returned</returns>
        public override DiskOperationError BurnDisk(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes)
        {
            DiskOperationError status;
            Device.CommandStatus st;

            bytes = 0;

            if (multi_session)
                return new BurnerError("Multi-session burning of DVD+RW is currently not supported");

            if (d.TrackCount != 1)
                return new BurnerError("multiple tracks targeted at a DVD+RW media, DVD+RW only supports a single track");

            if (d.ContainsAudio)
                return new BurnerError("cannot write audio tracks to a DVD+RW media, DVD+RW only supports data tracks");

            DiscInformation info;
            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st) ;

            status = m_drive.SetSpeed(speed);
            if (status != null)
                return status;

            //
            // Set the target media, this sets the start sector for the disk as well as all of
            // the pre-track gap information
            //
            d.StartSector = 0;
            d.TargetMedia = DiskBurnImage.DiskTargetType.DVD;

            bool doformat = false;
            if (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
                doformat = true;
            else
            {
                TrackInformation tinfo;
                st = MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, info.LastTrackInLastSession, out tinfo);

                IList<CapacityDescriptor> caplist;
                st = MyDevice.ReadFormatCapacities(false, out caplist);
                if (st != Device.CommandStatus.Success)
                    doformat = true;
                else if (caplist.Count == 0)
                    doformat = true;
                else if (caplist[0].NumberOfBlocks < d.LengthSectors)
                    doformat = true;
            }

            if (doformat)
            {
                status = FormatMoreRoom();
                if (status != null)
                    return status;
            }
           
            // Create a data generator for this stream of data.  For DVD is it easy.  There is always one
            // track that is a stream of 2048 byte sectors

            DataGenerator gen = new DataGenerator(MyDevice.GetLogger(), 64, MyDevice.MaximumTransferLength, 0, 0, 0);
            gen.AddSource("data", d[0], 0, d[0].DataToSend);

            status = m_drive.SendData(gen, d);
            bytes += (ulong)(d.LengthSectors * 2048);
            if (status != null)
                return status;

            status = m_drive.SynchronizeCache();
            if (status != null)
                return status;

            m_drive.ReadBufferLevels();

            st = MyDevice.ReadDiscInformation(out info);
            if (info.BackgroundFormatStatus == DiscInformation.BackgroundFormatStatusType.InProgress)
            {
                //
                // We have written our data so the format has progressed to the point where our data
                // is written.  Stop the format now and let another burn operation cause the format
                // to restart if necessary.
                //
                status = StopFormat();
                if (status != null)
                    return status;
            }

            // Close the session
            status = m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
            if (status != null)
                return status;

            return null;
        }

        public override DiskOperationError Erase(Drive.EraseType et)
        {
            return NotSupported("erase");
        }

        private CapacityDescriptor FindDesc(IList<CapacityDescriptor> caps, byte fmt)
        {
            foreach (CapacityDescriptor c in caps)
            {
                if (c.FormatType == fmt)
                    return c;
            }
            return null;
        }

        private CapacityDescriptor FindCapDesc()
        {
            Device.CommandStatus st;
            IList<CapacityDescriptor> caps;

            st = MyDevice.ReadFormatCapacities(true, out caps);
            if (st != Device.CommandStatus.Success)
                return null;

            //
            // Remove the current entry because we are only interested
            // in the ones following the current entry
            //
            return FindDesc(caps, FormatTypeFullFormat);
        }

        private DiskOperationError FormatMoreRoom()
        {
            Device.CommandStatus st;
            string str;

            // We need to format the unit, up to the sector needed
            if (MyDevice.GetLogger() != null)
            {
                str = "not enough formatted space - formatting more space" ;
                MyLogger.LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));
            }

            DiscInformation info;
            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st); ;

            CapacityDescriptor desc = FindCapDesc();
            if (desc == null)
            {
                str = "cannot find a format capacity entry via ReadFormatCapacity";
                MyDevice.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Error, 0, str));
                return new BurnerError("cannot find a suitable format descriptor via ReadFormatCapacity");
            }

            FormatParameterList plist = new FormatParameterList();
            plist.IMMED = 1;
            plist.FOV = 1;
            plist.DCRT = 1;
            plist.FormatType = (FormatParameterList.UnitFormatType)desc.FormatType;
            plist.NumberOfBlocks = 0xffffffff;
            plist.Parameter = 0x00;
            st = MyDevice.FormatUnit(plist);

            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            m_drive.WaitForDeviceReady(false);

            str = "background format is running, now back to writing data";
            MyDevice.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));

            return null;
        }

        private DiskOperationError StopFormat()
        {
            string str;
            DiskOperationError status;

            str = "data is all written, but background format is running, shutting down the format";
            MyDevice.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));

            status = m_drive.CloseTrackSession(Device.CloseTrackSessionType.CloseSession, 0);
            if (status != null)
                return status;

            str = "background format is shutdown, disk is good to go";
            MyDevice.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));

            return null;
        }
    }
}
