//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO ;
using Bwg.Burner;
using Bwg.Scsi;
using Bwg.Logging;

namespace Bwg.Burner
{
    /// <summary>
    /// This class represents a disk that is in the drive
    /// </summary>
    public class Disk
    {
        #region private members
        private bool m_valid;
        private Drive m_drive;
        private DiscInformation m_info;
        private MinuteSecondFrame m_leadin_start;
        private MinuteSecondFrame m_leadout_start;
        private CDTextEncoderDecoder m_encoder;
        #endregion

        #region public members

        /// <summary>
        /// The sessions that are on this disk
        /// </summary>
        public Session[] Sessions;

        #endregion

        #region contructors
        /// <summary>
        /// This class represents a disk that is in a drive.
        /// </summary>
        /// <param name="drv">the drive that holds this disk</param>
        public Disk(Drive drv)
        {
            m_drive = drv;
            m_info = null;
            m_valid = true;
            Sessions = new Session[0];

            GetInfo();
            ExtractSessions();
        }
        #endregion

        #region public properties

        /// <summary>
        /// The CDText associated with this disk
        /// </summary>
        public CDTextEncoderDecoder CDText
        {
            get
            {
                return m_encoder;
            }
        }

        /// <summary>
        /// If true, the disc has been finalized
        /// </summary>
        public bool IsFinalized
        {
            get
            {
                return m_info.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc;
            }
        }

        /// <summary>
        /// If true, the disc is incomplete
        /// </summary>
        public bool IsIncomplete
        {
            get
            {
                return m_info.DiscStatus == DiscInformation.DiscStatusType.IncompleteDisc;
            }
        }

        /// <summary>
        /// This method returns true if the disk is empty
        /// </summary>
        public bool IsEmpty
        {
            get
            {
                return m_info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc;
            }
        }

        /// <summary>
        /// This returns TRUE if this disc is an audio disc
        /// </summary>
        public bool IsAudio
        {
            get
            {
                if (Sessions.GetLength(0) != 1)
                    return false;

                foreach (Track t in Sessions[0].Tracks)
                {
                    if (!t.IsAudio)
                        return false;
                }
                return true;
            }
        }

        /// <summary>
        /// Start of the leadin
        /// </summary>
        public MinuteSecondFrame LeadInStart
        {
            get
            {
                return m_leadin_start;
            }
        }

        /// <summary>
        /// Start of the leadout
        /// </summary>
        public MinuteSecondFrame LeadOutStart
        {
            get
            {
                return m_leadout_start;
            }
        }

        /// <summary>
        /// The drive this disk is located in
        /// </summary>
        public Drive Drive
        {
            get
            {
                return m_drive;
            }
        }


        /// <summary>
        /// Returns TRUE if the disk structure is valid, FALSE otherwise
        /// </summary>
        public bool IsValid
        {
            get
            {
                return m_valid;
            }
        }

        /// <summary>
        /// The total number of tracks recorded on the disk.
        /// </summary>
        public int RecordedTrackCount
        {
            get
            {
                Device.CommandStatus st;

                if (!m_valid)
                    return -1;

                IList<TocEntry> toc;
                st = m_drive.Device.ReadToc(0x00, true, out toc);
                if (st != Device.CommandStatus.Success)
                {
                    m_valid = false;
                    return -1;
                }

                return toc.Count - 1;
            }
        }

        /// <summary>
        /// The type string for the media in the drive.
        /// </summary>
        public string MediaType
        {
            get
            {
                string ret = "Error";
                if (!m_valid)
                    return "Error";

                switch (m_drive.CurrentProfile)
                {
                    case Drive.SupportedProfiles.CD_R:
                        ret = "CD-R";
                        break;

                    case Drive.SupportedProfiles.CD_ROM:
                        ret = "CDROM";
                        break;

                    case Drive.SupportedProfiles.CD_RW:
                        ret = "CD-RW";
                        break;

                    case Drive.SupportedProfiles.DVD_PLUS_R:
                        ret = "DVD+R";
                        break;

                    case Drive.SupportedProfiles.DVD_PLUS_RW:
                        ret = "DVD+RW";
                        break;

                    case Drive.SupportedProfiles.DVD_R_SequentialRecording:
                        ret = "DVD-R";
                        break;

                    case Drive.SupportedProfiles.DVD_RAM:
                        ret = "DVD-RAM";
                        break;

                    case Drive.SupportedProfiles.DVD_ROM:
                        ret = "DVD-ROM";
                        break;

                    case Drive.SupportedProfiles.DVD_RW_SequentialRecording:
                    case Drive.SupportedProfiles.DVD_RW_RestrictedOverwrite:
                    case Drive.SupportedProfiles.DVD_RW_RestrictedOverwrite_And_Sequential:
                        ret = "DVD-RW";
                        break;

                    default:
                        ret = "Unknown";
                        break;
                }

                return ret;
            }
        }

        #endregion

        #region private methods

        private int GetSize(TocEntry toc, TrackInformation info)
        {
            return 2352;
        }

        private void GetInfo()
        {
            Device.CommandStatus st;
            m_encoder = null;

            if (m_drive == null)
                return;

            st = m_drive.Device.ReadDiscInformation(out m_info);
            if (st != Device.CommandStatus.Success)
            {
                m_valid = false;
                return;
            }

            TrackInformation tinfo;
            st = m_drive.Device.ReadTrackInformation(Device.ReadTrackType.TrackNumber, 0, out tinfo);
            if (st != Device.CommandStatus.Success)
                m_leadin_start = new MinuteSecondFrame(0, 0, 0);
            else
                m_leadin_start = new MinuteSecondFrame(tinfo.TrackStartAddress);

            // No need to read the TOC or anything else on an empty disk
            if (m_info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
                return;

            IList<TocEntry> toc;
            st = m_drive.Device.ReadToc(0x00, true, out toc);
            if (st != Device.CommandStatus.Success)
            {
                m_valid = false ;
                return ;
            }
            m_leadout_start = toc[toc.Count - 1].StartMSF;

            DiskOperationError status = m_drive.Initialize();
            if (status != null)
            {
                m_valid = false;
                return;
            }

            m_drive.ReadCDText(out m_encoder);
        }

        private Session AddSession(int number)
        {
            Session s;

            if (number - 1 < Sessions.GetLength(0))
                return Sessions[number - 1];

            Debug.Assert(number - 1 == Sessions.GetLength(0));
            s = new Session(number);

            Array.Resize(ref Sessions, number);

            Sessions[number - 1] = s;
            return s;
        }

        private void ExtractSessions()
        {
            Device.CommandStatus st;

            for (int i = 1; i <= RecordedTrackCount; i++)
            {
                TrackInformation info;
                st = m_drive.Device.ReadTrackInformation(Device.ReadTrackType.TrackNumber, (uint)i, out info);
                if (st != Device.CommandStatus.Success || info.SessionNumber > 256)
                {
                    m_valid = false;
                    return;
                }

                Session session = AddSession(info.SessionNumber);
                Track t = new Track(info);

                if (m_drive.IsCd && info.DataMode != TrackInformation.DataModeType.Audio)
                {
                    // For CD's there may be more information to assocaite with a track
                    HeaderData hdr;
                    SubheaderData subhdr;

                    st = m_drive.Device.ReadCD((uint)info.TrackStartAddress, out hdr);
                    if (st == Device.CommandStatus.Success)
                    {
                        t.Header = hdr;
                        if (t.Header.DataType == HeaderData.DataTypeType.Mode2Data)
                        {
                            st = m_drive.Device.ReadCD((uint)info.TrackStartAddress, out subhdr);
                            if (st == Device.CommandStatus.Success)
                                t.Subheader = subhdr;
                        }
                    }
                }

                session.AddTrack(t);
            }
        }
        #endregion
    }
}
