//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Bwg.Scsi ;

namespace Bwg.Burner
{
    /// <summary>
    /// This class represents a single track on a CD or DVD to be burned
    /// </summary>
    public class TrackBurnImage
    {
        #region private member variables
        /// <summary>
        /// The start sector for this track if we are doing SAO recording.  If we are doing
        /// TAO recording, the drive supplies the track address.
        /// </summary>
        private long m_start_sector;

        /// <summary>
        /// The track number for this track.
        /// </summary>
        private int m_track_number;

        /// <summary>
        /// If true, use pre-emphasis in recording the audio track
        /// </summary>
        private bool m_premphasis;

        /// <summary>
        /// This is the pre-track gap
        /// </summary>
        private long m_pretrack_gap;

        /// <summary>
        /// The ISRC number for the track
        /// </summary>
        private byte[] m_isrc;

        /// <summary>
        /// The list of CD text fields attached to this disk
        /// </summary>
        private IList<CDTextField> m_cdtext;

        #endregion

        #region public types
        /// <summary>
        /// The type of the track
        /// </summary>
        public enum TrackType
        {
            /// <summary>
            /// Raw track data
            /// </summary>
            Raw,

            /// <summary>
            /// The leadin information
            /// </summary>
            Leadin,

            /// <summary>
            /// The cd text information
            /// </summary>
            CDText,

            /// <summary>
            /// Audio track, the data source gives the type of audio
            /// </summary>
            Audio,

            /// <summary>
            /// Data track, 2048 bytes of data per sector
            /// </summary>
            Data_Mode_1,

            /// <summary>
            /// Data mode 2, formless, 2336 bytes of data per sector
            /// </summary>
            Data_Mode_2_Formless,

            /// <summary>
            /// Data mode 2, form 1, 2048 bytes of data per sector
            /// </summary>
            Data_Mode_2_Form_1,

            /// <summary>
            /// Data mode 2, form 2, 2324 bytes of data per sector
            /// </summary>
            Data_Mode_2_Form_2,
        } ;
        #endregion

        #region public data members
        /// <summary>
        /// The type of the track
        /// </summary>
        public readonly TrackType Type;

        /// <summary>
        /// The source of the data for this track
        /// </summary>
        public readonly DataSource Data;

        #endregion

        #region constructors

        /// <summary>
        /// Constructor for a new track object
        /// </summary>
        /// <param name="t">the track type (audio, data_mode1, etc.)</param>
        /// <param name="s">the data source for the track</param>
        public TrackBurnImage(TrackType t, DataSource s)
        {
            Type = t;
            Data = s;
            m_premphasis = false;
            m_pretrack_gap = 0;
            m_isrc = null ;
            m_cdtext = new List<CDTextField>();
        }

        /// <summary>
        /// Construct a new track object
        /// </summary>
        /// <param name="t">the track type (audio, data_mode1, etc.)</param>
        /// <param name="s">the data source for the track</param>
        /// <param name="isrc">the isrc number for the track</param>
        public TrackBurnImage(TrackType t, DataSource s, byte[] isrc) : this(t,s)
        {
            Debug.Assert(isrc.GetLength(0) == 12);

            m_isrc = isrc;
        }
        #endregion

        #region public properties

        /// <summary>
        /// This property returns the list of text fields associated with this 
        /// audio track.
        /// </summary>
        public IList<CDTextField> CDTextList
        {
            get
            {
                Debug.Assert(Type == TrackType.Audio);
                return m_cdtext;
            }
        }

        /// <summary>
        /// If true, this track has ISRC data
        /// </summary>
        public bool HasIsrc
        {
            get
            {
                return m_isrc != null;
            }
        }

        /// <summary>
        /// The ISRC data for this track
        /// </summary>
        public byte[] ISRC
        {
            get
            {
                return m_isrc;
            }
            set
            {
                if (value != null && value.GetLength(0) != 12)
                    throw new Exception("the ISRC must be a byte array of size 12");

                m_isrc = value;
            }
        }

        /// <summary>
        /// Return the name of the track (actually the data source name)
        /// </summary>
        public string Name
        {
            get
            {
                return Data.Name;
            }
        }

        /// <summary>
        /// This track number for this track
        /// </summary>
        public int TrackNumber
        {
            get
            {
                return m_track_number;
            }

            set
            {
                m_track_number = value;
            }
        }

        /// <summary>
        /// Return the start time for this track in MSF format.  This only applies to
        /// recording using SAO.  With TAO, the drive supplies the start time for a 
        /// track.
        /// </summary>
        public MinuteSecondFrame StartMSF
        {
            get
            {
                return new MinuteSecondFrame(m_start_sector);
            }
        }

        /// <summary>
        /// Return the start time for this track in sectors
        /// </summary>
        public long StartSector
        {
            get
            {
                return m_start_sector;
            }
            set
            {
                m_start_sector = value;
            }
        }

        /// <summary>
        /// Return the sector size in bytes for this track
        /// </summary>
        public int SectorSize
        {
            get
            {
                int s = 0;

                switch (Type)
                {
                    case TrackType.Audio:
                        s = 2352;
                        break;

                    case TrackType.Data_Mode_1:
                        s = 2048;
                        break;

                    case TrackType.Data_Mode_2_Formless:
                        s = 2336;
                        break;

                    case TrackType.Data_Mode_2_Form_1:
                        s = 2048;
                        break;

                    case TrackType.Data_Mode_2_Form_2:
                        s = 2324;
                        break;

                    case TrackType.Leadin:
                        s = 2048;
                        break;

                    case TrackType.CDText:
                        s = 96;
                        break;

                    default:
                        Debug.Assert(false, "unknown track type");
                        break;
                }

                return s;
            }
        }

        /// <summary>
        /// Return the length of this track in sectors, note sector size may differ based on the
        /// type of the track
        /// </summary>
        public long TotalLengthSectors
        {
            get
            {
                return Data.SectorCount + m_pretrack_gap;
            }
        }

        /// <summary>
        /// Return the length of the track data only.  Does not include any
        /// pregap length
        /// </summary>
        public long DataOnlyLength
        {
            get
            {
                return Data.SectorCount ;
            }
        }

        /// <summary>
        /// Returns the amount of data this track must send to the drive.
        /// </summary>
        public long DataToSend
        {
            get
            {
                return Data.SectorCount + m_pretrack_gap ;
            }
        }

        /// <summary>
        /// The size of the pre-track gap
        /// </summary>
        public long PreTrackGap
        {
            get
            {
                return m_pretrack_gap;
            }
            set
            {
                m_pretrack_gap = value;
            }
        }

        /// <summary>
        /// Applies only to audio tracks, if true record with pre-emphasis
        /// </summary>
        public bool PreEmphasis
        {
            get
            {
                Debug.Assert(Type == TrackType.Audio);
                return m_premphasis;
            }

            set
            {
                Debug.Assert(Type == TrackType.Audio);
                m_premphasis = value;
            }
        }

        #endregion

        #region public methods

        /// <summary>
        /// Find a CD text field given its type and language
        /// </summary>
        /// <param name="t">the type of the cd text field entry</param>
        /// <param name="blocknr">the language for the cd text field entry</param>
        /// <returns>the entry if it exists, or null</returns>
        public CDTextField FindCDTextField(CDTextField.TextFieldType t, byte blocknr)
        {
            foreach (CDTextField tmp in m_cdtext)
            {
                if (tmp.LanguageIndex == blocknr && tmp.Type == t)
                    return tmp;
            }
            return null;
        }

        /// <summary>
        /// Add a new CD text field to the track
        /// </summary>
        /// <param name="f">The field to add</param>
        /// <returns>true if it did not already exist and was added, false if it already exists</returns>
        public bool AddCDTextField(CDTextField f)
        {
            if (FindCDTextField(f.Type, f.LanguageIndex) != null)
                return false;

            m_cdtext.Add(f);
            return true;
        }

        /// <summary>
        /// Remove the CD text field of a given type and language
        /// </summary>
        /// <param name="t">the type of text field</param>
        /// <param name="blocknr">the language of interest</param>
        /// <returns></returns>
        public bool RemoveCDTextField(CDTextField.TextFieldType t, byte blocknr)
        {
            for (int i = 0; i < m_cdtext.Count; i++)
            {
                CDTextField tmp = m_cdtext[i];
                if (tmp.LanguageIndex == blocknr && tmp.Type == t)
                {
                    m_cdtext.RemoveAt(i);
                    return true;
                }
            }

            return false;
        }

        #endregion
    }
}
