//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Bwg.Logging;
using Bwg.Isofs;
using Bwg.Utils;

namespace Bwg.Controls
{
    /// <summary>
    /// This control displays an image of a disk that is being created via a DiskImageCreator
    /// </summary>
    public partial class DiskImageTreeView : TreeView
    {
        #region public member variables
        /// <summary>
        /// This is a list of controls to disable while processing a drop operation
        /// </summary>
        public IList<Control> DisableControls;

        #endregion

        #region private types

        /// <summary>
        /// This type of operation associated with an active label edit
        /// </summary>
        private enum EditMode
        {
            CreatingDirectory,
            RenamingDirectory
        } ;

        /// <summary>
        /// The delegate type for changing the name of a node to match the name of
        /// the attached DiskImageElement.
        /// </summary>
        /// <param name="node">the node to change</param>
        private delegate void ChangeNameDelegate(TreeNode node);

        #endregion

        #region private member variables
        private DiskImageElement.NameType m_nametype ;
        private DiskImageContents m_image;
        private Point m_menu_point;
        private EditMode m_mode;
        private System.Windows.Forms.Timer m_timer;
        private IList<TreeNode> m_name_change_list;
        private ShellIconMgr m_icon_mgr;
        #endregion

        #region constructors
        /// <summary>
        /// Create the control
        /// </summary>
        public DiskImageTreeView()
        {
            m_image = null;
            InitializeComponent();
            AllowDrop = true ;
            LabelEdit = true;

            m_name_change_list = new List<TreeNode>();

            ContextMenu = new ContextMenu();
            ContextMenu.MenuItems.Add("New", new EventHandler(CreateDirectoryHandler));
            ContextMenu.MenuItems.Add("Delete", new EventHandler(RemoveDirectoryHandler));
            ContextMenu.MenuItems.Add("Rename", new EventHandler(RenameDirectoryHandler));

            m_nametype = DiskImageElement.NameType.Local;

            DisableControls = new List<Control>();
            DisableControls.Add(this);
        }
        #endregion

        #region public properties

        /// <summary>
        /// This property gets/sets the type of anme to use in the disk image tree
        /// </summary>
        public DiskImageElement.NameType NameType
        {
            get
            {
                return m_nametype;
            }
            set
            {
                m_nametype = value;
                RefreshAll(Nodes);
            }
        }

        /// <summary>
        /// This property sets or removes the icon manager assocaited with this list view
        /// </summary>
        public ShellIconMgr IconManager
        {
            get
            {
                return m_icon_mgr;
            }
            set
            {
                m_icon_mgr = value;
                if (m_icon_mgr != null)
                    ImageList = m_icon_mgr.ImageList;
            }
        }

        /// <summary>
        /// The DiskImageCreator associated with the view
        /// </summary>
        public DiskImageContents ImageCreator
        {
            get
            {
                return m_image;
            }
            set
            {
                if (m_image != null)
                {
                    m_image.ImageChanged -= new EventHandler<ImageChangedEventArgs>(Creator_ImageChanged);
                    m_image.AddingDirectory -= new EventHandler<AddingDirectoryEventArgs>(Creator_AddingDirectory);
                }

                m_image = value;
                Nodes.Clear();

                if (m_image != null)
                {
                    if (m_image != null)
                    {
                        m_image.ImageChanged += new EventHandler<ImageChangedEventArgs>(Creator_ImageChanged);
                        m_image.AddingDirectory += new EventHandler<AddingDirectoryEventArgs>(Creator_AddingDirectory);
                    }
                    TreeNode node = new TreeNode("/");
                    node.Tag = m_image.RootDirectory;
                    node.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);
                    Nodes.Add(node);

                    Populate(node);
                }
            }
        }
        #endregion

        #region private and protected event handler

        private void Creator_AddingDirectory(object sender, AddingDirectoryEventArgs args)
        {
            // Find the node associated with the root
            TreeNode node = FindNode(Nodes, args.Root);
            TreeNode newdir = new TreeNode(args.Dir.GetName(m_nametype));
            newdir.Tag = args.Dir;
            node.Nodes.Add(newdir);
        }

        /// <summary>
        /// This method is called before a node in the disk image tree is expanded.  This method updates
        /// the image of the folder to an open folder image
        /// </summary>
        /// <param name="e">parameters for the expand, not used</param>
        protected override void OnBeforeExpand(TreeViewCancelEventArgs e)
        {
            e.Node.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, true);
            base.OnBeforeExpand(e);
        }

        /// <summary>
        /// This method is called after a node in the disk image tree is collapsed.  This method updates
        /// the image of the folder to a closed folder image
        /// </summary>
        /// <param name="e"></param>
        protected override void OnAfterCollapse(TreeViewEventArgs e)
        {
            e.Node.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);
            base.OnAfterCollapse(e);
        }

        /// <summary>
        /// Called when the mouse button is pressed in this control
        /// </summary>
        /// <param name="e">the location of the event</param>
        protected override void OnMouseDown(MouseEventArgs e)
        {
            m_menu_point = e.Location;
            if (HitTest(m_menu_point).Node != null)
                base.OnMouseDown(e);
        }

        private void CreateDirectoryHandler(object sender, EventArgs args)
        {
            TreeViewHitTestInfo info = HitTest(m_menu_point);
            if (info.Node != null)
            {
                TreeNode node = new TreeNode("New Directory");
                node.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);
                info.Node.Nodes.Add(node);
                info.Node.Expand();
                m_mode = EditMode.CreatingDirectory;
                node.BeginEdit();
            }
        }

        private void RemoveDirectoryHandler(object sender, EventArgs args)
        {
            TreeViewHitTestInfo info = HitTest(m_menu_point);
            if (info.Node != null)
            {
                if (info.Node.Tag is DiskImageElement)
                    m_image.Remove((DiskImageElement)info.Node.Tag);
            }
        }

        private void RenameDirectoryHandler(object sender, EventArgs args)
        {
            TreeViewHitTestInfo info = HitTest(m_menu_point);
            if (info.Node != null && info.Node.Tag is DiskImageElement)
            {
                DiskImageElement elem = (DiskImageElement)info.Node.Tag;
                if (elem.Parent != null)
                {
                    m_mode = EditMode.RenamingDirectory;
                    info.Node.BeginEdit();
                }
            }
        }

        static private void TimerTick(object sender, EventArgs args)
        {
            System.Windows.Forms.Timer t = (System.Windows.Forms.Timer)sender ;
            DiskImageTreeView view = (DiskImageTreeView)t.Tag;

            view.TimerTick();
        }

        private void ChangeNodeName(TreeNode n)
        {
            if (InvokeRequired)
            {
                object [] args = new object[1] ;
                args[0] = n ;
                Invoke(new ChangeNameDelegate(ChangeNodeName), args) ;
            }
            else
            {
                if (n.Tag is DiskImageElement)
                {
                    DiskImageElement elem = (DiskImageElement)n.Tag ;
                    n.Text = elem.GetName(m_nametype) ;
                }
            }
        }

        private void TimerTick()
        {
            foreach (TreeNode node in m_name_change_list)
                ChangeNodeName(node);

            m_timer.Stop();
        }


        private void ScheduleNameChange(TreeNode n)
        {
            m_name_change_list.Add(n);

            m_timer = new System.Windows.Forms.Timer();
            m_timer.Tag = this;
            m_timer.Tick += new EventHandler(TimerTick);
            m_timer.Interval = 10;
            m_timer.Start();
        }

        /// <summary>
        /// Called after the completion of a node name edit operation
        /// </summary>
        /// <param name="e">the node that was edited</param>
        protected override void OnAfterLabelEdit(NodeLabelEditEventArgs e)
        {
            if (e.Node.Parent == null || e.Label == null)
                return;

            DiskImageDirectory parent = (DiskImageDirectory)e.Node.Parent.Tag;
            if (m_mode == EditMode.CreatingDirectory)
            {
                DiskImageDirectory dir = m_image.AddNewDir(parent, e.Label);
                if (dir == null)
                    e.CancelEdit = true;
                else
                {
                    e.Node.Tag = dir;
                    ScheduleNameChange(e.Node);
                }
            }
            else if (m_mode == EditMode.RenamingDirectory)
            {
                DiskImageElement elem = (DiskImageElement)e.Node.Tag;
                if (m_image.RenameElement(parent, elem, e.Label))
                    ScheduleNameChange(e.Node);
                else
                    e.CancelEdit = true;
            }
            base.OnAfterLabelEdit(e);
        }

        /// <summary>
        /// This is called when a key is pressed
        /// </summary>
        /// <param name="e"></param>
        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete)
            {
                if (SelectedNode != null && SelectedNode.Tag != null && SelectedNode.Tag is DiskImageElement)
                    m_image.Remove((DiskImageElement)SelectedNode.Tag);
            }
            else
            {
                base.OnKeyDown(e);
            }
        }

        /// <summary>
        /// Called when a file or directory is dropped on the image view
        /// </summary>
        /// <param name="drgevent">the arguments giving the directory or file dropped</param>
        protected override void OnDragDrop(DragEventArgs drgevent)
        {
            base.OnDragDrop(drgevent);

            BeginUpdate();
            m_image.WithinUpdate = true;

            object obj = drgevent.Data.GetData("FileNameW", true);
            string[] namearray = (string[])obj;

            Application.UseWaitCursor = true;
            foreach (Control c in DisableControls)
                c.Enabled = false;

            // Now we must add the file or directory to the ISO image based
            // on the drop location
            Point pt = PointToClient(new Point(drgevent.X, drgevent.Y));
            TreeViewHitTestInfo info = HitTest(pt);

            // Find the node in the directory tree
            DiskImageDirectory dir = (DiskImageDirectory)info.Node.Tag;

            foreach (string fname in namearray)
            {
                if (File.Exists(fname))
                {
                    string filename = Path.GetFileName(fname);
                    string isopath = dir.GetPath(m_nametype) + "/" + filename;
                    try
                    {
                        m_image.Add(isopath, fname);
                    }
                    catch (Exception ex)
                    {
                        string str = "Error adding file '" + fname + "' to image - " + ex.Message;
                        MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
                else if (Directory.Exists(fname))
                {
                    string filename = Path.GetFileName(fname);
                    string isopath = dir.GetPath(m_nametype) + "/" + filename;
                    try
                    {
                        m_image.Add(isopath, fname);
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show("Error adding files to image - " + ex.Message, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
                else
                {
                    string msg = "The dropped element '" + fname + "' does not exist";
                    MessageBox.Show(msg, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

            EndUpdate();
            m_image.WithinUpdate = false;

            Application.UseWaitCursor = false;
            foreach (Control c in DisableControls)
                c.Enabled = true;
        }

        /// <summary>
        /// Called when a drag and drop operation is drug over the image window
        /// </summary>
        /// <param name="drgevent"></param>
        protected override void OnDragOver(DragEventArgs drgevent)
        {
            base.OnDragOver(drgevent);

            Point pt = PointToClient(new Point(drgevent.X, drgevent.Y));
            TreeViewHitTestInfo info = HitTest(pt);
            if (info.Node != null && (info.Node.Tag is DiskImageDirectory) && drgevent.Data.GetDataPresent("FileNameW", true))
                drgevent.Effect = DragDropEffects.Copy;
            else
                drgevent.Effect = DragDropEffects.None;
        }
        #endregion

        #region private member functions

        private void RefreshAll(TreeNodeCollection nodes)
        {
            foreach (TreeNode node in nodes)
            {
                if (node.Tag != null && node.Tag is DiskImageElement)
                {
                    DiskImageElement elem = (DiskImageElement)node.Tag;
                    node.Text = elem.GetName(m_nametype);

                    if (node.Nodes.Count > 0)
                        RefreshAll(node.Nodes);
                }
            }
        }

        private TreeNode FindNodeByElem(TreeNodeCollection nodes, DiskImageElement elem)
        {
            foreach (TreeNode n in nodes)
            {
                if (n.Tag != null && n.Tag is DiskImageElement && (DiskImageElement)n.Tag == elem)
                    return n;

                TreeNode check = FindNodeByElem(n.Nodes, elem);
                if (check != null)
                    return check;
            }

            return null;
        }

        private TreeNode FindNode(TreeNodeCollection nodes, DiskImageElement elem)
        {
            foreach (TreeNode n in nodes)
            {
                if (n.Tag == elem)
                    return n;

                TreeNode tn = FindNode(n.Nodes, elem);
                if (tn != null)
                    return tn;
            }

            return null;
        }

        private void Creator_ImageChanged(object sender, ImageChangedEventArgs args)
        {
            TreeNode node = FindNodeByElem(Nodes, args.Element);

            if (node != null)
            {
                switch (args.Change)
                {
                    case ImageChangedEventArgs.ChangeType.Removed:
                        node.Remove();
                        break;

                    case ImageChangedEventArgs.ChangeType.Renamed:
                        node.Text = ((DiskImageElement)node.Tag).GetName(m_nametype);
                        break;
                }
            }
        }

        private void Populate(TreeNode node)
        {
            if (node.Tag is DiskImageDirectory)
            {
                DiskImageDirectory dir = (DiskImageDirectory)node.Tag;

                foreach (DiskImageElement elem in dir)
                {
                    if (elem is DiskImageDirectory)
                    {
                        TreeNode newnode = new TreeNode(elem.GetName(m_nametype));
                        newnode.Tag = elem;
                        newnode.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);

                        elem.Tag = newnode;

                        node.Nodes.Add(newnode);

                        elem.NameChangeEvent += new EventHandler<NameChangeEventArgs>(NodeNameChanged);

                        Populate(newnode);
                    }
                }
            }
        }

        private void NodeNameChanged(object sender, NameChangeEventArgs args)
        {
            if (sender is DiskImageElement)
            {
                DiskImageElement elem = (DiskImageElement)sender;
                TreeNode treenode = (TreeNode)elem.Tag;
                treenode.Text = args.NewName;
            }
        }

        #endregion
    }
}
