//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using Bwg.Burner;

namespace Bwg.Controls
{
    /// <summary>
    /// This form displays and allows the user to edit settings associated with the disk
    /// or drive
    /// </summary>
    public partial class DriveSettings : Form
    {
        /// <summary>
        /// The debug level if debug messages are displayed
        /// </summary>
        public int DebugLevel;
        
        /// <summary>
        /// If true, leave the disk open for future sessions
        /// </summary>
        public bool LeaveSessionOpen;

        /// <summary>
        /// If true, display debug messages
        /// </summary>
        public bool DebugMessagesFlag;

        /// <summary>
        /// If true, display warning messages
        /// </summary>
        public bool WarningMessagesFlag;

        /// <summary>
        /// This indicates the burntype
        /// </summary>
        public Drive.BurnType BurnType;

        /// <summary>
        /// The speed to use in burning a disk
        /// </summary>
        public SpeedInfo BurnSpeedValue;

        /// <summary>
        /// Constructor for the form that displays user settings
        /// </summary>
        /// <param name="debug">if true, display debug messages</param>
        /// <param name="level">the level of debug message to display</param>
        /// <param name="warning">if true, display warning messages</param>
        /// <param name="leave">if true, leave the disk in a state where other sessions can be added</param>
        /// <param name="btype">this parameter gives the current burn type</param>
        /// <param name="speeds">the list of speeds supported</param>
        /// <param name="istao">if true, it indicates the media supports track at once burning</param>
        /// <param name="issao">if true, it indicates the media supports session at once burning</param>
        /// <param name="isms">if true, it indicates the media supports multi-session burning</param>
        public DriveSettings(bool debug, int level, bool warning, bool leave, Drive.BurnType btype, 
                             SpeedInfo [] speeds, bool istao, bool issao, bool isms)
        {
            DebugMessagesFlag = debug;
            WarningMessagesFlag = warning;
            LeaveSessionOpen = leave;
            BurnType = btype;
            DebugLevel = level;

            InitializeComponent();

            if (DebugMessagesFlag)
                DebugMessages.Checked = true;
            else
                DebugMessages.Checked = false;

            if (WarningMessagesFlag)
                WarningMessages.Checked = true;
            else
                WarningMessages.Checked = false;

            if (LeaveSessionOpen)
                LeaveOpen.Checked = true;
            else
                FinalizeDisk.Checked = true;


            for(int i = 0 ; i < DebugLevelList.Items.Count ; i++)
            {
                int result ;
                string str = (string)DebugLevelList.Items[i] ;
                str = str.Substring(6) ;

                if (Int32.TryParse(str, out result) && result == DebugLevel)
                {
                    DebugLevelList.SelectedIndex = i ;
                    break ;
                }
            }

            if (speeds.GetLength(0) > 0)
            {
                for (int i = 0; i < speeds.GetLength(0); i++)
                    BurnSpeed.Items.Add(speeds[i]);

                BurnSpeed.SelectedIndex = 0;
                BurnSpeedValue = speeds[0];
            }
            else
            {
                BurnSpeed.Items.Add("Maximum");
                BurnSpeedValue = null;
            }

            if (istao && issao || (!istao && !issao))
            {
                //
                // Ok, if both TAO is not current and SAO is not current, something is wrong or
                // the drive is giving me bad data or I am not interpreseting this situation correctly.  In
                // this case I assume that either method of writing is ok.
                //
                BurnTypeList.Enabled = true;
                switch (BurnType)
                {
                    case Drive.BurnType.DontCare:
                        BurnTypeList.SelectedIndex = 0;
                        break;

                    case Drive.BurnType.Session:
                        BurnTypeList.SelectedIndex = 1;
                        break;

                    case Drive.BurnType.Track:
                        BurnTypeList.SelectedIndex = 2;
                        break;
                }
            }
            else if (istao)
            {
                BurnTypeList.SelectedIndex = 0;
                BurnTypeList.Enabled = false;
            }
            else if (issao)
            {
                BurnTypeList.SelectedIndex = 1;
                BurnTypeList.Enabled = false;
            }

            if (!isms)
            {
                FinalizeDisk.Checked = true;
                FinalizeDisk.Enabled = false;
                LeaveOpen.Enabled = false;
            }
        }

        private void DebugMessages_CheckedChanged(object sender, EventArgs e)
        {
            if (DebugMessages.Checked)
                DebugMessagesFlag = true;
            else
                DebugMessagesFlag = false;
        }

        private void WarningMessages_CheckedChanged(object sender, EventArgs e)
        {
            if (WarningMessages.Checked)
                WarningMessagesFlag = true;
            else
                WarningMessagesFlag = false;
        }

        private void LeaveOpen_CheckedChanged(object sender, EventArgs e)
        {
            LeaveSessionOpen = true;
        }

        private void FinalizeDisk_CheckedChanged(object sender, EventArgs e)
        {
            LeaveSessionOpen = false;
        }

        private void BurnTypeList_SelectedIndexChanged(object sender, EventArgs e)
        {
            switch (BurnTypeList.SelectedIndex)
            {
                case 0:
                    BurnType = Drive.BurnType.DontCare;
                    break;

                case 1:
                    BurnType = Drive.BurnType.Session;
                    break;

                case 2:
                    BurnType = Drive.BurnType.Track;
                    break;
            }
        }

        private void DebugLevelList_SelectedIndexChanged(object sender, EventArgs e)
        {
            int result ;
            string str = (string)DebugLevelList.SelectedItem;
            str = str.Substring(6);

            if (Int32.TryParse(str, out result) && result >= 0 && result <= 9)
                DebugLevel = result;
            else
                DebugLevel = 9;
        }

        private void BurnSpeed_SelectedIndexChanged(object sender, EventArgs e)
        {
            BurnSpeedValue = null;
            if (BurnSpeed.SelectedItem is SpeedInfo)
                BurnSpeedValue = (SpeedInfo)BurnSpeed.SelectedItem;
        }
    }
}