//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.Schema;
using System.IO;
using System.Reflection;

namespace Bwg.Controls
{
    /// <summary>
    /// This class reads help text from an XML file and returns the requested
    /// help strings based on an associated help tag.
    /// </summary>
    public class HelpTextManager
    {
        #region private member variables
        private IDictionary<string, string> m_help_text;
        #endregion

        #region constructors
        /// <summary>
        /// Construct the help manager
        /// </summary>
        public HelpTextManager()
        {
            m_help_text = new Dictionary<string, string>();
        }
        #endregion

        #region public member functions
        /// <summary>
        /// Return the help associated with the given tag, returns an empty
        /// string if there is no help associated with the tag
        /// </summary>
        /// <param name="tag">the help tag</param>
        /// <returns>the help string</returns>
        public string GetHelpString(string tag)
        {
            string ret = "" ;

            if (m_help_text.ContainsKey(tag))
                ret = m_help_text[tag] ;

            return ret ;
        }

        /// <summary>
        /// Load the help file
        /// </summary>
        /// <param name="filename">the name of the file file</param>
        /// <returns>true if the file was read sucessfully, false othersise</returns>
        public bool LoadHelpFile(string filename)
        {
            bool ret = true;

            try
            {
                XmlSchema xsc ;

                string xsdname = FindDataFileName("HelpFileSchema.xsd");
                XmlSchemaSet xscset = new XmlSchemaSet();
                using(Stream strm = new FileStream(xsdname, FileMode.Open, FileAccess.Read))
                {
                    xsc = XmlSchema.Read(strm, new ValidationEventHandler(MyValidate));
                    xscset.Add(xsc);
                    xscset.Compile();
                }

                string xmlhelp = FindDataFileName(filename);
                using (Stream strm = new FileStream(xmlhelp, FileMode.Open, FileAccess.Read))
                {
                    string tag = "", text = "";
                    XmlReaderSettings settings = new XmlReaderSettings();
                    settings.ValidationType = ValidationType.Schema;
                    settings.ValidationFlags |= XmlSchemaValidationFlags.ReportValidationWarnings;
                    settings.Schemas = xscset;
                    settings.ValidationEventHandler += new ValidationEventHandler(MyXmlValidate);
                    XmlReader rdr = XmlReader.Create(strm, settings);

                    while (rdr.Read())
                    {
                        if (rdr.NodeType == XmlNodeType.Element)
                        {
                            IXmlSchemaInfo info = rdr.SchemaInfo;
                            if (rdr.LocalName == "HelpTag")
                                tag = rdr.ReadElementString();
                            else if (rdr.LocalName == "HelpText")
                            {
                                text = rdr.ReadElementString();
                                m_help_text[tag] = CleanText(text);
                            }
                        }
                    }
                }
            }
            catch (Exception)
            {
                ret = false;
                m_help_text.Clear();
            }

            return ret;
        }

#endregion

        #region private member functions
        private static void MyValidate(object sender, ValidationEventArgs args)
        {
            // TODO - add logger and log validation errors to logger
        }

        private static void MyXmlValidate(object sender, ValidationEventArgs args)
        {
            // TODO - add logger and log validation errors to logger
        }

        private string CleanText(string s)
        {
            return s.Trim();
        }

        private string FindDataFileName(string name)
        {
            Assembly a = GetType().Assembly;
            string basepath = Path.GetDirectoryName(a.Location);

#if !DEBUG
            basepath = Path.GetDirectoryName(basepath) ;
            basepath = Path.Combine(basepath, "data") ;
#endif
            return Path.Combine(basepath, name);
        }
        #endregion
    }
}
