//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111S1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Bwg.Logging ;

namespace Bwg.Controls
{
    /// <summary>
    /// This window is a control that can log the messages coming from a memory
    /// message sink.
    /// </summary>
    public partial class LoggerWindow : ListView
    {
        #region private member variables
        private MemorySink m_sink;
        #endregion

        #region private types
        private delegate void AddMessageDelegate(UserMessage m);
        #endregion

        #region constructor
        /// <summary>
        /// Create the logger windows
        /// </summary>
        public LoggerWindow()
        {
            m_sink = null;
            InitializeComponent();
            FullRowSelect = true;

            ContextMenu = new ContextMenu();
            ContextMenu.MenuItems.Add(new MenuItem("Copy", new EventHandler(CopyEventHandler)));
            ContextMenu.MenuItems.Add(new MenuItem("Select All", new EventHandler(SelectAllEventHandler)));
            ContextMenu.MenuItems.Add(new MenuItem("Save...", new EventHandler(SaveEventHandler)));
            ContextMenu.MenuItems.Add(new MenuItem("Clear", new EventHandler(ClearEventHandler)));
        }
        #endregion

        #region public member functions
        /// <summary>
        /// Setup the columns for the logger window
        /// </summary>
        public void SetupColumns()
        {
            View = View.Details;

            Columns.Clear();
            ColumnHeader h;

            h = new ColumnHeader();
            h.Text = "Date/Time";
            Columns.Add(h);

            h = new ColumnHeader();
            h.Text = "Type";
            Columns.Add(h);

            h = new ColumnHeader();
            h.Text = "Message";
            Columns.Add(h);

            Columns[0].Width = 140;
            Columns[1].Width = 78;
            OnSizeChanged(null);
        }

        /// <summary>
        /// This method is called when the window changes size.  This method resizes the columns
        /// in the window, keeping the first two columns the same and varying the size of the message
        /// column.
        /// </summary>
        /// <param name="e"></param>
        protected override void  OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);

            if (Columns.Count >= 3)
            {
                Columns[2].Width = Width - Columns[0].Width - Columns[1].Width - 12;
            }
        }
        #endregion

        #region public properties
        /// <summary>
        /// This property sets the message sink for the logger windows
        /// </summary>
        public MemorySink Sink
        {
            get
            {
                return m_sink;
            }
            set
            {
                m_sink = value;
                if (m_sink != null)
                    m_sink.MessageAdded += new EventHandler<MessageAddedArgs>(MessageAdded);
            }
        }
        #endregion

        #region private member functions
        private string GetItemText(ListViewItem item)
        {
            StringBuilder b = new StringBuilder(256) ;

            for (int i = 0; i < item.SubItems.Count; i++)
            {
                if (b.Length > 0)
                    b.Append(' ');

                b.Append(item.SubItems[i].Text);
            }
            b.Append("\r\n");

            return b.ToString();
        }

        private string GetLogText()
        {
            string text = "";

            if (SelectedItems.Count == 0)
            {
                foreach (ListViewItem item in Items)
                    text += GetItemText(item);
            }
            else
            {
                foreach (ListViewItem item in SelectedItems)
                    text += GetItemText(item);
            }

            return text;
        }

        private void SelectAllEventHandler(object sender, EventArgs args)
        {
            foreach (ListViewItem item in Items)
                item.Selected = true;
        }

        private void CopyEventHandler(object sender, EventArgs args)
        {
            string text = GetLogText() ;
            if (text.Length > 0)
                Clipboard.SetText(text);
        }

        private void SaveEventHandler(object sender, EventArgs args)
        {
            SaveFileDialog dialog = new SaveFileDialog();

            dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            dialog.Filter = "txt files (*.txt)|*.txt|All files (*.*)|*.*";
            dialog.FilterIndex = 1;
            dialog.RestoreDirectory = true;
            if (dialog.ShowDialog() != DialogResult.OK)
                return;

            try
            {
                FileStream strm = new FileStream(dialog.FileName, FileMode.Create, FileAccess.Write);
                TextWriter writer = new StreamWriter(strm);
                writer.Write(GetLogText());
                writer.Flush();
                writer.Dispose();
                strm.Dispose();
            }
            catch (Exception ex)
            {
                string str = "cannot write log to file - " + ex.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void MessageAdded(object sender, MessageAddedArgs args)
        {
            AddMessageToWindow(args.AddedMessage);
        }

        private void AddMessageToWindow(UserMessage m)
        {
            if (InvokeRequired)
            {
                AddMessageDelegate dlg = new AddMessageDelegate(AddMessageToWindow);
                object[] args = new object[1];
                args[0] = m;
                Invoke(dlg, args);
            }
            else
            {
                ListViewItem item = new ListViewItem(m.When.ToString());
                string str = m.MType.ToString();
                if (m.MType == UserMessage.Category.Debug)
                    str += "-" + m.Level.ToString();
                item.SubItems.Add(str);
                item.SubItems.Add(m.Text);
                Items.Add(item);
            }
        }

        private void ClearEventHandler(object sender, EventArgs args)
        {
            Items.Clear();
        }

        #endregion
    }
}
