//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Threading;
using Bwg.DiskImage;
using Bwg.Logging;
using Bwg.Scsi;
using Bwg.Burner;
using Bwg.Controls;
using Bwg.Utils;

namespace Bwg.CopyDisk
{
    public partial class CopyDiskForm : BwgBurnBaseForm
    {
        #region private types
        private enum DiskSwapStatus
        {
            Waiting,
            Swapped,
            Canceled
        } ;

        private enum CopySteps
        {
            DiskToImage,                // Copy the contents of a disk to an image file,
                                        // Uses m_drive_name to find the drive
                                        // Uses m_image_name for the image name

            ImageToDisk,                // Copy the contents of an image to a disk
                                        // Uses m_second_drive_name to find the drive
                                        // Uses m_image_name for the image name

            DiskToDisk,                 // Copy the contents of a disk to another disk
                                        // Uses m_drive_name to find the source drive
                                        // Uses m_second_drive_name to find the dest drive

            Swap,                       // Prompts the user to swap disks
                                        // Uses m_drive_name for the drive

            RemoveImage,                // Deletes the image file
                                        // Uses m_image_name for the image to delete
        } ;

        private enum AskYesNoState
        {
            Yes,
            No,
            Waiting
        } ;
        #endregion

        #region private member vars
        private Drive.BurnType m_burn_type;
        private string m_temp_dir;
        private System.Windows.Forms.Timer m_timer ;
        private DateTime m_start;
        private IList<CopySteps> m_copy_tasks;
        private string m_image_name;
        private string m_drive_name;
        private string m_second_drive_name;
        private Thread m_burner_thread;
        private AskYesNoState m_yes_no_state;
        private DiskSwapStatus m_swapped_state;

        #endregion

        private delegate void TimerTickDelegate();
        private delegate void DoneDelegate(DiskOperationError status);
        private delegate void ProgressHandlerDelegate(object sender, ProgressChangedEventArgs args);
        private delegate AskYesNoState AskYesNoDelegate(string msg);
        private delegate void PromptDiskSwapDelegate(string drive);

        public CopyDiskForm()
        {
            InitializeComponent();

            LoggerWindow.SetupColumns();
            LoggerWindow.Sink = m_sink;

            m_logger.LogMessage(new UserMessage(UserMessage.Category.Info, 0, "Disk Copy Version " + GetVersion()));

            // Set the logger and populate the dropdown with the drives found
            SourceDisk.Logger = m_logger;
            SourceDisk.SelectedIndexChanged += new EventHandler(SourceDiskSelector_IndexChanged);
            SourceDisk.InitializeDeviceList(true);

            DestinationDisk.Logger = m_logger;
            DestinationDisk.SelectedIndexChanged += new EventHandler(DestinationDiskSelector_IndexChanged);
            DestinationDisk.InitializeDeviceList(true);

            m_timer = new System.Windows.Forms.Timer();

            m_burn_type = (Drive.BurnType)GetStartupInteger(PropertyNameCopyDiskBurnType, (int)Drive.BurnType.DontCare);
            m_temp_dir = GetStartupString(PropertyNameTemporaryDirectory, Path.GetTempPath());
        }

        private void SourceDiskSelector_IndexChanged(object sender, EventArgs args)
        {
            Drive srcdrv;

            if (SourceDisk.IsCreateImageSelected)
            {
                SourceDesc.Text = "Source Image File";
            }
            else
            {
                Device dev = new Device(m_logger);
                if (!dev.Open(SourceDisk.SelectedDeviceName))
                {
                    SourceDesc.Text = "Error opening drive";
                    return;
                }

                srcdrv = new Drive(dev);
                DiskOperationError status = srcdrv.Initialize();
                if (status != null)
                {
                    SourceDesc.Text = status.Message;
                    srcdrv.Dispose();
                    srcdrv = null;
                    return;
                }

                SourceDesc.Text = srcdrv.StatusString;

                srcdrv.Dispose();
                srcdrv = null;
            }
        }

        private void DestinationDiskSelector_IndexChanged(object sender, EventArgs args)
        {
            Drive destdrv;

            if (DestinationDisk.IsCreateImageSelected)
            {
                DestinationDesc.Text = "Destination Image File";
            }
            else
            {
                Device dev = new Device(m_logger);
                if (!dev.Open(DestinationDisk.SelectedDeviceName))
                {
                    DestinationDesc.Text = "Error opening drive";
                    return;
                }

                destdrv = new Drive(dev);
                DiskOperationError status = destdrv.Initialize();
                if (status != null)
                {
                    DestinationDesc.Text = status.Message;
                    destdrv.Dispose();
                    destdrv = null;
                    return;
                }

                DestinationDesc.Text = destdrv.StatusString;
                destdrv.Dispose();
                destdrv = null;
            }
        }

        private void CopyButton_Click(object sender, EventArgs e)
        {
            DialogResult res;

            m_copy_tasks = null ;

            m_drive_name = SourceDisk.SelectedDeviceName;
            m_second_drive_name = DestinationDisk.SelectedDeviceName;

            if (SourceDisk.SelectedDeviceName == DestinationDisk.SelectedDeviceName)
            {
                if (SourceDisk.IsCreateImageSelected)
                {
                    MessageBox.Show("The source and the destination are both images.  This is not a valid selection",
                                    "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                //
                // Do a two stage copy with an image in between
                //
                m_image_name = FindTempImageName();

                m_copy_tasks = new List<CopySteps>();
                m_copy_tasks.Add(CopySteps.DiskToImage);
                m_copy_tasks.Add(CopySteps.Swap);
                m_copy_tasks.Add(CopySteps.ImageToDisk);
            }
            else if (SourceDisk.IsCreateImageSelected)
            {
                OpenFileDialog dialog = new OpenFileDialog();
                dialog.Multiselect = false;
                dialog.Filter = "BwgBurn Image Files (*.bim)|*.bim|ISO Image Files (*.iso)|*.iso|All files (*.*)|*.*";
                dialog.RestoreDirectory = true;
                dialog.FilterIndex = 0;

                if (dialog.ShowDialog() == DialogResult.OK)
                {
                    m_image_name = dialog.FileName;
                    m_copy_tasks = new List<CopySteps>();
                    m_copy_tasks.Add(CopySteps.ImageToDisk);
                }
                else
                {
                    MessageBox.Show("Disk Copy Canceled", "Canceled", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
            else if (DestinationDisk.IsCreateImageSelected)
            {
                SaveFileDialog dialog = new SaveFileDialog();
                dialog.Filter = "BwgBurn Image Files (*.bim)|*.bim|ISO Image Files (*.iso)|*.iso|All files (*.*)|*.*";
                dialog.RestoreDirectory = true;
                dialog.AddExtension = true;
                dialog.CheckPathExists = true;
                dialog.DefaultExt = "iso";
                dialog.FilterIndex = 0;

                res = dialog.ShowDialog();
                if (res == DialogResult.OK)
                {
                    m_image_name = dialog.FileName;
                    m_copy_tasks = new List<CopySteps>();
                    m_copy_tasks.Add(CopySteps.DiskToImage);
                }
                else
                {
                    MessageBox.Show("Disk Copy Canceled", "Canceled", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
            else
            {
                res = MessageBox.Show("Copying directly from one disk to another can be error prone.  Can I use an intermediate image file?",
                                      "Question", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question) ;

                if (res == DialogResult.Yes)
                {
                    m_image_name = FindTempImageName();
                    m_copy_tasks = new List<CopySteps>();
                    m_copy_tasks.Add(CopySteps.DiskToImage) ;
                    m_copy_tasks.Add(CopySteps.ImageToDisk) ;
                    m_copy_tasks.Add(CopySteps.RemoveImage);
                }
                else if (res == DialogResult.No)
                {
                    m_copy_tasks = new List<CopySteps>();
                    m_copy_tasks.Add(CopySteps.DiskToDisk) ;
                }
                else
                {
                    MessageBox.Show("Disk Copy Canceled", "Canceled", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }

            if (m_copy_tasks != null)
            {
                CopyButton.Enabled = false;
                Settings.Enabled = false;
                RunTasks();
            }
        }

        private void Done(DiskOperationError status)
        {
            m_timer.Stop();

            CopyButton.Enabled = true;
            Settings.Enabled = true;

            string str = "";
            string caption = "Error!";
            MessageBoxIcon icon = MessageBoxIcon.Error;

            if (status == null)
            {
                str = "The copy operation completed sucessfully\r\n";
                caption = "Operation Complete";
                icon = MessageBoxIcon.Information;
            }
            else
            {
                str = "The copy operation failed - " + status.Message;
            }
            MessageBox.Show(str, caption, MessageBoxButtons.OK, icon);
        }

        private void RunTasksProc()
        {
            DiskOperationError status;
            object[] args;

            foreach (CopySteps step in m_copy_tasks)
            {
                status = null;

                switch (step)
                {
                    case CopySteps.DiskToDisk:
                        status = CopyDiskToDisk(m_drive_name, m_second_drive_name);
                        break;

                    case CopySteps.DiskToImage:
                        status = CopyDiskToImage(m_drive_name, m_image_name);
                        break;

                    case CopySteps.ImageToDisk:
                        status = CopyImageToDisk(m_image_name, m_second_drive_name);
                        break;

                    case CopySteps.Swap:
                        status = PropmtUserToSwapDisks(m_drive_name);
                        break;

                    case CopySteps.RemoveImage:
                        File.Delete(m_image_name);
                        break;
                }

                if (status != null)
                {
                    args = new object[1];
                    args[0] = status;
                    Invoke(new DoneDelegate(Done), args);
                }
            }

            args = new object[1];
            args[0] = null;
            Invoke(new DoneDelegate(Done), args);
        }

        private void TimerTickHandler(object sender, EventArgs args)
        {
            TimerHandler();
        }

        private void TimerHandler()
        {
            if (InvokeRequired)
            {
                Invoke(new TimerTickDelegate(TimerHandler));
            }
            else
            {
                TimeSpan elapsed = DateTime.Now - m_start;
                Elapsed.Text = "Elapsed Time: " + elapsed.Minutes.ToString() + ":" + elapsed.Seconds.ToString("D2");
            }
        }

        private void RunTasks()
        {
            m_burner_thread = new Thread(new ThreadStart(RunTasksProc));
            m_burner_thread.Start();

            m_start = DateTime.Now;
            m_timer.Interval = 1000;
            m_timer.Tick += new EventHandler(TimerTickHandler);
            m_timer.Tag = this;
            m_timer.Start();
        }

        private string FindTempImageName()
        {
            string ret;

            if (m_temp_dir == Path.GetTempPath())
                ret = Path.GetTempFileName();
            else
            {
                bool loop = true ;
                string name = "AAAAAAAA.AAA";

                ret = m_temp_dir;

                while (loop)
                {
                    string fname = Path.Combine(m_temp_dir, name);
                    if (!File.Exists(fname))
                    {
                        ret = fname;
                        break;
                    }

                    if (name == "ZZZZZZZZ.ZZZ")
                        return string.Empty;

                    for (int pos = name.Length - 1; pos >= 0; pos--)
                    {
                        if (name[pos] != 'Z')
                        {
                            char ch = (char)(name[pos] + 1);

                            if (pos == name.Length - 1)
                                name = name.Substring(0, pos) + ch;
                            else if (pos == 0)
                                name = ch + name.Substring(1);
                            else
                                name = name.Substring(0, pos) + ch + name.Substring(pos + 1);

                            break;
                        }
                    }
                }
            }

            return ret;
        }

        private void DisplayDiskSwapPrompt(string diskname)
        {
            if (InvokeRequired)
            {
                object[] args = new object[1];
                args[0] = diskname;
                Invoke(new PromptDiskSwapDelegate(DisplayDiskSwapPrompt), args);
            }
            else
            {
                string str = "Please load the blank disk to be burned in the drive '" + diskname + "'." ;
                DialogResult r = MessageBox.Show(str, "Load Disk", MessageBoxButtons.OKCancel, MessageBoxIcon.Question);
                if (r == DialogResult.Cancel)
                    m_swapped_state = DiskSwapStatus.Canceled;
                else
                    m_swapped_state = DiskSwapStatus.Swapped;
            }
        }

        private DiskOperationError PropmtUserToSwapDisks(string diskname)
        {
            m_swapped_state = DiskSwapStatus.Waiting;
            DisplayDiskSwapPrompt(diskname);
            while (m_swapped_state == DiskSwapStatus.Waiting)
                Thread.Sleep(100);

            if (m_swapped_state == DiskSwapStatus.Canceled)
                return null;

            return null;
        }

        private DiskOperationError CopyDiskToImage(string disksrc, string imagedest)
        {
            DiskOperationError status;

            Device dev = new Device(m_logger);
            if (!dev.Open(disksrc))
                return new DeviceIoctlError("CreateFile", dev.LastError) ;

            Drive drv = new Drive(dev);
            status = drv.Initialize();
            if (status != null)
                return status;

            OsFileWriter wtr = new OsFileWriter();
            if (!wtr.Open(imagedest, OsFileWriter.OpenType.Create))
                return new DeviceIoctlError("Open", (uint)wtr.LastError);

            status = CopyDiskToImage(drv, wtr);

            wtr.Close();
            drv.Dispose();

            return status;
        }

        private DiskOperationError CopyImageToDisk(string imagesrc, string diskdest)
        {
            DiskOperationError status;

            Device dev = new Device(m_logger);
            if (!dev.Open(diskdest))
                return new DeviceIoctlError("CreateFile", dev.LastError);

            Drive drv = new Drive(dev);
            status = drv.Initialize();
            if (status != null)
                return status;

            OsFileReader rdr = new OsFileReader();
            if (!rdr.Open(imagesrc))
                return new DeviceIoctlError("Open", (uint)rdr.LastError);

            status = CopyImageToDisk(rdr, drv);

            rdr.Close();
            drv.Dispose();

            return status;
        }

        private DiskOperationError CopyDiskToDisk(string src, string dest)
        {
            DiskOperationError status;

            Device srcdev = new Device(m_logger);
            if (!srcdev.Open(src))
                return new DeviceIoctlError("CreateFile", srcdev.LastError);

            Drive srcdrv = new Drive(srcdev);
            status = srcdrv.Initialize();
            if (status != null)
                return status;

            Device destdev = new Device(m_logger);
            if (!destdev.Open(dest))
            {
                srcdrv.Dispose();
                return new DeviceIoctlError("CreateFile", destdev.LastError);
            }

            Drive destdrv = new Drive(destdev);
            status = destdrv.Initialize();
            if (status != null)
            {
                destdev.Dispose();
                srcdrv.Dispose();
                return status;
            }

            status = CopyDiskToDisk(srcdrv, destdrv);
            srcdrv.Dispose();
            destdrv.Dispose();

            return status;
        }

        private DiskOperationError CopyDiskToDisk(Drive src, Drive dest)
        {
            DiskOperationError status;

            status = CheckErase(dest);
            if (status != null)
                return status;

            DiskContents image = new DriveDiskContents(m_logger, src);
            image.ProgressChanged += new ProgressChangedEventHandler(BurnProgressChanged);
            status = image.BurnDisk(dest, m_burn_type);
            image.ProgressChanged -= new ProgressChangedEventHandler(BurnProgressChanged);

            return status;
        }

        private DiskOperationError CopyDiskToImage(Drive src, OsFileWriter dest)
        {
            DiskContents image = new DriveDiskContents(m_logger, src);
            image.ProgressChanged += new ProgressChangedEventHandler(BurnProgressChanged);
            DiskOperationError status = image.WriteFile(dest);
            image.ProgressChanged -= new ProgressChangedEventHandler(BurnProgressChanged);

            return status;
        }

        private DiskOperationError CopyImageToDisk(OsFileReader src, Drive dest)
        {
            DiskOperationError status;

            status = CheckErase(dest);
            if (status != null)
                return status;

            DiskContents image = new FileDiskContents(m_logger, src);
            image.ProgressChanged += new ProgressChangedEventHandler(BurnProgressChanged);
            status = image.BurnDisk(dest, Drive.BurnType.Track);
            image.ProgressChanged -= new ProgressChangedEventHandler(BurnProgressChanged);

            return status;
        }

        private DiskOperationError CheckErase(Drive drv)
        {
            DiscInformation info;
            Device.CommandStatus st;
            DiskOperationError status = null;

            st = drv.Device.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return Drive.CreateErrorObject(st, drv.Device);

            if (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
                return null;

            if (drv.IsEraseSupported)
            {
                if (AskYesNo("The destination disk is not blank, is it ok to erase this disk?") == AskYesNoState.No)
                {
                    status = new BurnerError("Cannot copy to destination disk, it is not empty");
                }
                else
                {
                    status = drv.Erase(Drive.EraseType.Minimal);
                }
            }
            else
            {
                status = new BurnerError("Cannot copy to destination disk, it is not blank and not erasable");
            }

            return status;
        }

        AskYesNoState AskYesNo(string msg)
        {
            AskYesNoState st = AskYesNoState.Waiting;
            if (InvokeRequired)
            {
                object[] args = new object[1];
                args[0] = msg;
                m_yes_no_state = AskYesNoState.Waiting;
                Invoke(new AskYesNoDelegate(AskYesNo), args);

                while (m_yes_no_state == AskYesNoState.Waiting) Thread.Sleep(250);

                st =  m_yes_no_state;
            }
            else
            {
                DialogResult res;

                res = MessageBox.Show(msg, "Question", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (res == DialogResult.Yes)
                    m_yes_no_state = AskYesNoState.Yes;
                else
                    m_yes_no_state = AskYesNoState.No;
            }

            return st;
        }

        private void BurnProgressChanged(object sender, ProgressChangedEventArgs args)
        {
            if (InvokeRequired)
            {
                object[] param = new object[2];
                param[0] = sender;
                param[1] = args;
                Invoke(new ProgressHandlerDelegate(BurnProgressChanged), param);
            }
            else
            {
                ProgressBar.Value = args.ProgressPercentage;
                if (args.UserState is ProgressText)
                {
                    ProgressText pt = (ProgressText)args.UserState;
                    ProgressBar.Text = pt.LongName;
                    TopLevelControl.Text = pt.ShortName + " ( " + args.ProgressPercentage.ToString() + "% )";
                }
                else if (args.UserState is string)
                {
                    ProgressBar.Text = (string)args.UserState;
                }
            }
        }

        private void Settings_Click(object sender, EventArgs e)
        {
            bool dbgflag, wrnflag ;

            if (m_logger.HasSink(UserMessage.Category.Debug))
                dbgflag = true ;
            else
                dbgflag = false ;

            if (m_logger.HasSink(UserMessage.Category.Warning))
                wrnflag = true ;
            else
                wrnflag = false ;

            SpeedInfo[] speeds = null;
            if (!DestinationDisk.IsCreateImageSelected)
            {
                Device dev = new Device(m_logger);
                if (dev.Open(DestinationDisk.SelectedDeviceName))
                {
                    Drive drv = new Drive(dev);
                    if (drv.Initialize() == null)
                    {
                        if (drv.GetWriteSpeeds(out speeds) != null)
                            speeds = null;
                    }
                    drv.Dispose();
                }
                dev.Dispose();
            }

            CopySettings form = new CopySettings(m_temp_dir, m_logger.GetLevel(UserMessage.Category.Debug), dbgflag, wrnflag, m_burn_type, speeds);
            if (form.ShowDialog() == DialogResult.OK)
            {
                if (form.DebugMessagesFlag)
                {
                    m_logger.SetSink(UserMessage.Category.Debug, m_sink);
                    m_logger.SetLevel(UserMessage.Category.Debug, form.DebugLevel);
                }
                else
                    m_logger.RemoveSink(UserMessage.Category.Debug);

                if (form.WarningMessagesFlag)
                    m_logger.SetSink(UserMessage.Category.Warning, m_sink);
                else
                    m_logger.RemoveSink(UserMessage.Category.Warning);

                SaveLoggerSettings();

                m_temp_dir = form.TempDir;
                SetStartupString(PropertyNameTemporaryDirectory, m_temp_dir);

                m_burn_type = form.BurnType;
                SetStartupInteger(PropertyNameCopyDiskBurnType, (int)m_burn_type);
            }
        }
    }
}