//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;
using Bwg.Audio;
using Bwg.Burner;
using Bwg.Scsi;

namespace Bwg.CreateAudioDisk
{
    /// <summary>
    /// This control is a
    /// </summary>
    public partial class SongList : ListView
    {
        #region public data members
        /// <summary>
        /// This is the audio manager that this control uses to determine if a file of
        /// a given type is supported.
        /// </summary>
        public AudioManager Manager;

        #endregion

        #region private data members
        /// <summary>
        /// This is the disk image we are building
        /// </summary>
        private DiskBurnImage m_disk;

        /// <summary>
        /// The form indiating we are processing a song (scanning)
        /// </summary>
        ProcessingSongForm m_processing_form;
        #endregion

        #region constructors
        /// <summary>
        /// This method creates the song list control
        /// </summary>
        public SongList()
        {
            InitializeComponent();
            CreateColumns();

            View = View.Details;
            FullRowSelect = true;
            AllowDrop = true;
        }
        #endregion

        #region public properties
        public DiskBurnImage Disk
        {
            get
            {
                return m_disk;
            }
            set
            {
                m_disk = value;
                if (m_disk != null)
                    UpdateDisplay();
            }
        }
        #endregion

        #region public methods

        /// <summary>
        /// This method moves the given list view item up one spot
        /// in the list.
        /// </summary>
        /// <param name="item"></param>
        public void MoveUp(ListViewItem item)
        {
            if (item.Index == 0)
                return;

            m_disk.MoveTrackUp((TrackBurnImage)item.Tag);
            UpdateDisplay();
        }

        public void MoveDown(ListViewItem item)
        {
            if (item.Index == Items.Count - 1)
                return;

            m_disk.MoveTrackDown((TrackBurnImage)item.Tag);
            UpdateDisplay();
        }

        public void Delete(ListViewItem item)
        {
            m_disk.RemoveTrack((TrackBurnImage)item.Tag);
            UpdateDisplay();
        }

        public void AddSongs(string[] names)
        {
            m_processing_form = new ProcessingSongForm();
            m_processing_form.Show(this);
            foreach (string fname in names)
            {
                m_processing_form.SongTitle.Text = Path.GetFileName(fname);
                Application.DoEvents();

                if (File.Exists(fname))
                {
                    AddSong(fname);
                }
                else
                {
                    string msg = "The song '" + fname + "' does not exist";
                    MessageBox.Show(msg, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
            m_processing_form.Dispose();
            m_processing_form = null;
        }

        /// <summary>
        /// This method adds a song to the song list
        /// </summary>
        /// <param name="fullpath"></param>
        public void AddSong(string fullpath)
        {
            short lmin, lmax, rmin, rmax;
            AudioReader reader = Manager.GetReader(fullpath);
            if (reader != null)
            {
                try
                {
                    // This is the first operation we should do
                    reader.ScanProgressChanged += new ProgressChangedEventHandler(ScanChanged);
                    reader.ScanInputFile(out lmin, out lmax, out rmin, out rmax);
                }
                catch (Exception ex)
                {
                    reader.ScanProgressChanged -= new ProgressChangedEventHandler(ScanChanged);
                    MessageBox.Show(ex.Message, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }
                reader.ScanProgressChanged -= new ProgressChangedEventHandler(ScanChanged);

                AudioDataSource src = new AudioDataSource(reader);
                TrackBurnImage t = new TrackBurnImage(TrackBurnImage.TrackType.Audio, src);
                m_disk.AddTrack(t);

                AddTrackToDisplay(t);
            }

        }
        #endregion

        #region private methods

        private void AddTrackToDisplay(TrackBurnImage t)
        {
            // File name
            string name = Path.GetFileName(t.Name);
            ListViewItem item = new ListViewItem(name);
            item.Tag = t;

            // File type
            AudioDataSource src = (AudioDataSource)t.Data;
            item.SubItems.Add(src.FileTypeString);

            // Start time (we display the start time of the actual audio, not of the pregap
            MinuteSecondFrame start = t.StartMSF + t.PreTrackGap;
            item.SubItems.Add(start.ToString("M:S.F"));

            // Length
            MinuteSecondFrame tlen = new MinuteSecondFrame(t.TotalLengthSectors);
            item.SubItems.Add(tlen.ToString("M:S.F"));

            // Gap
            tlen = new MinuteSecondFrame(t.PreTrackGap);
            item.SubItems.Add(tlen.ToString("M:S.F"));

            // min/max values
            string values = "?/?/?/?" ;
            item.SubItems.Add(values);

            // Add the item to the list
            Items.Add(item);
        }

        private void UpdateDisplay()
        {
            Items.Clear();

            foreach (TrackBurnImage t in m_disk)
                AddTrackToDisplay(t);
        }

        private void ScanChanged(object sender, ProgressChangedEventArgs args)
        {
            if (m_processing_form != null)
            {
                m_processing_form.ScanProgressBar.Value = args.ProgressPercentage;
            }
        }

        private void CreateColumns()
        {
            Columns.Add("File");
            Columns.Add("Type");
            Columns.Add("Start");
            Columns.Add("Length");
            Columns.Add("Gap");
        }

        /// <summary>
        /// Called when the the size of the control changes
        /// </summary>
        /// <param name="e"></param>
        protected override void  OnSizeChanged(EventArgs e)
        {
            Columns[0].Width = Width * 5 / 10;
            Columns[1].Width = Width * 2 / 10;
            Columns[2].Width = Width * 1 / 10;
            Columns[3].Width = Width * 1 / 10;
            Columns[4].Width = Width - Columns[0].Width - Columns[1].Width - 4;
        }
        #endregion

        #region protected event handlers
        /// <summary>
        /// This method is called when a drag and drop operation causes something to be 
        /// dropped over this window.
        /// </summary>
        /// <param name="drgevent">the contents dropped on this window</param>
        protected override void OnDragDrop(DragEventArgs drgevent)
        {
            base.OnDragDrop(drgevent);
            object obj = drgevent.Data.GetData("FileNameW", true);
            AddSongs((string[])obj);
        }

        /// <summary>
        /// This method is called when a drag and drop operation is in progress and the
        /// contents is dragged over this window.
        /// </summary>
        /// <param name="drgevent">the object describing what is being dragged over this window</param>
        protected override void OnDragOver(DragEventArgs drgevent)
        {
            base.OnDragOver(drgevent);

            if (drgevent.Data.GetDataPresent("FileNameW", true))
            {
                bool dropok = false;

                if (Manager == null)
                    dropok = true;
                else
                {
                    object obj = drgevent.Data.GetData("FileNameW", true);
                    string[] namearray = (string[])obj;

                    foreach (string fname in namearray)
                    {
                        string ext = Path.GetExtension(fname);
                        if (ext.Length == 0)
                            continue;

                        if (File.Exists(fname) && Manager.AudioDriverExists(ext))
                        {
                            dropok = true;
                            break;
                        }
                    }
                }

                if (dropok)
                    drgevent.Effect = DragDropEffects.Copy;
                else
                    drgevent.Effect = DragDropEffects.None;
            }
        }
        #endregion
    }
}
