//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Threading ;
using System.Diagnostics;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using Bwg.Logging;
using Bwg.Isofs;
using Bwg.Controls;
using Bwg.Scsi;
using Bwg.Burner;
using Bwg.Utils ;

namespace Bwg.CreateDataDisk
{
    /// <summary>
    /// The main window for the create data disk application
    /// </summary>
    public partial class CreateDataDiskForm : BwgBurnBaseForm
    {
        #region public variables

        /// <summary>
        /// The initial project for this application.  This is only applied at start time
        /// </summary>
        public string InitialProject;

        /// <summary>
        /// If true, hide the image settings button.  This only applies at start time
        /// </summary>
        public bool HideImageSettings;

        #endregion

        #region private member variables
        [FlagsAttribute]
        private enum BurnRequestOptions
        {
            Erase = 0x01,                       // Erase the disk before burning
            MultiSession = 0x02,                // Leave the disk open for additional sessions
        } ;

        /// <summary>
        /// The image to burn to a disk.
        /// </summary>
        private DiskImageContents m_image;

        /// <summary>
        /// This object is responsible for creating the disk image to be recored onto the disk.  This
        /// object also maintains information about the size of the current image.
        /// </summary>
        private DiskImageCreator m_creator;

        private DiskImageElement.NameType m_nametype;
        private char m_drive_letter;
        private Device m_device;
        private Drive m_drive;
        private System.Windows.Forms.Timer m_timer;
        private DateTime m_start;
        private Thread m_burner_thread;
        private bool m_multi_session;
        private Drive.BurnType m_burn_type;
        private SpeedInfo m_burn_speed;
        private ShellIconMgr m_icon_mgr;
        private bool m_remaining;
        #endregion

        #region private types
        private delegate void TimerTickDelegate();
        private delegate void DoneDelegate(DiskOperationError status);
        private delegate void ProgressChangedDelegate(object sender, ProgressChangedEventArgs args);
        #endregion

        #region constructor
        /// <summary>
        /// Create the data disk form
        /// </summary>
        public CreateDataDiskForm()
        {
            HideImageSettings = false;
            InitialProject = string.Empty;
            m_nametype = DiskImageElement.NameType.Local;

            m_icon_mgr = new ShellIconMgr();
            m_remaining = false;

            InitializeComponent();
            
            LoggerWindow.Sink = m_sink;
            LoggerWindow.SetupColumns();

            m_logger.LogMessage(new UserMessage(UserMessage.Category.Info, 0, "Data Disk Creator Version " + GetVersion()));

            DeviceSelector.SelectedIndexChanged += new EventHandler(DeviceSelector_SelectedIndexChanged);
            DeviceSelector.Logger = m_logger;
            DeviceSelector.InitializeDeviceList(true);

            m_image = new DiskImageContents(m_logger);
            m_creator = new DiskImageCreator(m_logger, m_image);

            DiskImageTree.NameType = m_nametype;
            DiskImageTree.IconManager = m_icon_mgr;
            DiskImageTree.ImageCreator = m_image;
            DiskImageTree.AfterSelect += new TreeViewEventHandler(DiskImageTree_AfterSelect);

            DiskImageList.NameType = m_nametype;
            DiskImageList.IconManager = m_icon_mgr;
            DiskImageList.ImageCreator = m_image;
            DiskImageList.Directory = m_image.RootDirectory;

            ComputerTree.IconManager = m_icon_mgr;
            ComputerList.IconManager = m_icon_mgr;

            ComputerTree.AfterSelect += new TreeViewEventHandler(ComputerTree_AfterSelect);
            ComputerTree.Populate();

            ComputerList.DoubleClick += new EventHandler(ComputerList_DoubleClick);

            m_image.SizeChanged += new EventHandler<EventArgs>(CreatorChanged);

            m_image.BeforeDirectoryAdd += new EventHandler<EventArgs>(BeforeDirectoryAdd);
            m_image.AfterDirectoryAdd += new EventHandler<EventArgs>(AfterDirectoryAdd);
            m_image.AddingDirectory += new EventHandler<AddingDirectoryEventArgs>(AddingDirectoryToImage);
            m_image.AddingFile += new EventHandler<AddingFileEventArgs>(AddingFileToImage);

            DiskImageTree.DisableControls.Add(ComputerTree);
            DiskImageTree.DisableControls.Add(ComputerList);
            DiskImageTree.DisableControls.Add(DriveSettings);
            DiskImageTree.DisableControls.Add(ImageSettings);
            DiskImageTree.DisableControls.Add(BurnDiskButton);
            DiskImageTree.DisableControls.Add(DeviceSelector);

            DiskImageList.DisableControls.Add(ComputerTree);
            DiskImageList.DisableControls.Add(ComputerList);
            DiskImageList.DisableControls.Add(DriveSettings);
            DiskImageList.DisableControls.Add(ImageSettings);
            DiskImageList.DisableControls.Add(BurnDiskButton);
            DiskImageList.DisableControls.Add(DeviceSelector);

            BurnProgress.Value = 0;
            Operation.Text = "Ready";

            m_timer = new System.Windows.Forms.Timer();

            m_burn_type = (Drive.BurnType)GetStartupInteger(PropertyNameCreateDataDiskBurnType, (int)Drive.BurnType.DontCare);
            m_multi_session = GetStartupBool(PropertyNameCreateDataDiskMultiSession, false);

            m_burner_thread = null;

            AllViewItemsOff();
            LocalNamesMenuItem.Checked = true;
        }
        #endregion

        #region public methods

        /// <summary>
        /// Reinitialize the window based on the command line arguments
        /// </summary>
        public void ReInitialize()
        {
            if (HideImageSettings)
                ImageSettings.Visible = false;
            else
                ImageSettings.Visible = true;

            if (InitialProject != string.Empty)
                LoadImageFromFile(InitialProject);
        }

        /// <summary>
        /// Add an extension to the list of files to display
        /// </summary>
        /// <param name="ext"></param>
        public void AddExtension(string ext)
        {
            ComputerList.AddExt(ext);
            m_creator.AddExt(ext);
        }

        #endregion

        #region private methods

        private void DiskImageTree_DragDropComplete(object sender, EventArgs args)
        {
            DiskImageDirectory dir = DiskImageList.Directory;
            DiskImageList.Directory = dir;
        }

        private void ComputerList_DoubleClick(object sender, EventArgs args)
        {
            ListViewItem item = ComputerList.SelectedItems[0];
            if (item.Tag is string)
            {
                string dir = (string)item.Tag;
                if (Directory.Exists(dir))
                {
                    TreeNode n = ComputerTree.SelectedNode;
                    if (!n.IsExpanded)
                        n.Expand();

                    TreeNode child = n.FirstNode;
                    while (child != null)
                    {
                        if (child.Tag is string && (string)child.Tag == dir)
                        {
                            ComputerTree.SelectedNode = child;
                            return;
                        }
                        child = child.NextNode;
                    }
                }
            }
        }

        /// <summary>
        /// This method is called when the main form is closing.  This method determines if this
        /// is ok and tells the form if it is not.
        /// </summary>
        /// <param name="e">the event args, used to return the close ok status</param>
        protected override void OnClosing(CancelEventArgs e)
        {
            if (m_burner_thread != null)
            {
                DialogResult res = MessageBox.Show("A burn operation is in progress. Do you really want to quit this program and abort the burn?",
                                                    "Abort?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (res == DialogResult.No)
                {
                    e.Cancel = true;
                    return; 
                }

                WaitingOnDrive form = new WaitingOnDrive();
                form.Show(this);
                Application.DoEvents();

                m_burner_thread.Abort();
                while (m_burner_thread.IsAlive)     // Wait for thread to die
                    Application.DoEvents();

                m_drive.WaitForDeviceReady(false);
                m_drive.Device.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
                m_drive.Device.StartStopUnit(false, Device.PowerControl.NoChange, Device.StartState.EjectDisc);

                form.Close();
            }
            base.OnClosing(e);
        }

        /// <summary>
        /// This method is called when the form changes size
        /// </summary>
        /// <param name="e">not used</param>
        protected override void OnSizeChanged(EventArgs e)
        {
            if (Width > 800)
                DirectoryAdd.Width = Width - 400;
            else
                DirectoryAdd.Width = 300;

            base.OnSizeChanged(e);
        }

        private void BeforeDirectoryAdd(object sender, EventArgs args)
        {
            Operation.Text = "Adding Files";
        }

        private void AfterDirectoryAdd(object sender, EventArgs args)
        {
            Operation.Text = "Ready";
            DirectoryAdd.Text = "";
            FileAdd.Text = "";
        }

        private void AddingDirectoryToImage(object sender, AddingDirectoryEventArgs args)
        {
            DirectoryAdd.Text = args.Dir.FileSystemLocation;
            Application.DoEvents();
        }

        private void AddingFileToImage(object sender, AddingFileEventArgs args)
        {
            FileAdd.Text = args.LocalFileName;
            Application.DoEvents();
        }

        private void CreatorChanged(object sender, EventArgs args)
        {
            long myused;
            long capacity;

            if (m_image == null)
                myused = 0;
            else
                myused = m_creator.SizeInBlocks;

            if (m_drive != null)
            {
                if (m_remaining)
                    capacity = m_drive.RemainingCapacityInWritableSectors;
                else
                    capacity = m_drive.CapacityInSectors;
            }
            else if (DeviceSelector.IsCreateImageSelected)
            {
                capacity = myused;
            }
            else
            {
                capacity = Drive.D74minDiskCapacity;
            }

            long usedmeg = myused / 512;
            long availmsg = capacity / 512;

            if (m_drive == null)
                DiskSpace.Maximum = (int)Drive.D74minDiskCapacity;
            else
                DiskSpace.Maximum = (int)m_drive.CapacityInSectors;

            string txt ;

            txt = "Used " + usedmeg.ToString() + " Mb of " + availmsg.ToString() + " Mb";
            if (m_remaining)
                txt += " remaining on disk";
            else
                txt += " total on disk";

            DiskSpace.Text = txt;

            if (myused > capacity)
            {
                DiskSpace.Value = DiskSpace.Maximum;
                DiskSpace.BarInnerColorDark = Color.Red;
                DiskSpace.BarInnerColorLight = Color.LightPink;
            }
            else
            {
                DiskSpace.Value = (int)(myused);
                DiskSpace.BarInnerColorLight = Color.FromArgb(192, 255, 255);
                DiskSpace.BarInnerColorDark = Color.Purple;
            }

            BurnProgress.Value = 0;
            DirectoryAdd.Text = string.Empty;
            FileAdd.Text = String.Empty;
        }

        private void DiskImageTree_AfterSelect(object sender, TreeViewEventArgs args)
        {
            if (args.Node != null)
            {
                if (args.Node.Tag is DiskImageDirectory)
                {
                    DiskImageDirectory dir = (DiskImageDirectory)args.Node.Tag ;
                    DiskImageList.Directory = dir ;
                    ImageDirName.Text = dir.GetPath(m_nametype);
                }
            }
        }

        private void ComputerTree_AfterSelect(object sender, TreeViewEventArgs args)
        {
            if (args.Node != null)
            {
                if (args.Node.Tag is string)
                {
                    ComputerList.Path = (string)args.Node.Tag;
                    ComputerDirName.Text = (string)args.Node.Tag;
                }
            }
        }

        private void DeviceSelector_SelectedIndexChanged(object sender, EventArgs e)
        {
            eraseToolStripMenuItem.Enabled = false;

            if (DeviceSelector.SelectedItem == null || DeviceSelector.IsCreateImageSelected)
            {
                if (m_drive != null)
                    m_drive.Dispose();

                if (m_device != null)
                    m_device.Dispose();

                m_drive = null;
                m_device = null;

                if (DeviceSelector.IsCreateImageSelected)
                    m_drive_status.Text = "Create a disk image";
                else
                    m_drive_status.Text = "No Drive Is Selected";

                m_drive_letter = '-' ;

            }
            else
            {
                string dname = (string)DeviceSelector.SelectedItem;
                if (m_drive != null)
                {
                    m_drive.Dispose();
                    m_drive = null;
                }

                if (m_device != null)
                {
                    m_device.Dispose();
                    m_device = null;
                }

                m_drive_letter = dname[0];
                m_device = new Device(m_logger);
                if (!m_device.Open(m_drive_letter))
                {
                    m_device.Dispose();
                    m_device = null;
                }
                else
                {
                    m_drive = new Drive(m_device);
                    DiskOperationError status = m_drive.Initialize();
                    if (status != null)
                    {
                        m_drive.Dispose();
                        m_device.Dispose();
                        m_drive = null;
                        m_device = null;
                    }
                }

                // Set the status string
                if (m_drive == null)
                    m_drive_status.Text = "No Drive Is Selected";
                else
                {
                    eraseToolStripMenuItem.Enabled = m_drive.IsEraseSupported;
                    m_drive_status.Text = m_drive.StatusString;
                }
            }

            string tip;
            if (m_drive != null)
            {
                tip = m_drive.StatusString + "\n";
                if (m_drive.MediaManufacturer.Length > 0)
                    tip += "Media Manufacturer: " + m_drive.MediaManufacturer + "\n";
                if (m_drive.MediaType.Length > 0)
                    tip += "Media Type: " + m_drive.MediaType + "\n";
                if (m_drive.MediaVersion != 0xff)
                    tip += "Media Version: " + m_drive.MediaVersion.ToString() + "\n";

                SpeedInfo[] speeds;
                if (m_drive.GetWriteSpeeds(out speeds) != null || speeds.GetLength(0) == 0)
                    m_burn_speed = null;
                else
                    m_burn_speed = speeds[0];
            }
            else
            {
                tip = "No drive selected";
            }
            DiskImageToolTip.SetToolTip(DeviceSelector, tip);
            CreatorChanged(null, null);
        }

        private void ImageSettings_Click(object sender, EventArgs e)
        {
            ImageSettings settings = (ImageSettings)m_creator.Settings.Clone();
            ImageSettingsForm form = new ImageSettingsForm(settings);
            if (form.ShowDialog() == DialogResult.OK)
            {
                if (m_creator.ContainsFilesSized(2147483648) && settings.IsISO9660)
                {
                    string str = "You have changed the image type to include an ISO 9660 type.  You also have files" ;
                    str += " in the image that exceed 2 Gbytes in size.  These files cannot be written to an ISO 9660" ;
                    str += " image.  Do you want to remove these files from the image? (Yes will remove the files, No" ;
                    str += " will prevent the new settings from taking effect." ;
                    DialogResult r = MessageBox.Show(str, "Invalid Files", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                    if (r == DialogResult.No)
                    {
                        MessageBox.Show("Settings changes ignored", "Files too large", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }

                    int cnt = m_creator.RemoveFilesSized(2147483648);
                    str = cnt.ToString() + " files removed." ;
                    MessageBox.Show(str, "Files Removed", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
                m_creator.Settings = settings;

                Form f = new UpdatingView();
                f.Show();
                Application.DoEvents();
                DiskImageList.NameType = DiskImageList.NameType;
                DiskImageTree.NameType = DiskImageTree.NameType;
                CreatorChanged(null, null);
                UpdateViewMenu();
                f.Close();
            }
        }

        private void TimerTick()
        {
            TimeSpan elapsed = DateTime.Now - m_start;
            DirectoryAdd.Text = "Elapsed Time: " + elapsed.Minutes.ToString() + ":" + elapsed.Seconds.ToString("D2");

            if (m_drive != null && m_drive.IsBufferValid)
            {
                if (DriveBuffer.Text == string.Empty)
                {
                    DriveBuffer.Text = "Buffer Usage";
                    DriveBuffer.AddPercent = true;
                }

                DriveBuffer.Minimum = 0;
                DriveBuffer.Maximum = m_drive.BufferLength;
                int used = m_drive.BufferLength - m_drive.BufferAvailable;
                if (used < m_drive.BufferLength)
                    DriveBuffer.Value = used;
            }
            else
            {
                DriveBuffer.Text = string.Empty;
                DriveBuffer.AddPercent = false;
            }
            Application.DoEvents();
        }

        private static void TimerTickHandler(object sender, EventArgs args)
        {
            System.Windows.Forms.Timer t = (System.Windows.Forms.Timer)sender;
            CreateDataDiskForm form = (CreateDataDiskForm)t.Tag;
            if (form.InvokeRequired)
            {
                form.Invoke(new TimerTickDelegate(form.TimerTick));
            }
            else
                form.TimerTick();
        }

        private void ProgressHandler(object sender, ProgressChangedEventArgs e)
        {
            if (InvokeRequired)
            {
                object[] args = new object[2];
                args[0] = sender;
                args[1] = e;
                Invoke(new ProgressChangedDelegate(ProgressHandler), args);
            }
            else
            {
                if (e.UserState is ProgressText)
                {
                    ProgressText progtext = (ProgressText)e.UserState;
                    BurnProgress.Text = progtext.LongName;
                    TopLevelControl.Text = progtext.ShortName + " ( " + e.ProgressPercentage.ToString() + " % )" ;
                }
                else if (e.UserState is string)
                {
                    BurnProgress.Text = (string)e.UserState;
                    TopLevelControl.Text = BurnProgress.Text;
                }

                BurnProgress.Value = e.ProgressPercentage;
            }
        }

        private void EnableCreateImageControls(bool st)
        {
            DeviceSelector.Enabled = st;
            DiskImageList.Enabled = st;
            DiskImageTree.Enabled = st;
            ComputerTree.Enabled = st;
            ComputerList.Enabled = st;
            ImageSettings.Enabled = st;
            DriveSettings.Enabled = st;
            BurnDiskButton.Enabled = st;
        }

        private void BurnImage(object arg)
        {
            Debug.Assert(arg != null && arg is string);

            string filename = (string)arg;
            DiskOperationError status = null ;

            m_creator.ValidationProgressChanged += new ProgressChangedEventHandler(ProgressHandler);
            if (!m_creator.Validate())
            {
                m_creator.ValidationProgressChanged -= new ProgressChangedEventHandler(ProgressHandler);
                ProgressHandler(null, new ProgressChangedEventArgs(100, null));
                status = new BurnerError("Error validating the disk image, see log window for more details");
            }
            else
            {
                m_creator.CreateProgressChanged += new ProgressChangedEventHandler(ProgressHandler);
                if (!m_creator.CreateImage(filename, 0))
                    status = new BurnerError("Error creating disk image, see log window for more details");
                m_creator.CreateProgressChanged -= new ProgressChangedEventHandler(ProgressHandler);
            }

            object[] args = new object[1];
            args[0] = status;
            Invoke(new DoneDelegate(BurnDone), args);
        }

        private void CreateImage()
        {
            EnableCreateImageControls(false);
            SaveFileDialog dialog = new SaveFileDialog();

            dialog.AddExtension = true;
            dialog.CheckPathExists = true;
            dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            dialog.RestoreDirectory = true;
            dialog.Filter = "ISO Image files (*.iso)|*.iso|All files (*.*)|*.*";
            dialog.FilterIndex = 0 ;
            dialog.OverwritePrompt = true ;
            dialog.ValidateNames = true ;
            DialogResult ret = dialog.ShowDialog();
            if (ret == DialogResult.OK)
            {
                // Start a timer to display elapsed time
                m_start = DateTime.Now;
                m_timer.Interval = 1000;
                m_timer.Tick += new EventHandler(TimerTickHandler);
                m_timer.Tag = this;
                m_timer.Start();

                // Now, create a thread to burn the disk
                m_burner_thread = new Thread(new ParameterizedThreadStart(BurnImage));
                m_burner_thread.Start(dialog.FileName);
            }
            else
            {
                EnableCreateImageControls(true);
            }
        }

        private void BurnDiskButton_Click(object sender, EventArgs e)
        {
            BurnRequestOptions options = BurnRequestOptions.MultiSession ;

            if (DeviceSelector.IsCreateImageSelected)
            {
                CreateImage();
                return;
            }

            if (m_drive == null)
            {
                MessageBox.Show("CD/DVD drive not selected, please select a drive");
                return;
            }

            DiskOperationError status  = m_drive.Initialize();
            if (status != null)
            {
                string str = "Error intializing drive - " + status.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            DiscInformation info;
            Device.CommandStatus st = m_device.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
            {
                MessageBox.Show("Error reading disc information - see the log for more information", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            if (m_drive.IsCDRW && info.DiscStatus != DiscInformation.DiscStatusType.EmptyDisc)
            {
                if (info.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc || !m_drive.IsCDRW)
                {
                    // Propmt the user to format
                    DialogResult res = MessageBox.Show("This disk is not erased.  Do you wish to erase this disk?",
                                                       "Erase Disk?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                    if (res == DialogResult.No)
                    {
                        MessageBox.Show("Please insert a blank disk and press burn again", "Not A Blank Disk", MessageBoxButtons.OK, MessageBoxIcon.Information);
                        return;
                    }
                    options |= BurnRequestOptions.Erase ;
                }
                else if (info.DiscStatus == DiscInformation.DiscStatusType.IncompleteDisc)
                {
                    AskAppendEraseQuestion form = new AskAppendEraseQuestion() ;
                    DialogResult res = form.ShowDialog(this);
                    if (res == DialogResult.No)
                    {
                        options |= BurnRequestOptions.Erase ;
                    }
                    else if (res == DialogResult.Cancel)
                    {
                        MessageBox.Show("Please insert a blank disk and press burn again", "Not A Blank Disk", MessageBoxButtons.OK, MessageBoxIcon.Information);
                        return;
                    }
                }
            }
            else if (m_drive.CurrentProfile == Drive.SupportedProfiles.DVD_PLUS_R && info.DiscStatus == DiscInformation.DiscStatusType.IncompleteDisc)
            {
                string str = "The media in the drive is a DVD+R with a session already recored on the disk.  ";
                str += "While the current session can be appended to the media, most computers are not able to read ";
                str += "the data from any session past the first session.  Are you sure you want to append this data ";
                str += "to the current media?";
                DialogResult res = MessageBox.Show(str, "Additional Sessions?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (res == DialogResult.No)
                {
                    MessageBox.Show("Please insert a blank disk and press burn again", "Not A Blank Disk", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    return;
                }
            }

            if (m_drive.IsBurnSupported)
            {
                EnableCreateImageControls(false);

                // Start a timer to display elapsed time
                m_start = DateTime.Now;
                m_timer.Interval = 1000;
                m_timer.Tick += new EventHandler(TimerTickHandler);
                m_timer.Tag = this;
                m_timer.Start();

                // Now, create a thread to burn the disk
                m_burner_thread = new Thread(new ParameterizedThreadStart(BurnProc));
                m_burner_thread.Start(options);
            }
            else
            {
                string str = "A disk of type '" + m_drive.CurrentProfile.ToString() + "' is in the drive, this type is not ";
                str += "supported by the create data disk utility.";
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);

                m_device.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
            }
        }

        private void BurnDone(DiskOperationError status)
        {
            string msg;
            m_timer.Stop();

            m_burner_thread = null;

            EnableCreateImageControls(true);

            string caption = "Error!";
            MessageBoxIcon localicon ;

            if (status == null)
            {
                msg = "The burn operation completed sucessfully\r\n";
                if (m_drive != null)
                {
                    msg += "Target Speed: " + m_burn_speed + "\r\n";
                    msg += "Actual Speed: " + m_drive.BurnSpeed;
                }
                caption = "Complete";
                localicon = MessageBoxIcon.Information;
            }
            else
            {
                msg = status.Message;
                caption = "Error!";
                localicon = MessageBoxIcon.Error;
            }

            MessageBox.Show(msg, caption, MessageBoxButtons.OK, localicon);
        }

        private void BurnProc(object param)
        {
            DiskOperationError status ;

            Debug.Assert(param is BurnRequestOptions);
            BurnRequestOptions options = (BurnRequestOptions)param;

            object[] args = new object[1];

            //
            // Erase the disk if required.  
            //
            if ((options & BurnRequestOptions.Erase) != 0)
            {
                try
                {
                    m_drive.ProgressChanged += new ProgressChangedEventHandler(ProgressHandler);
                    status = m_drive.Erase(Drive.EraseType.Minimal);
                    m_drive.ProgressChanged -= new ProgressChangedEventHandler(ProgressHandler);
                }
                catch (ThreadAbortException ex)
                {
                    m_drive.WaitForDeviceReady(false);
                    throw ex;
                }
            }

            m_creator.ValidationProgressChanged += new ProgressChangedEventHandler(ProgressHandler);
            if (!m_creator.Validate())
            {
                m_creator.ValidationProgressChanged -= new ProgressChangedEventHandler(ProgressHandler);
                ProgressHandler(null, new ProgressChangedEventArgs(100, null));
                args[0] = new BurnerError("image validation failed");
                Invoke(new DoneDelegate(BurnDone), args);
                return;
            }
            m_creator.ValidationProgressChanged -= new ProgressChangedEventHandler(ProgressHandler);

            //
            // Burn the image to the disk
            //
            DiskBurnImage dsk = new DiskBurnImage(m_logger);
            ImageGeneratorSource src = new ImageGeneratorSource(m_creator);
            TrackBurnImage t = new TrackBurnImage(TrackBurnImage.TrackType.Data_Mode_1, src);
            dsk.AddTrack(t);

            try
            {
                m_drive.ProgressChanged += new ProgressChangedEventHandler(ProgressHandler);
                status = m_drive.BurnDisk(m_burn_type, dsk, m_multi_session, m_burn_speed);
                m_drive.ProgressChanged -= new ProgressChangedEventHandler(ProgressHandler);
            }
            catch (ThreadAbortException)
            {
                //
                // If the burn was aborted, insure that the session is closed
                //
                m_drive.Device.CloseTrackSession(false, Device.CloseTrackSessionType.CloseSession, 0);
                return;
            }

            m_drive.Device.StartStopUnit(false, Device.PowerControl.NoChange, Device.StartState.EjectDisc);
            args[0] = status;
            Invoke(new DoneDelegate(BurnDone), args);
        }

        private void DriveSettings_Click(object sender, EventArgs e)
        {
            bool warning, debug ;

            if (m_drive == null)
            {
                MessageBox.Show("There is no drive selected, or the drive cannot be accessed", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            if (m_logger.HasSink(UserMessage.Category.Debug))
                debug = true;
            else
                debug = false;

            if (m_logger.HasSink(UserMessage.Category.Warning))
                warning = true;
            else
                warning = false;

            int level = m_logger.GetLevel(UserMessage.Category.Debug);

            SpeedInfo[] speeds;
            DiskOperationError status = m_drive.GetWriteSpeeds(out speeds);
            if (status != null)
            {
                string str = "Cannot retreive speed data from drive - " + status.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            DriveSettings form = new DriveSettings(debug, level, warning, m_multi_session, m_burn_type, speeds,
                                                   m_drive.IsTrackAtOnceSupported, m_drive.IsSessionAtOnceSupported,
                                                   m_drive.IsMultiSessionSupported);
            if (form.ShowDialog(this) == DialogResult.OK)
            {
                m_multi_session = form.LeaveSessionOpen;
                if (form.DebugMessagesFlag)
                {
                    m_logger.SetSink(UserMessage.Category.Debug, m_sink);
                    m_logger.SetLevel(UserMessage.Category.Debug, form.DebugLevel);
                }
                else
                    m_logger.RemoveSink(UserMessage.Category.Debug);

                if (form.WarningMessagesFlag)
                    m_logger.SetSink(UserMessage.Category.Warning, m_sink);
                else
                    m_logger.RemoveSink(UserMessage.Category.Warning);

                m_burn_type = form.BurnType;
                m_burn_speed = form.BurnSpeedValue;

                SaveLoggerSettings();
                SetStartupInteger(PropertyNameCreateDataDiskBurnType, (int)m_burn_type);
                SetStartupBool(PropertyNameCreateDataDiskMultiSession, m_multi_session);
            }
        }

        private void SaveImageToFile(string filename)
        {
            try
            {
                FileStream strm = new FileStream(filename, FileMode.Create, FileAccess.Write);
                BinaryFormatter formatter = new BinaryFormatter();
                formatter.Serialize(strm, m_image);
                strm.Close();

                //
                // We just loaded this image, it is not dirty
                //
                m_image.IsDirty = false;
            }
            catch (Exception ex)
            {
                string str = "Error writing project file '" + filename + "' - " + ex.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

        }

        private void LoadImageFromFile(string filename)
        {
            try
            {
                FileStream strm = new FileStream(filename, FileMode.Open, FileAccess.Read);
                BinaryFormatter formatter = new BinaryFormatter();
                m_image = (DiskImageContents)formatter.Deserialize(strm);
                strm.Close();

                m_image.Initialize(m_logger);
                m_image.FileName = filename;

                m_image.IsDirty = false;
                DiskImageTree.ImageCreator = m_image;

                m_creator = new DiskImageCreator(m_logger, m_image);
                CreatorChanged(null, null);
            }
            catch (Exception ex)
            {
                string str = "Error reading project file '" + filename + "' - " + ex.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void newtToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_image.IsDirty)
            {
                DialogResult res;

                res = MessageBox.Show("The current project has not been saved.  Do you want to save this current project?",
                                      "Save?", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
                if (res == DialogResult.Yes)
                    saveToolStripMenuItem_Click(null, null);
                else if (res == DialogResult.Cancel)
                    return;
            }

            m_image = new DiskImageContents(m_logger);
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DialogResult res;

            if (m_image.IsDirty)
            {

                res = MessageBox.Show("The current project has not been saved.  Do you want to save this current project?",
                                      "Save?", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
                if (res == DialogResult.Yes)
                    saveToolStripMenuItem_Click(null, null);
                else if (res == DialogResult.Cancel)
                    return;
            }

            OpenFileDialog dialog = new OpenFileDialog();
            dialog.Multiselect = false;
            dialog.Filter = "Data Disk Project Files (*.bwg)|*.bwg|All files (*.*)|*.*";
            dialog.RestoreDirectory = true;
            dialog.AddExtension = true;
            dialog.CheckFileExists = true;
            dialog.DefaultExt = "bwg";
            dialog.FilterIndex = 0;

            res = dialog.ShowDialog();
            if (res == DialogResult.OK)
            {
                LoadImageFromFile(dialog.FileName);
                m_image.Logger = m_logger;
            }
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_image.FileName.Length == 0)
                saveAsToolStripMenuItem_Click(sender, e);

            SaveImageToFile(m_image.FileName);
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DialogResult res;

            SaveFileDialog dialog = new SaveFileDialog();
            dialog.Filter = "Data Disk Project Files (*.bwg)|*.bwg|All files (*.*)|*.*";
            dialog.RestoreDirectory = true;
            dialog.AddExtension = true;
            dialog.CheckPathExists = true;
            dialog.DefaultExt = "bwg";
            dialog.FilterIndex = 0;

            res = dialog.ShowDialog();
            if (res == DialogResult.OK)
            {
                m_image.FileName = dialog.FileName;
                SaveImageToFile(m_image.FileName);
            }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void UpdateViewMenu()
        {
            bool local = LocalNamesMenuItem.Checked;
            bool iso = ISONamesMenuItem.Checked;
            bool joliet = JolietNamesMenuItem.Checked;
            bool udf = UDFNamesMenuItem.Checked;

            AllViewItemsOff();

            if (local)
                LocalNamesMenuItem.Checked = true;
            else if (iso)
                ISONamesMenuItem.Checked = true;
            else if (joliet)
                JolietNamesMenuItem.Checked = true;
            else
                UDFNamesMenuItem.Checked = true;
        }

        private void AllViewItemsOff()
        {
            LocalNamesMenuItem.Checked = false;
            ISONamesMenuItem.Checked = false;
            JolietNamesMenuItem.Checked = false;
            UDFNamesMenuItem.Checked = false;

            LocalNamesMenuItem.Enabled = true;
            ISONamesMenuItem.Enabled = true;
            JolietNamesMenuItem.Enabled = true;
            UDFNamesMenuItem.Enabled = true;

            if (m_creator.Settings.ISOImage == Bwg.Isofs.ImageSettings.ISOImageType.None)
            {
                ISONamesMenuItem.Enabled = false;
                JolietNamesMenuItem.Enabled = false;
            }

            if (m_creator.Settings.UDFImage == Bwg.Isofs.ImageSettings.UDFImageType.None)
            {
                UDFNamesMenuItem.Enabled = false;
            }

        }

        private void localNamesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AllViewItemsOff();
            LocalNamesMenuItem.Checked = true;
            Form f = new UpdatingView();
            f.Show();
            Application.DoEvents();
            DiskImageTree.NameType = DiskImageElement.NameType.Local;
            DiskImageList.NameType = DiskImageElement.NameType.Local;
            f.Close();
        }

        private void iSO9660NamesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AllViewItemsOff();
            ISONamesMenuItem.Checked = true;
            Form f = new UpdatingView();
            f.Show();
            Application.DoEvents();
            DiskImageTree.NameType = DiskImageElement.NameType.ISO9660;
            DiskImageList.NameType = DiskImageElement.NameType.ISO9660;
            f.Close();
        }

        private void jolietNamesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AllViewItemsOff();
            JolietNamesMenuItem.Checked = true;
            Form f = new UpdatingView();
            f.Show();
            Application.DoEvents();
            DiskImageTree.NameType = DiskImageElement.NameType.Joliet;
            DiskImageList.NameType = DiskImageElement.NameType.Joliet;
            f.Close();
        }

        private void uDFNamesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AllViewItemsOff();
            UDFNamesMenuItem.Checked = true;


            Form f = new UpdatingView();
            f.Show();
            Application.DoEvents();
            DiskImageTree.NameType = DiskImageElement.NameType.UDF;
            DiskImageList.NameType = DiskImageElement.NameType.UDF;
            f.Close();
        }

        private void DiskSpace_Click(object sender, EventArgs e)
        {
            m_remaining = !m_remaining;
            CreatorChanged(null, null);
        }


        private void fullToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_drive.IsEraseSupported)
            {
                EnableCreateImageControls(false);

                // Start a timer to display elapsed time
                m_start = DateTime.Now;
                m_timer.Interval = 1000;
                m_timer.Tick += new EventHandler(TimerTickHandler);
                m_timer.Tag = this;
                m_timer.Start();

                // Now, create a thread to burn the disk
                m_burner_thread = new Thread(new ParameterizedThreadStart(EraseProc));
                m_burner_thread.Start(true);
            }
            else
            {
                string str = "A disk of type '" + m_drive.CurrentProfile.ToString() + "' is in the drive, this type does not ";
                str += "support erase.";
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);

                m_device.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
            }
        }

        private void quickToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_drive.IsEraseSupported)
            {
                DiscInformation info;
                Device.CommandStatus st;

                st = m_drive.Device.ReadDiscInformation(out info);
                if (st == Device.CommandStatus.Success && info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
                {
                    MessageBox.Show("The media is already erased!", "Information!", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    return;
                }

                EnableCreateImageControls(false);

                // Start a timer to display elapsed time
                m_start = DateTime.Now;
                m_timer.Interval = 1000;
                m_timer.Tick += new EventHandler(TimerTickHandler);
                m_timer.Tag = this;
                m_timer.Start();

                // Now, create a thread to burn the disk
                m_burner_thread = new Thread(new ParameterizedThreadStart(EraseProc));
                m_burner_thread.Start(false);
            }
            else
            {
                string str = "A disk of type '" + m_drive.CurrentProfile.ToString() + "' is in the drive, this type does not ";
                str += "support erase.";
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);

                m_device.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
            }
        }

        private void EraseDone(DiskOperationError status)
        {
            string msg;
            m_timer.Stop();

            m_burner_thread = null;

            EnableCreateImageControls(true);

            string caption = "Error!";
            MessageBoxIcon localicon;

            if (status == null)
            {
                msg = "The erase operation completed sucessfully";
                caption = "Complete";
                localicon = MessageBoxIcon.Information;
            }
            else
            {
                msg = "The erase operation failed - " + status.Message;
                caption = "Error!";
                localicon = MessageBoxIcon.Error;
            }

            // This forces a refresh all data associated with the media in the drive
            DeviceSelector_SelectedIndexChanged(null, null);


            MessageBox.Show(msg, caption, MessageBoxButtons.OK, localicon);
        }
        
        private void EraseProc(object param)
        {
            DiskOperationError status;

            Debug.Assert(param is bool);
            bool isfull = (bool)param;

            try
            {
                m_drive.ProgressChanged += new ProgressChangedEventHandler(ProgressHandler);
                status = m_drive.Erase(isfull ? Drive.EraseType.Full : Drive.EraseType.Minimal);
                m_drive.ProgressChanged -= new ProgressChangedEventHandler(ProgressHandler);
            }
            catch (ThreadAbortException ex)
            {
                m_drive.WaitForDeviceReady(false);
                throw ex;
            }

            object[] args = new object[1];
            args[0] = status;
            Invoke(new DoneDelegate(EraseDone), args);
        }
        #endregion
    }
}
