//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Diagnostics;

namespace Bwg.Database.Freedb
{
    public class FreeDBAlbumInfo : ICloneable
    {
        #region private variables
        /// <summary>
        /// The title of the album
        /// </summary>
        private string m_title;

        /// <summary>
        /// The artist that performed this album
        /// </summary>
        private string m_artist;

        /// <summary>
        /// The year the album was recorded
        /// </summary>
        private int m_year;

        /// <summary>
        /// The Genre for the album
        /// </summary>
        private string m_genre;

        /// <summary>
        /// The category for this album
        /// </summary>
        private string m_category;

        /// <summary>
        /// If true, this data is directory from freedb
        /// </summary>
        private bool m_from_freedb;

        /// <summary>
        /// The tracks on this disk.  The final track will be the leadout
        /// </summary>
        public FreeDBTrackInfo[] m_tracks;
        #endregion

        #region public properties

        /// <summary>
        /// This property is true if the data contained in this object came from
        /// freedb.org and has not been changed.
        /// </summary>
        public bool FromFreedb
        {
            get
            {
                return m_from_freedb;
            }
            set
            {
                m_from_freedb = value;
            }
        }

        /// <summary>
        /// This property returns true if the album is filled in and can be submitted
        /// to the free database.
        /// </summary>
        public bool ReadyForSubmit
        {
            get
            {
                if (m_artist.Length == 0)
                    return false;

                if (m_title.Length == 0)
                    return false;

                for(int i = 0 ; i < m_tracks.GetLength(0) - 1 ; i++)
                {
                    FreeDBTrackInfo tinfo = m_tracks[i] ;
                    if (tinfo.Title.Length >=5 && tinfo.Title.Substring(0, 5) == "Track")
                        return false;
                }
                return true;
            }
        }

            /// <summary>
            /// The total seconds for this track
            /// </summary>
            public int TotalSeconds
        {
            get
            {
                long l = 0;

                for (int i = 0; i < m_tracks.GetLength(0) - 1; i++)
                {
                    FreeDBTrackInfo t = m_tracks[i];
                    l += t.LengthFrame;
                    l += t.LengthSecond * 75;
                    l += t.LengthMinute * 60 * 75;
                }

                return (int)(l / 75);
            }
        }

        /// <summary>
        /// This property is a string that represents the title of the album
        /// </summary>
        public string Title
        {
            get
            {
                return m_title;
            }
            set
            {
                if (m_title != value)
                {
                    m_from_freedb = false;
                    m_title = value;
                }
            }
        }

        /// <summary>
        /// This property is a string that represents the artist for the album
        /// </summary>
        public string Artist
        {
            get
            {
                return m_artist;
            }
            set
            {
                if (m_artist != value)
                {
                    m_from_freedb = false;
                    m_artist = value;
                }
            }
        }

        /// <summary>
        /// The year the ablum was recorded.
        /// </summary>
        public int Year
        {
            get
            {
                return m_year;
            }
            set
            {
                if (m_year != value)
                {
                    m_from_freedb = false;
                    m_year = value;
                }
            }
        }

        /// <summary>
        /// The genre for the ablum
        /// </summary>
        public string Genre
        {
            get
            {
                return m_genre;
            }
            set
            {
                if (m_genre != value)
                {
                    m_from_freedb = false;
                    m_genre = value;
                }
            }
        }

        /// <summary>
        /// The number of trackis in the album (including the leadout)
        /// </summary>
        public int TrackCount
        {
            get
            {
                return m_tracks.GetLength(0);
            }
        }

        /// <summary>
        /// The category for the album
        /// </summary>
        public string Category
        {
            get
            {
                return m_category;
            }
            set
            {
                m_category = value;
            }
        }

        /// <summary>
        /// This is a user tag to be used for whatever purpose.
        /// </summary>
        public object Tag;

        #endregion

        #region constructors
        public FreeDBAlbumInfo()
        {
            m_tracks = new FreeDBTrackInfo[0];
            m_title = string.Empty;
            m_artist = string.Empty;

            m_category = string.Empty;
            m_genre = string.Empty;
            m_year = 0;
        }
        #endregion

        #region public methods

        public void SetTrackTitle(int index, string title)
        {
            if (m_tracks[index].Title != title)
            {
                m_tracks[index].Title = title;
                m_from_freedb = false;
            }
        }

        public void SetTrackArtist(int index, string artist)
        {
            if (m_tracks[index].Artist != artist)
            {
                m_tracks[index].Artist = artist;
                m_from_freedb = false;
            }
        }

        public override string ToString()
        {
            return m_title;
        }

        public Object Clone()
        {
            FreeDBAlbumInfo obj = new FreeDBAlbumInfo();

            obj.Year = Year;
            obj.Tag = Tag;
            obj.Title = Title;
            obj.Artist = Artist;
            obj.Genre = Genre;
            obj.Category = Category;

            for (int i = 0; i < TrackCount; i++)
            {
                FreeDBTrackInfo tinfo = m_tracks[i];
                obj.AddTrack(tinfo.Title, tinfo.StartMinute, tinfo.StartSecond, tinfo.StartFrame,
                                          tinfo.LengthMinute, tinfo.LengthSecond, tinfo.LengthFrame);

            }

            return obj;
        }

        public FreeDBTrackInfo this[int index]
        {
            get
            {
                return m_tracks[index];
            }
        }

        public void AddTrack(string title, byte sm, byte ss, byte sf, byte lm, byte ls, byte lf)
        {
            FreeDBTrackInfo t = new FreeDBTrackInfo(sm, ss, sf, lm, ls, lf);
            t.Title = title;
            Array.Resize(ref m_tracks, m_tracks.GetLength(0) + 1);
            m_tracks[m_tracks.GetLength(0) - 1] = t;
        }
        #endregion
    }
}
