//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Bwg.Logging;
using Bwg.Utils;
using Bwg.Scsi;
using Bwg.Burner;

namespace Bwg.DiskImage
{
    public sealed class FileDiskContents : DiskContents
    {
        #region private member variables
        private OsFileReader m_file;
        private byte m_minor_version;
        private byte m_major_version;
        private IDictionary<TrackInfo, long> m_track_offsets;
        #endregion

        #region constructors
        public FileDiskContents(Logger l, OsFileReader rdr) : base(l)
        {
            m_file = rdr;
        }
        #endregion

        #region protected methods
        protected override DiskOperationError Initialize()
        {
            DiskOperationError status;

            //
            // Read the disk and sessions headers
            //
            status = Read();
            if (status != null)
                return status;

            //
            // Analyze the image to find the track ranges
            //
            AnalyzeTrackStartOffsets();

            return null;
        }

        internal override int MaxBufferSize
        {
            get
            {
                return 65536 * 4;           // 256K size buffer
            }
        }

        internal override DiskOperationError ReadTrackData(int current, int toread, int secsize, IntPtr buf)
        {
            int amount;

            //
            // Find the track assocaited with the given sector
            //
            TrackInfo tinfo = FindTrack(current, toread);
            if (tinfo == null)
                throw new Exception("Invalid image file - cannot find track from sector");

            long trstart = FindTrackDataStart(tinfo, current - tinfo.Start);
            if (!m_file.Seek(trstart, OsFileReader.SeekType.Begin))
                return new DeviceIoctlError("Seek", (uint)m_file.LastError);

            if (!m_file.ReadData(buf, toread * secsize, out amount))
                return new DeviceIoctlError("ReadData", (uint)m_file.LastError);

            if (amount != toread * secsize)
                return new BurnerError("Unexpected end of file reading disk image file");

            return null;
        }

        #endregion

        #region private member functions

        private void AnalyzeTrackStartOffsets()
        {
            m_track_offsets = new Dictionary<TrackInfo, long>();

            long offset = 2048 + m_sessions.GetLength(0) * 2048;

            foreach (SessionInfo sinfo in m_sessions)
            {
                foreach (TrackInfo tinfo in sinfo.Tracks)
                {
                    m_track_offsets[tinfo] = offset;
                    Debug.WriteLine("Track " + tinfo.Number.ToString() + ", offset is " + offset.ToString());
                    offset += (long)tinfo.Size * (long)tinfo.SectorSize;
                }
            }
        }

        private long FindTrackDataStart(TrackInfo tinfo, int sector)
        {
            return m_track_offsets[tinfo] + (long)sector * (long)tinfo.SectorSize;
        }

        private TrackInfo FindTrack(int sector, int size)
        {
            foreach(SessionInfo sinfo in m_sessions)
            {
                foreach (TrackInfo tinfo in sinfo.Tracks)
                {
                    if (sector >= tinfo.Start && sector + size <= tinfo.Start + tinfo.Size)
                        return tinfo;
                }
            }

            return null;
        }

        private DiskOperationError Read()
        {
            byte[] data = new byte[2048];
            int amount ;

            if (!m_file.ReadData(ref data, out amount))
                return new DeviceIoctlError("ReadData", (uint)m_file.LastError);

            if (amount != data.GetLength(0))
                return new BurnerError("ReadData, unexpected EOF reading disk image");

            if (data[0] != Signature1 || data[1] != Signature2 || data[2] != Signature3 || data[3] != Signature4)
                return new BurnerError("Invalid image file, not a BIM image");

            if (data[4] != 1 || data[5] != 0)
                return new BurnerError("Unknown BIM image version, version " + data[4].ToString() + "not supported");

            m_minor_version = data[5];
            m_major_version = data[4];

            if (data[7] != 1)
                return new BurnerError("No support for multi-session disks (YET)");

            m_disk_type = (ImageType)data[6];

            for (int i = 1; i <= data[7]; i++)
            {
                DiskOperationError status = ReadSession(i);
                if (status != null)
                    return status;
            }

            return null;
        }

        private DiskOperationError ReadSession(int num)
        {
            byte[] data = new byte[2048];
            int offset, amount;

            if (!m_file.ReadData(ref data, out amount))
                return new DeviceIoctlError("ReadData", (uint)m_file.LastError);

            if (amount != data.GetLength(0))
                return new BurnerError("ReadData, unexpected EOF reading disk image");

            if (data[0] != num)
                return new BurnerError("Invalid image file, expected session " + num.ToString() + ", got session " + data[0].ToString());

            if (data[1] > 99)
                return new BurnerError("Invalid image file, invalid track count " + data[1].ToString() + " in session " + num.ToString());

            offset = 16;
            SessionInfo info = new SessionInfo(num);
            for (int i = 1; i <= data[1]; i++)
            {
                int tno, tstart, tsize;
                TrackBurnImage.TrackType ttype;

                tno = data[offset] | (data[offset + 1] << 8) | (data[offset + 2] << 16) | (data[offset + 3] << 24);
                offset += 4;

                ttype = (TrackBurnImage.TrackType)(data[offset] | (data[offset + 1] << 8) | (data[offset + 2] << 16) | (data[offset + 3] << 24));
                offset += 4;

                tstart = data[offset] | (data[offset + 1] << 8) | (data[offset + 2] << 16) | (data[offset + 3] << 24);
                offset += 4;

                tsize = data[offset] | (data[offset + 1] << 8) | (data[offset + 2] << 16) | (data[offset + 3] << 24);
                offset += 4;

                TrackInfo tinfo = new TrackInfo(ttype, tno, tstart, tsize);
                info.AddTrack(tinfo);
            }

            Array.Resize(ref m_sessions, m_sessions.GetLength(0) + 1);
            m_sessions[m_sessions.GetLength(0) - 1] = info;

            return null;
        }
        #endregion
    }
}
