//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.ComponentModel;
using Bwg.Utils ;
using Bwg.Logging ;
using Bwg.Scsi ;
using Bwg.Burner ;

namespace Bwg.DiskImage
{
    internal class SessionInfo
    {
        public readonly int Number;
        public TrackInfo [] Tracks ;

        /// <summary>
        /// This event is signaled when the progress has changed
        /// </summary>
        public ProgressChangedEventHandler ProgressChanged;

        #region constructors
        public SessionInfo(int num)
        {
            Number = num;
            Tracks = new TrackInfo[0] ;
        }
        #endregion

        #region public methods
        public DiskOperationError WriteHeader(OsFileWriter wrtr)
        {
            byte[] data = new byte[2048];
            int offset, amount;

            data[0] = (byte)Number;                     // Byte 0, session #
            data[1] = (byte)Tracks.GetLength(0);        // Byte 1, number of tracks, for session 1, includes leadin
            offset = 16 ;

            foreach (TrackInfo tinfo in Tracks)
            {
                Debug.Assert(offset < data.GetLength(0));

                //
                // Write the track header.  A track header consists of a four byte track #,
                // a 4 byte starting sector, and a 4 byte size in sectors
                //
                // Yes, I know that track #s cannot be four bytes, but keeping this four bytes
                // makes the file easier to read and therefore easier to debug.
                //
                data[offset++] = (byte)((tinfo.Number >> 0) & 0xff);
                data[offset++] = (byte)((tinfo.Number >> 8) & 0xff);
                data[offset++] = (byte)((tinfo.Number >> 16) & 0xff);
                data[offset++] = (byte)((tinfo.Number >> 24) & 0xff);

                data[offset++] = (byte)tinfo.TrackType;
                data[offset++] = 0;
                data[offset++] = 0;
                data[offset++] = 0;

                data[offset++] = (byte)((tinfo.Start >> 0) & 0xff);
                data[offset++] = (byte)((tinfo.Start >> 8) & 0xff);
                data[offset++] = (byte)((tinfo.Start >> 16) & 0xff);
                data[offset++] = (byte)((tinfo.Start >> 24) & 0xff);

                data[offset++] = (byte)((tinfo.Size >> 0) & 0xff);
                data[offset++] = (byte)((tinfo.Size >> 8) & 0xff);
                data[offset++] = (byte)((tinfo.Size >> 16) & 0xff);
                data[offset++] = (byte)((tinfo.Size >> 24) & 0xff);
            }

            if (!wrtr.WriteData(data, out amount) || amount != data.GetLength(0))
                return new DeviceIoctlError("WriteData", (uint)wrtr.LastError);

            return null;
        }

        public DiskOperationError Write(OsFileWriter wrtr, DiskContents dc)
        {
            DiskOperationError status ;

            foreach (TrackInfo tinfo in Tracks)
            {
                status = CopyOneTrack(wrtr, dc, tinfo, Tracks.GetLength(0));
                if (status != null)
                    return status;
            }
            return null;
        }

        public void AddTrack(TrackInfo info)
        {
            Array.Resize(ref Tracks, Tracks.GetLength(0) + 1);
            Tracks[Tracks.GetLength(0) - 1] = info;
        }
        #endregion

        #region private methods
        private int GetOptimalBufferSize(DiskContents dc, TrackInfo info)
        {
            int bufsize = dc.MaxBufferSize;
            int sectsize = info.SectorSize;
            int blocks = bufsize / sectsize;

            return blocks * sectsize;
        }

        private DiskOperationError CopyOneTrack(OsFileWriter wrtr, DiskContents dc, TrackInfo tinfo, int total)
        {
            DiskOperationError status;
            int size = GetOptimalBufferSize(dc, tinfo);
            int blks = size / tinfo.SectorSize ;
            IntPtr buf = Marshal.AllocHGlobal(size);
            int current = 0;
            int amount;

            Debug.WriteLine("Copying track " + tinfo.Number.ToString());
            if (ProgressChanged != null)
            {
                ProgressText pt = new ProgressText("Copying Track " + tinfo.Number.ToString() + " of " + total.ToString(), "Copying");
                ProgressChangedEventArgs e = new ProgressChangedEventArgs(0, pt);
                ProgressChanged(this, e);
            }

            while (current < tinfo.Size)
            {
                int toread = blks ;
                if (toread > tinfo.Size - current)
                    toread = tinfo.Size - current ;

                if (ProgressChanged != null)
                {
                    int pcnt = current * 100 / tinfo.Size;
                    ProgressText pt = new ProgressText("Copying Track " + tinfo.Number.ToString() + " of " + total.ToString(), "Copying");
                    ProgressChangedEventArgs e = new ProgressChangedEventArgs(pcnt, pt);
                    ProgressChanged(this, e);
                }

                status = dc.ReadTrackData(current + tinfo.Start, toread, tinfo.SectorSize, buf);
                if (status != null)
                    return status;

                if (!wrtr.WriteData(buf, toread * tinfo.SectorSize, out amount))
                    return new DeviceIoctlError("WriteData", (uint)wrtr.LastError);

                if (amount != toread * tinfo.SectorSize)
                    return new BurnerError("WriteData, write error did not write sufficient data");

                current += toread;
            }

            if (ProgressChanged != null)
            {
                ProgressText pt = new ProgressText("Copying Track " + tinfo.Number.ToString() + " of " + total.ToString(), "Copying");
                ProgressChangedEventArgs e = new ProgressChangedEventArgs(100, pt);
                ProgressChanged(this, e);
            }
            Marshal.FreeHGlobal(buf);
            return null;
        }
        #endregion
    }
}
