using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Bwg.DvdMaster;

namespace Bwg.DvdReader
{
    public class Reader
    {
        #region private constants
        private int BytesPerSector = 2048;
        #endregion

        #region private member variables
        /// <summary>
        /// The base directory for the DVD (should contains the VIDEO_TS subdirectory)
        /// </summary>
        private string m_base_dir;

        /// <summary>
        /// The disc we are populating
        /// </summary>
        private DvdDisc m_disc;

        /// <summary>
        /// The list of messages generated during the reading process
        /// </summary>
        private IList<string> m_messages;
        #endregion

        #region constructor
        /// <summary>
        /// Create a new reader object and read the contents of the DVD into disc object
        /// </summary>
        /// <param name="dir"></param>
        /// <param name="disc"></param>
        public Reader(string dir, DvdDisc disc)
        {
            m_base_dir = dir;
            m_disc = disc;
        }
        #endregion

        #region public methods
        /// <summary>
        /// This method reads the contents of a disc into the disc structure
        /// </summary>
        /// <returns></returns>
        public bool ReadDiscContents()
        {
            bool status;

            m_messages = new List<string>();
            m_messages.Add("Reading DVD from base directory '" + m_base_dir + "'");

            //
            // Read the video manager info file.
            //
            status = ReadVideoManager("VIDEO_TS.IFO");
            if (!status)
            {
                //
                // The default version may not be readable
                //
                status = ReadVideoManager("VIDEO_TS.BUP");
                if (!status)
                    return false;
            }

            return true;
        }
        #endregion

        #region private methods
        private bool ReadVideoManager(string filename)
        {
            string fname = Path.Combine(m_base_dir, "VIDEO_TS");
            fname = Path.Combine(fname, filename);

            FileStream strm = new FileStream(fname, FileMode.Open, FileAccess.Read);
            if (!ReadVideoManagerInfoManagement(strm))
            {
                strm.Close() ;
                return false;
            }

            if (!ReadVideoManagerPartOfTitleSearch(strm))
            {
                strm.Close();
                return false;
            }

            if (!ReadVideoManagerProgramChainUnits(strm))
            {
                strm.Close();
                return false;
            }

            return true;
        }

        private bool ReadVideoManagerProgramChainUnits(FileStream strm)
        {
            byte[] data;

            if (m_disc.VideoManagerInformationManagement.VmgmPgciUtStartSector == 0)
                return true;

            long pos = m_disc.VideoManagerInformationManagement.VmgmPgciUtStartSector * BytesPerSector;
            if (strm.Seek(pos + 4, SeekOrigin.Begin) != pos + 4)
                return false;

            int count = (strm.ReadByte() << 24) | (strm.ReadByte() << 16) | (strm.ReadByte() << 8) | strm.ReadByte() + 1;
            if (strm.Seek(pos, SeekOrigin.Begin) != pos)
                return false;

            data = new byte[count];
            if (strm.Read(data, 0, count) != count)
                return false;

            m_disc.VideoManagerProgramChainUnits = new VMGM_PGCI_UT(data);

            return true;
        }

        private bool ReadVideoManagerPartOfTitleSearch(FileStream strm)
        {
            byte [] data ;

            if (m_disc.VideoManagerInformationManagement.VmgPttSrptStartSector == 0)
                return true;

            long pos = m_disc.VideoManagerInformationManagement.VmgPttSrptStartSector * BytesPerSector ;
            if (strm.Seek(pos, SeekOrigin.Begin) != pos)
                return false ;

            int count = (strm.ReadByte() << 8) | strm.ReadByte() ;
            if (strm.Seek(pos, SeekOrigin.Begin) != pos)
                return false ;

            int bytecnt = 8 + 12 * count ;
            data = new byte[bytecnt];
            if (strm.Read(data, 0, bytecnt) != bytecnt)
                return false;

            m_disc.VideoManagerPartOfTitleSearch = new VMG_PTT_SRPT(data);

            return true;
        }

        private bool ReadVideoManagerInfoManagement(FileStream strm)
        {
            byte[] data = new byte[1024];

            if (strm.Seek(0, SeekOrigin.Begin) != 0)
                return false;

            if (strm.Read(data, 0, 1024) != 1024)
                return false;

            m_disc.VideoManagerInformationManagement = new VMGM_MAT(data);
            return true;
        }
        #endregion
    }
}
