//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Security;
using System.Diagnostics;
using Bwg.Logging;

namespace Bwg.Isofs
{
    /// <summary>
    /// This class represents the contents of a disk image.  It contains a set of mappings
    /// from disk image files and directories to directories and files on the local computer.  While
    /// not every directory in the image is mapped directly to a directory on the computer, each
    /// file in the image is mapped to a file on the computer.
    /// </summary>
    [Serializable()]
    public class DiskImageContents
    {
        #region private member variables

        /// <summary>
        /// Those chars that are valid A characters
        /// </summary>
        [NonSerialized()]
        private const string ValidAChars = " !\"%&'()*+,-./0123456789;:<=>?ABCDEFGHIJKLMNOPQRSTUVWXYZ_";

        /// <summary>
        /// This value is true if the contents have been changed since the file has last been saved.
        /// </summary>
        [NonSerialized()]
        private bool m_dirty;

        /// <summary>
        /// The logger associated with this image
        /// </summary>
        [NonSerialized()]
        private Logger m_logger;

        /// <summary>
        /// The name of the file where the data is stored
        /// </summary>
        [NonSerialized()]
        private string m_filename;

        /// <summary>
        /// This object contains all of the variable settings associated with the image
        /// </summary>
        private ImageSettings m_settings;

        /// <summary>
        /// This is the top level directory for the ISO file system.  This directory
        /// will not contain a local file system location
        /// </summary>
        private DiskImageDirectory m_toplevel;

        /// <summary>
        /// This list of allowed extensions for this image.  If this is null, all extension are
        /// allowed.
        /// </summary>
        private IList<string> m_allowed_exts;

        /// <summary>
        /// If true, we are within a large update and we do not update the size until this is over
        /// </summary>
        private bool m_within_update;

        #endregion

        #region public constants

        /// <summary>
        /// This is the maximum size of a file that can exist on an ISO image
        /// </summary>
        public const long MaximumISOFileSize = 2147483648;

        #endregion

        #region public member variables

        /// <summary>
        /// This event is fired when the disk image size is changed to adding
        /// or removing files
        /// </summary>
        [NonSerialized()]
        public EventHandler<EventArgs> SizeChanged;

        /// <summary>
        /// This event is fired when an element in the disk image is removed or
        /// renamed.
        /// </summary>
        [NonSerialized()]
        public EventHandler<ImageChangedEventArgs> ImageChanged;

        /// <summary>
        /// This event is fired just before a directory add to the image
        /// </summary>
        [NonSerialized()]
        public EventHandler<EventArgs> BeforeDirectoryAdd;

        /// <summary>
        /// This event is fired after a directory add to the image
        /// </summary>
        [NonSerialized()]
        public EventHandler<EventArgs> AfterDirectoryAdd;

        /// <summary>
        /// This event is fired when a new directory is being added to the image
        /// </summary>
        [NonSerialized()]
        public EventHandler<AddingDirectoryEventArgs> AddingDirectory;

        /// <summary>
        /// This event is fired when a new file is added to the directory
        /// </summary>
        [NonSerialized()]
        public EventHandler<AddingFileEventArgs> AddingFile;

        #endregion

        #region public properties

        /// <summary>
        /// This property is the logger for displaying messages to the user
        /// </summary>
        public Logger Logger
        {
            get
            {
                return m_logger;
            }
            set
            {
                m_logger = value;
            }
        }

        /// <summary>
        /// Returns true if the contents or settings are dirty (changed since saved)
        /// </summary>
        public bool IsDirty
        {
            get
            {
                return m_dirty || m_settings.IsDirty;
            }
            set
            {
                m_dirty = value;
                m_settings.IsDirty = value;
            }
        }

        /// <summary>
        /// Return the settings assocaited with this content
        /// </summary>
        public ImageSettings Settings
        {
            get
            {
                return m_settings;
            }
            set
            {
                m_settings = value;
                m_toplevel.UpdateSettings(m_settings);
            }
        }

        /// <summary>
        /// The name of the file where these project settings are stored
        /// </summary>
        public string FileName
        {
            get
            {
                return m_filename;
            }
            set
            {
                m_filename = value;
            }
        }

        /// <summary>
        /// The root directory of the disk image
        /// </summary>
        public DiskImageDirectory RootDirectory
        {
            get
            {
                return m_toplevel;
            }
        }

        /// <summary>
        /// This property returns the total number of files in the image contents
        /// </summary>
        public int FileCount
        {
            get
            {
                return RootDirectory.RecursiveFileCount;
            }
        }

        /// <summary>
        /// This property returns the total number of directories in this image
        /// </summary>
        public int DirectoryCount
        {
            get
            {
                return RootDirectory.RecursiveDirectoryCount;
            }
        }

        /// <summary>
        /// This property controls the SizeChanged event.  If this value is set to true
        /// SizeChanged events are not sent, assuming a large update.  If this value is false
        /// SizeChanged events are sent on every image size change.  Also, if this value is set
        /// to false, it forces a SizeChanged event as it is assumed a large update has just
        /// finished.
        /// 
        /// NOTE: This is kind of a kludge to deal with the runtimes associated with the large
        ///       images associated with BluRay discs.  A better solution is to always send the
        ///       size changed event, but have the SizeInBlocks for the DiskImageCreator class
        ///       be more efficient in dealing with incremental adds of files and directories.
        /// </summary>
        public bool WithinUpdate
        {
            get
            {
                return m_within_update;
            }
            set
            {
                m_within_update = value;
                if (m_within_update == false && SizeChanged != null)
                {
                    SizeChanged(this, new EventArgs());
                }
            }
        }

        #endregion

        #region constructor
        /// <summary>
        /// Create a new disk image object.
        /// </summary>
        public DiskImageContents(Logger l)
        {
            m_logger = l;
            m_settings = new ImageSettings();
            m_toplevel = new DiskImageDirectory(m_settings);
            m_filename = string.Empty;
            m_allowed_exts = null;
        }
        #endregion

        #region public member functions

        /// <summary>
        /// This methods returns true if the disk image contains any flies that are equal to
        /// or greater than the size given.
        /// </summary>
        /// <param name="size">the size of interest</param>
        /// <returns>true if a file is found of the size given or greater, otherwise false</returns>
        public bool ContainsFilesSized(long size)
        {
            if (m_toplevel == null)
                return false;

            return m_toplevel.ContainsFilesSized(size);
        }

        /// <summary>
        /// Remove all files of equal or greater to the size given
        /// </summary>
        /// <param name="size">remove all files of equal or greater size</param>
        public int RemoveFilesSized(long size)
        {
            int ret = 0;

            if (m_toplevel != null)
                ret = m_toplevel.RemoveFilesSized(size);

            return ret;
        }

        /// <summary>
        /// This method adds an extension to the list of allowed extensions
        /// </summary>
        /// <param name="ext">the extension to add</param>
        public void AddExt(string ext)
        {
            if (m_allowed_exts == null)
                m_allowed_exts = new List<string>();

            m_allowed_exts.Add(ext);
        }

        /// <summary>
        /// This method initializes this object after it has been serialized from disk.
        /// </summary>
        /// <param name="logger"></param>
        public void Initialize(Logger logger)
        {
            m_logger = logger;
            m_filename = string.Empty;
        }

        /// <summary>
        /// Add a new element to the disk.  If this element is a directory, all files
        /// and directories contained in the given directory are also added to the
        /// image.
        /// </summary>
        /// <param name="isoloc">the location of the element in the disk image</param>
        /// <param name="local">the location of the element in the location file system.</param>
        /// <returns></returns>
        public DiskImageElement Add(string isoloc, string local)
        {
            DiskImageElement ret = null;
            EventArgs args = new EventArgs();

            DiskImageDirectory root = FindRoot(ref isoloc, "", 0);
            if (File.Exists(local))
            {
                ret = AddFile(root, isoloc, local);
            }
            else if (Directory.Exists(local))
            {
                if (BeforeDirectoryAdd != null)
                    BeforeDirectoryAdd(this, args);
                ret = AddDirectory(root, isoloc, local);
                if (AfterDirectoryAdd != null)
                    AfterDirectoryAdd(this, args);
            }

            if (ret != null && SizeChanged != null && !m_within_update)
            {
                SizeChanged(this, args);
            }

            return ret;
        }

        /// <summary>
        /// Rename an element in the disk image
        /// </summary>
        /// <param name="parent">the directory containing the element to be renamed</param>
        /// <param name="elem">the element to be renamed</param>
        /// <param name="newname">the new name for the element</param>
        /// <returns></returns>
        public bool RenameElement(DiskImageDirectory parent, DiskImageElement elem, string newname)
        {
            Debug.Assert(false);
            elem.LocalName = newname;
            return true;
        }

        /// <summary>
        /// Add a new directory to the disk image.  There will be no file system location
        /// attached to the directory.
        /// </summary>
        /// <param name="parent">The diretory to contain the new directory</param>
        /// <param name="name">the name of the new directory</param>
        /// <returns></returns>
        public DiskImageDirectory AddNewDir(DiskImageDirectory parent, string name)
        {
            m_dirty = true;

            DiskImageDirectory dir = new DiskImageDirectory(m_settings);
            dir.LocalName = name;
            parent.Add(dir);

            if (SizeChanged != null)
            {
                EventArgs args = new EventArgs();
                SizeChanged(this, args);
            }

            return dir;
        }

        /// <summary>
        /// Remove an element (file or directory) from a directory in an image
        /// </summary>
        /// <param name="elem">the element to remove</param>
        public void Remove(DiskImageElement elem)
        {
            // We recursively look for the specific element
            m_dirty = true;
            DiskImageDirectory parent = elem.Parent;
            parent.Remove(elem);

            // Tell anyone listening that we removed an element
            if (ImageChanged != null)
            {
                ImageChangedEventArgs args = new ImageChangedEventArgs(elem, ImageChangedEventArgs.ChangeType.Removed);
                ImageChanged(this, args);
            }

            // Tell anyone listening that we changed the size of the image
            if (SizeChanged != null)
            {
                EventArgs args = new EventArgs();
                SizeChanged(this, args);
            }
        }
        #endregion

        #region private member functions

        private DiskImageDirectory FindRoot(ref string path, string filename, uint lineno)
        {
            string str;

            if (path[0] != '/')
            {
                if (filename.Length > 0)
                    str = filename + ": line " + lineno.ToString() + ": ";
                else
                    str = "-add: ";
                str += "all ISO paths must be absolute paths (begin with a '/') - '" + path + "'";
                m_logger.LogMessage(new UserMessage(UserMessage.Category.Error, 1, str));
                return null;
            }

            // Remove the leading '/' character
            path = path.Substring(1);

            char[] seps = new char[1] { '/' };
            string[] elems = path.Split(seps);
            int i = 0;
            DiskImageDirectory here = RootDirectory;

            //
            // This is a path like "/blah".  Since '/' is always insured to exist, the path
            // "/blah" is always insured to exist.  If the name blah is not a valid name for
            // this image, it will get corrected else where (in AddDirectory)
            //
            if (elems.GetLength(0) == 1)
            {
                path = elems[0];
                return m_toplevel;
            }

            while (i < elems.GetLength(0) - 1)
            {
                // Look for the name the user gave us
                DiskImageElement nextelem = here.Find(DiskImageElement.NameType.Local, elems[i], false);
                if (nextelem == null)
                {
                    if (filename.Length > 0)
                        str = filename + ": line " + lineno.ToString() + ": ";
                    else
                        str = "-add: ";
                    str += "an element of the path does not exist - '" + here.GetPath(DiskImageElement.NameType.Local) + "/" + elems[i] + "'";
                    m_logger.LogMessage(new UserMessage(UserMessage.Category.Error, 1, str));
                    return null;
                }

                if (!(nextelem is DiskImageDirectory))
                {
                    if (filename.Length > 0)
                        str = filename + ": line " + lineno.ToString() + ": ";
                    else
                        str = "-add: ";
                    str += "an element of the path is not a directory - '" + path + "'";
                    m_logger.LogMessage(new UserMessage(UserMessage.Category.Error, 1, str));
                    return null;
                }
                i++;
                here = (DiskImageDirectory)nextelem;
            }

            path = elems[i];
            return here;
        }

        private DiskImageFile AddFile(DiskImageDirectory root, string name, string local)
        {
            //
            // Check the extension list if it is defined.  No extension list means all files are
            // allowed on the volume.
            //
            if (m_allowed_exts != null)
            {
                string myext = Path.GetExtension(local);
                if (myext.Length > 0)
                {
                    if (myext[0] == '.')
                        myext = myext.Substring(1).ToLower();

                    bool found = false;
                    foreach (string ext in m_allowed_exts)
                    {
                        if (ext.ToLower() == myext)
                        {
                            found = true;
                            break;
                        }
                    }

                    if (!found)
                        return null;
                }
            }

            FileInfo info = new FileInfo(local);
            if (info.Length >= 2147483648 && m_settings.IsISO9660)
                throw new Exception("cannot add a file over 2 Gbytes to an ISO image, select 'Image Settings' to set the image type to UDF only");

            //
            // See if the file already exits
            //
            if (root.Contains(DiskImageElement.NameType.Local, local))
                return null;

            m_dirty = true;
            DiskImageFile file = new DiskImageFile(m_settings, local);
            root.Add(file);

            if (AddingFile != null)
            {
                AddingFileEventArgs args = new AddingFileEventArgs(local);
                AddingFile(this, args);
            }

            return file;
        }

        private DiskImageDirectory AddDirectory(DiskImageDirectory root, string name, string local)
        {
            m_dirty = true;
            DiskImageDirectory dir = new DiskImageDirectory(m_settings, local);

            root.Add(dir);

            if (AddingDirectory != null)
            {
                AddingDirectoryEventArgs args = new AddingDirectoryEventArgs(root, dir);
                AddingDirectory(this, args);
            }

            string[] files = Directory.GetFiles(local);
            foreach (string f in files)
                AddFile(dir, Path.GetFileName(f), f);

            string[] dirs = Directory.GetDirectories(local);
            foreach (string d in dirs)
                AddDirectory(dir, Path.GetFileName(d), d);

            return dir;
        }

        private bool IsValidIsoPath(string path, string filename, uint lineno)
        {
            string copy = path;
            return FindRoot(ref copy, filename, lineno) != null;
        }

        private bool IsValidLocalPath(string path, string filename, uint lineno)
        {
            string str;
            UserMessage m;
            bool ret = true;

            if (Directory.Exists(path))
            {
                ret = true;
            }
            else if (File.Exists(path))
            {
                FileInfo info = null;

                try
                {
                    info = new FileInfo(path);
                }
                catch (SecurityException)
                {
                    str = filename + ": line " + lineno.ToString();
                    str += "referenced file '" + path + "' - access denied";
                    m = new UserMessage(UserMessage.Category.Error, 1, str);
                    m_logger.LogMessage(m);
                    ret = false;
                }
                catch (ArgumentException)
                {
                    str = filename + ": line " + lineno.ToString();
                    str += "referenced file '" + path + "' - invalid path name";
                    m = new UserMessage(UserMessage.Category.Error, 1, str);
                    m_logger.LogMessage(m);
                    ret = false;
                }
                catch (UnauthorizedAccessException)
                {
                    str = filename + ": line " + lineno.ToString();
                    str += "referenced file '" + path + "' - access denied";
                    m = new UserMessage(UserMessage.Category.Error, 1, str);
                    m_logger.LogMessage(m);
                    ret = false;
                }
                catch (PathTooLongException)
                {
                    str = filename + ": line " + lineno.ToString();
                    str += "referenced file '" + path + "' - invalid path name, too long";
                    m = new UserMessage(UserMessage.Category.Error, 1, str);
                    m_logger.LogMessage(m);
                    ret = false;
                }
                catch (NotSupportedException)
                {
                    str = filename + ": line " + lineno.ToString();
                    str += "referenced file '" + path + "' - invalid path name, contains invalid ':' character";
                    m = new UserMessage(UserMessage.Category.Error, 1, str);
                    m_logger.LogMessage(m);
                    ret = false;
                }

                if (!info.Exists)
                {
                    str = filename + ": line " + lineno.ToString();
                    str += "referenced file '" + path + "' - no such file or directory";
                    m = new UserMessage(UserMessage.Category.Error, 1, str);
                    m_logger.LogMessage(m);
                    ret = false;
                }
            }

            return ret;
        }



        #endregion

        #region internal static methods
        /// <summary>
        /// Validates that a string is a valid string of A-characters
        /// </summary>
        /// <param name="str">the string to check</param>
        /// <returns>true if the string is valid, false otherwise</returns>
        public static bool CheckValidAChars(string str)
        {
            foreach (char ch in str)
            {
                if (ValidAChars.IndexOf(ch) == -1)
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Validates that a string is a valid string of D-characters
        /// </summary>
        /// <param name="str">the string to check</param>
        /// <returns>true if the string is valid, false otherwise</returns>
        public static bool CheckValidDChars(string str)
        {
            foreach (char ch in str)
            {
                if (!Char.IsDigit(ch) && !Char.IsUpper(ch) && !Char.IsLower(ch))
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Validates that a string is a valid string of D-characters and seperators
        /// </summary>
        /// <param name="str">the string to check</param>
        /// <returns>true if the string is valid, false otherwise</returns>
        public static bool CheckValidDCharsSep1Sep2(string str)
        {
            foreach (char ch in str)
            {
                if (!Char.IsDigit(ch) && !Char.IsLower(ch) && !Char.IsUpper(ch) && ch != '.' && ch != ';')
                    return false;
            }
            return true;
        }

        #endregion
    }
}
