//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.IO;
using System.Diagnostics;
using Bwg.Logging;

namespace Bwg.Isofs
{
    /// <summary>
    /// This class represents a directory in the disk image
    /// </summary>
    [Serializable()]
    public class DiskImageDirectory : DiskImageElement, IEnumerable<DiskImageElement>
    {
        #region public member variables

        /// <summary>
        /// The one based index of this entry in the Path Table.
        /// </summary>
        [NonSerialized()]
        public ushort PathTableEntryNumber;

        /// <summary>
        /// This event is fired when a directory in a disk image changes
        /// </summary>
        [NonSerialized()]
        public EventHandler<DiskImageDirectoryChangedArgs> Changed;

        #endregion

        #region public types
        /// <summary>
        /// The type of location
        /// </summary>
        public enum DirectoryLocationType
        {
            /// <summary>
            /// The ISO 9660 directory element (using ASCII characters)
            /// </summary>
            ISO9660 = 0,

            /// <summary>
            /// The Joliet directory element (using UNICODE characters)
            /// </summary>
            Joliet = 1,

            /// <summary>
            /// The UDF FileEntry block for this directory element
            /// </summary>
            UdfFileEntry = 2,

            /// <summary>
            /// The UDF file identifier descriptor for this directory element
            /// </summary>
            UdfFileIdentifierDescriptor = 3
        } ;
        #endregion

        #region private member variables
        /// <summary>
        /// This list contains the contents of the directory
        /// </summary>
        private IList<DiskImageElement> m_contents;

        private const string ValidDirNameChars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
        #endregion

        #region constructors

        /// <summary>
        /// Create a directory with the given name, and no mapping to
        /// the local file system.
        /// </summary>
        /// <param name="settings">the settings associated with the image being created</param>
        public DiskImageDirectory(ImageSettings settings)
            : base(settings, "")
        {
            m_contents = new List<DiskImageElement>();
        }

        /// <summary>
        /// Create a directory with the given name mapped to the given
        /// directory on the file system.
        /// </summary>
        /// <param name="settings">the settings associated with the image being created</param>
        /// <param name="fsloc">location of this directory on the local file system</param>
        public DiskImageDirectory(ImageSettings settings, string fsloc)
            : base(settings, fsloc)
        {
            m_contents = new List<DiskImageElement>();
        }
        #endregion

        #region public properties

        /// <summary>
        /// The number of subdirectories contained in this directory
        /// </summary>
        public int DirectoryCount
        {
            get
            {
                int total = 0;

                foreach (DiskImageElement elem in m_contents)
                {
                    if (elem is DiskImageDirectory)
                        total++;
                }
                return total;
            }
        }

        /// <summary>
        /// The number of files contained in this directory
        /// </summary>
        public int FileCount
        {
            get
            {
                int total = 0;

                foreach (DiskImageElement elem in m_contents)
                {
                    if (elem is DiskImageFile)
                        total++;
                }
                return total;
            }
        }

        /// <summary>
        /// The total number of files + dirs contained directly by this directory
        /// </summary>
        public int Count
        {
            get
            {
                return m_contents.Count;
            }
        }

        /// <summary>
        /// This total number of files + dirs contained recursively by this directory
        /// </summary>
        public int RecursiveTotalCount
        {
            get
            {
                return RecursiveDirectoryCount + RecursiveFileCount;
            }
        }

        /// <summary>
        /// This property returns the total number of directories in this tree of the
        /// image hierarchy including the directory here.
        /// </summary>
        public int RecursiveDirectoryCount
        {
            get
            {
                int total = 1;          // For myself

                foreach (DiskImageElement elem in m_contents)
                {
                    if (elem is DiskImageDirectory)
                        total += ((DiskImageDirectory)elem).RecursiveDirectoryCount;
                }

                return total;
            }
        }

        /// <summary>
        /// THis property returns the total number of files in this tree of the image hierarchy.
        /// </summary>
        public int RecursiveFileCount
        {
            get
            {
                int total = 0;

                foreach (DiskImageElement elem in m_contents)
                {
                    if (elem is DiskImageFile)
                        total++;
                    else if (elem is DiskImageDirectory)
                        total += ((DiskImageDirectory)elem).RecursiveFileCount;
                }

                return total;
            }
        }

        /// <summary>
        /// The level of this directory in the disk hierarchy
        /// </summary>
        public int Level
        {
            get
            {
                if (Parent == null)
                    return 1;

                return Parent.Level + 1;
            }
        }

        /// <summary>
        /// Returns the size of the extent required to represent this directory
        /// in bytes.
        /// </summary>
        public int GetExtentSizeInBytes(DiskImageElement.NameType type)
        {
            int sizebytes = 68;            // entries for . and ..
            Encoding enc = CreatorTask.GetEncoding(type);

            foreach (DiskImageElement elem in this)
            {
                byte[] name = enc.GetBytes(elem.GetName(type));
                int dirsize = name.GetLength(0) + 33;
                if ((name.GetLength(0) % 2) == 0)
                    dirsize++;

                int curblock = sizebytes / 2048;
                if ((sizebytes + dirsize) / 2048 != curblock)
                {
                    // This element will cross the sector boundary, which is not
                    // allowed.  We pad the current sector to the end and start with
                    // the next sector.
                    sizebytes = (curblock + 1) * 2048;
                }
                sizebytes += dirsize;
            }
            return sizebytes;
        }

        /// <summary>
        /// Returns the size of the extend required to represent this directory
        /// in 2048 byte blocks.
        /// </summary>
        public int GetExtentSizeInBlocks(DiskImageElement.NameType type)
        {
            int sizebytes = GetExtentSizeInBytes(type);
            int blocks = sizebytes / 2048;
            if ((sizebytes % 2048) != 0)
                blocks++;

            return blocks;
        }
        #endregion

        #region public member functions
        /// <summary>
        /// This method returns true if this directory, or an sub-directory contains a file
        /// that is of the size given or larger
        /// </summary>
        /// <param name="size">the size of interest</param>
        /// <returns>true if a file is found of the size of interest, false otherwise</returns>
        public bool ContainsFilesSized(long size)
        {
            foreach (DiskImageElement elem in m_contents)
            {
                if (elem is DiskImageDirectory)
                {
                    DiskImageDirectory dir = (DiskImageDirectory)elem;
                    if (dir.ContainsFilesSized(size))
                        return true;
                }
                else if (elem is DiskImageFile)
                {
                    DiskImageFile f = (DiskImageFile)elem;
                    if (f.FileSizeInBytes >= size)
                        return true;
                }
            }

            return false;
        }

        /// <summary>
        /// This method removes any file that is greater than the file size given.
        /// </summary>
        /// <param name="size">the size of interest</param>
        /// <returns>the number of files removed</returns>
        public int RemoveFilesSized(long size)
        {
            IList<DiskImageFile> files = new List<DiskImageFile>();

            int cnt = 0;
            foreach (DiskImageElement elem in m_contents)
            {
                if (elem is DiskImageDirectory)
                {
                    DiskImageDirectory dir = (DiskImageDirectory)elem;
                    cnt += dir.RemoveFilesSized(size);
                }
                else if (elem is DiskImageFile)
                {
                    DiskImageFile f = (DiskImageFile)elem;
                    if (f.FileSizeInBytes > size)
                        files.Add(f);
                }
            }

            cnt += files.Count;
            foreach (DiskImageFile f in files)
                Remove(f);

            return cnt;
        }


        /// <summary>
        /// Remove the contents of the directory
        /// </summary>
        public void Clear()
        {
            m_contents.Clear();

            if (Changed != null)
            {
                DiskImageDirectoryChangedArgs args =
                    new DiskImageDirectoryChangedArgs(
                        DiskImageDirectoryChangedArgs.ChangeType.RemoveAllElements, null);
                Changed(this, args);
            }
        }

        /// <summary>
        /// Add a new element to the directory
        /// </summary>
        /// <param name="elem">the element to add</param>
        public void Add(DiskImageElement elem)
        {
            AddElement(elem);
            elem.Parent = this;
            if (Changed != null)
            {
                DiskImageDirectoryChangedArgs args =
                    new DiskImageDirectoryChangedArgs(
                        DiskImageDirectoryChangedArgs.ChangeType.AddElement, elem);
                Changed(this, args);
            }
        }

        /// <summary>
        /// Remove an element from the directory in the disk image
        /// </summary>
        /// <param name="elem">the element to remove</param>
        public void Remove(DiskImageElement elem)
        {
            m_contents.Remove(elem);
            elem.Parent = null;

            if (Changed != null)
            {
                DiskImageDirectoryChangedArgs args =
                    new DiskImageDirectoryChangedArgs(
                        DiskImageDirectoryChangedArgs.ChangeType.RemoveElement, elem);
                Changed(this, args);
            }
        }

        /// <summary>
        /// Find an element in a directory by its name
        /// </summary>
        /// <param name="type">the type of name we are searching for</param>
        /// <param name="name">the name of the element to find</param>
        /// <param name="mappedonly">if true, used only cached, mapped names</param>
        /// <returns>the element if it exists, null if it does not</returns>
        public DiskImageElement Find(DiskImageElement.NameType type, string name, bool mappedonly)
        {
            foreach (DiskImageElement elem in this)
            {
                if (mappedonly)
                {
                    if (elem.GetNameIfSet(type) == name)
                        return elem;
                }
                else
                {
                    if (elem.GetName(type) == name)
                        return elem;
                }
            }
            return null;
        }

        /// <summary>
        /// Returns true if this directory contains an element with the given
        /// local name and image mapped name.
        /// </summary>
        /// <param name="type">the type of name to search for</param>
        /// <param name="name">the name to search for</param>
        /// <returns></returns>
        public bool Contains(DiskImageElement.NameType type, string name)
        {
            foreach (DiskImageElement elem in this)
            {
                if (elem.GetName(type) == name)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Flush the cached names of this element and any elements contained by this element
        /// </summary>
        public override void UpdateSettings(ImageSettings settings)
        {
            UpdateMySettings(settings);
            foreach (DiskImageElement elem in m_contents)
                elem.UpdateSettings(settings);
        }

        /// <summary>
        /// Print a directory node to the logger
        /// </summary>
        /// <param name="logger">the logger to print to</param>
        /// <param name="c">the code for the message to the logger</param>
        /// <param name="indent">the level of identation</param>
        /// <param name="recurse">if true, recursively print any child elements</param>
        public override void PrintNode(Logger logger, Bwg.Logging.UserMessage.Category c, int indent, bool recurse)
        {
            string str = "";

            for (int i = 0; i < indent; i++)
                str += ' ';

            if (Parent == null)
                str += "/";
            else
                str += GetName(NameType.Local);

            logger.LogMessage(new UserMessage(c, 0, str));
            if (recurse)
            {
                foreach (DiskImageElement elem in this)
                    elem.PrintNode(logger, c, indent + 4, recurse);
            }
        }

        /// <summary>
        /// Get the files contained in this directory element, and optionally
        /// contained in any directory elements below this element in the
        /// hierarchy
        /// </summary>
        /// <param name="list">the destination for the files</param>
        /// <param name="recurse">if true, check any child nodes for files</param>
        public void GetFiles(ref IList<DiskImageFile> list, bool recurse)
        {
            foreach (DiskImageElement elem in this)
            {
                if (elem is DiskImageFile)
                    list.Add((DiskImageFile)elem);
            }

            if (recurse)
            {
                foreach (DiskImageElement elem in this)
                {
                    if (elem is DiskImageDirectory)
                    {
                        DiskImageDirectory dir = (DiskImageDirectory)elem;
                        dir.GetFiles(ref list, true);
                    }
                }
            }
        }

        /// <summary>
        /// Get the sub-directories contained in this directory.  If the recurse
        /// flag is true, find an sub-directories anywhere below the current element
        /// in the hierarchy.
        /// </summary>
        /// <param name="list">the destination for the dirs</param>
        /// <param name="recurse">if true, descend through subdirectories</param>
        public void GetDirectories(ref IList<DiskImageDirectory> list, bool recurse)
        {
            foreach (DiskImageElement elem in this)
            {
                if (elem is DiskImageDirectory)
                    list.Add((DiskImageDirectory)elem);
            }

            if (recurse)
            {
                foreach (DiskImageElement elem in this)
                {
                    if (elem is DiskImageDirectory)
                    {
                        DiskImageDirectory dir = (DiskImageDirectory)elem;
                        dir.GetDirectories(ref list, true);
                    }
                }
            }
        }

        /// <summary>
        /// Get the size of the path table entry for this directory in bytes
        /// </summary>
        /// <returns>size of the path table entry in bytes</returns>
        public int GetPathTableEntrySize(DiskImageElement.NameType type)
        {
            int len;

            Encoding enc;
            if (type == NameType.ISO9660)
                enc = Encoding.ASCII;
            else
                enc = Encoding.BigEndianUnicode;

            string namestr = GetName(type);
            if (namestr.Length == 0)
            {
                // This is the root directory
                len = 9;
            }
            else
            {
                byte[] name = enc.GetBytes(namestr);
                len = 8 + name.GetLength(0);
            }

            if ((len % 2) == 1)
                len++;

            return len;
        }

        /// <summary>
        /// Returns an enumerator that will enumerate through the contents of this
        /// directory
        /// </summary>
        /// <returns>enumerator</returns>
        public IEnumerator<DiskImageElement> GetEnumerator()
        {
            return m_contents.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that will enumerate through the contents of this
        /// directory
        /// </summary>
        /// <returns>enumerator</returns>
        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            IEnumerable e = (IEnumerable)this;
            return e.GetEnumerator();
        }

        /// <summary>
        /// This method sets the location of this directory element given the type of
        /// location to set and the physical location.  The location is a block number.
        /// </summary>
        /// <param name="type">the type of location to set</param>
        /// <param name="extent">the extent for the directory data</param>
        public void SetDirectoryLocation(DirectoryLocationType type, ExtentAd extent)
        {
            SetExtent((int)type, 0, extent);
        }

        /// <summary>
        /// This method retreives the location of the directory element given the type
        /// of location to retreive.
        /// </summary>
        /// <param name="type">the type of location</param>
        /// <returns>the location of the directory</returns>
        public ExtentAd GetDirectoryLocation(DirectoryLocationType type)
        {
            return GetExtent((int)type, 0);
        }

        #endregion

        #region private methods

        /// <summary>
        /// Return an ISO9660 valid name based on the settings for the IS09660 image
        /// </summary>
        /// <returns></returns>
        protected override string GetISO9660Name()
        {
            string result = GetLocalName();

            // Map all '.' characters to '_' so directories have not
            // dot characters in the name
            result = RemoveIllegalCharacters(m_settings.ISO9660Settings, result, ValidDirNameChars);

            // Now, check the length
            if (result.Length > m_settings.ISO9660Settings.MaximumDirectoryLength && m_settings.ISO9660Settings.MaximumDirectoryLength != 0)
                result = ShortenName(NameType.ISO9660, result, m_settings.ISO9660Settings.MaximumDirectoryLength);
            else
            {
                //
                // The shorten process already checks for duplicate names.  If we do not have to shorten
                // we may have created a duplicate name by removing illegal characters.  This is true even if we do 
                // not change any characters in the name as some other name may have been shortened to this name or
                // mapped to this name by removing illegal chars.
                //
                if (result != string.Empty)
                    result = CheckForDuplicateNames(NameType.ISO9660, result);
            }

            return result;
        }

        private string GetFileName(string n)
        {
            int index = n.LastIndexOf('.');
            if (index == -1)
                return n;

            return n.Substring(0, index);
        }

        private string GetFileExt(string n)
        {
            string ret = string.Empty;

            int index = n.LastIndexOf('.');
            if (index == -1)
                return string.Empty;

            index++;
            while (index < n.Length && n[index] != ';')
                ret += n[index++];

            return ret;
        }

        private int GetFileVersion(string n)
        {
            int var;

            int index = n.LastIndexOf(';');
            if (index == -1)
                return 0;

            if (!Int32.TryParse(n.Substring(index + 1), out var))
                return 0;

            return var;
        }

        private string PadName(string name, int len)
        {
            Debug.Assert(name.Length <= len);

            if (name.Length == len)
                return name;

            string ret = name;
            while (ret.Length < len)
                ret += ' ';

            return ret;
        }

        private int CompareNames(string n1, string n2)
        {
            string f1, f2;

            f1 = GetFileName(n1);
            f2 = GetFileName(n2);

            f1 = PadName(f1, Math.Max(f1.Length, f2.Length));
            f2 = PadName(f2, Math.Max(f1.Length, f2.Length));
            if (f1.CompareTo(f2) < 0)
                return -1;
            if (f1.CompareTo(f2) > 0)
                return 1;

            // The names are identical, next compare the extensions
            f1 = GetFileExt(n1);
            f2 = GetFileExt(n2);

            f1 = PadName(f1, Math.Max(f1.Length, f2.Length));
            f2 = PadName(f1, Math.Max(f1.Length, f2.Length));
            if (f1.CompareTo(f2) < 0)
                return -1;
            if (f1.CompareTo(f2) > 0)
                return 1;

            // The file names and extensios are identical, next compare the versions
            int v1, v2;
            v1 = GetFileVersion(n1);
            v2 = GetFileVersion(n2);

            if (v1 < v2)
                return -1;
            if (v1 > v2)
                return 1;

            return 0;
        }

        private void AddElement(DiskImageElement elem)
        {
            if (m_contents.Count == 0)
                m_contents.Add(elem);
            else
            {
                int i = 0;
                bool added = false;

                while (i < m_contents.Count)
                {
                    if (CompareNames(elem.GetName(NameType.Local), m_contents[i].GetName(NameType.Local)) < 0)
                    {
                        added = true;
                        m_contents.Insert(i, elem);
                        break;
                    }
                    i++;
                }

                if (!added)
                    m_contents.Add(elem);
            }
        }
        #endregion
    }
}
