//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using Bwg.Logging ;

namespace Bwg.Isofs
{
    /// <summary>
    /// This class represents is a base class for an element that is contained
    /// is a disk image being created.  Both a file and a directory element is
    /// derived from this class.
    /// </summary>
    [Serializable()]
    public abstract class DiskImageElement
    {
        #region private data members
        /// <summary>
        /// The name of the file or directory on the image.  If this is an empty string
        /// the trailing portion of the file system name is used as the base for creating
        /// the mapped name.  If this is not an empty string, this is used as the base for
        /// creating the mapped name.
        /// </summary>
        private string m_fs_name;

        /// <summary>
        /// This string contains the file system location for this file.
        /// </summary>
        private string m_filesystem_location ;

        /// <summary>
        /// The location of this image element on the disk.  For files, there is one
        /// and only one location.  For directories there may be more than one.
        /// </summary>
        private IDictionary<int, ExtentAd[]> m_extent_location;

        /// <summary>
        /// Settings associated with the image being created.  These settings are used to map
        /// names from their local values to values that are suitable for the targeted image
        /// </summary>
        protected ImageSettings m_settings;

        private string m_iso_name;
        private string m_joliet_name;
        private string m_udf_name;


        #endregion

        #region public data members
        /// <summary>
        /// This is the parent directory that contains this element
        /// </summary>
        public DiskImageDirectory Parent;

        /// <summary>
        /// An event that is triggered if the name of this element is changed
        /// </summary>
        [NonSerialized()]
        public EventHandler<NameChangeEventArgs> NameChangeEvent;

        /// <summary>
        /// An external object associated with the node, usually related to a GUI
        /// </summary>
        [NonSerialized()]
        public object Tag;

        #endregion

        #region public types
        /// <summary>
        /// The type of name we are interested in
        /// </summary>
        public enum NameType
        {
            /// <summary>
            /// The name from the local file system (or assigned local name)
            /// </summary>
            Local,

            /// <summary>
            /// The name mapped per ISO9660 standard (and associated settings)
            /// </summary>
            ISO9660,

            /// <summary>
            /// The name mapped per Joliet extensions to ISO9660
            /// </summary>
            Joliet,

            /// <summary>
            /// The name mapped per UDF standards (no mapping required really)
            /// </summary>
            UDF
        } ;
    #endregion

        #region constructors
        /// <summary>
        /// Create a new file system element for the ISO fs.
        /// </summary>
        /// <param name="settings">the settings associated with the image</param>
        /// <param name="fsloc">the location of this element in the local file system</param>
        public DiskImageElement(ImageSettings settings, string fsloc)
        {
            Parent = null;
            m_filesystem_location = fsloc;
            m_fs_name = string.Empty;
            m_settings = settings;

            m_iso_name = string.Empty;
            m_joliet_name = string.Empty;
            m_udf_name = string.Empty;

            m_extent_location = new Dictionary<int, ExtentAd[]>();
        }
        #endregion

        #region public properties

        /// <summary>
        /// This property returns the local name for this element.  If this local name is not
        /// equal to string.Empty, this name is used instead of the name used in the computer
        /// filesystem.
        /// </summary>
        public string LocalName
        {
            get
            {
                return m_fs_name;
            }
            set
            {
                m_fs_name = value;
            }
        }

        /// <summary>
        /// This property is the location of this element in the file system of the local
        /// computer.  If this is a FILE element, this value is required.  If this is a 
        /// directory element, this value may be string.Empty indicating a folder on the
        /// IMAGE that does not exist physically on the computer file system.
        /// </summary>
        public string FileSystemLocation
        {
            get
            {
                return m_filesystem_location ;
            }
        }

        /// <summary>
        /// This method returns the depth of this element in the directory hierarchy
        /// </summary>
        public int Depth
        {
            get
            {
                if (Parent != null)
                    return Parent.Depth + 1;

                return 0;
            }
        }
        #endregion

        #region protected methods

        /// <summary>
        /// Flush an cached mapped names for this element
        /// </summary>
        protected void UpdateMySettings(ImageSettings settings)
        {
            m_iso_name = string.Empty ;
            m_joliet_name = string.Empty ;
            m_udf_name = string.Empty  ;
            m_settings = settings ;
        }

        /// <summary>
        /// Return the local file system name for this element
        /// </summary>
        /// <returns>the local name of this element</returns>
        protected string GetLocalName()
        {
            if (m_fs_name != string.Empty)
                return m_fs_name;

            return Path.GetFileName(m_filesystem_location);
        }

        /// <summary>
        /// Return the ISO9660 name
        /// </summary>
        /// <returns></returns>
        protected abstract string GetISO9660Name();

        /// <summary>
        /// This method returns the location of an element.  An element may belong to more than a single
        /// element set so the which paramenter gives the element set we are interested in.  In addition the
        /// location (especially for large files) may consist of more than a single segment.
        /// </summary>
        /// <param name="which">the element set</param>
        /// <param name="index">the index of this paremeter in the set</param>
        /// <param name="extent">the extent for this portion of the element</param>
        protected void SetExtent(int which, int index, ExtentAd extent)
        {
            ExtentAd[] locarray;

            if (!m_extent_location.TryGetValue(which, out locarray))
            {
                locarray = new ExtentAd[0];
            }

            if (index >= locarray.GetLength(0))
                Array.Resize(ref locarray, index + 1);

            locarray[index] = extent;
            m_extent_location[which] = locarray;
        }

        /// <summary>
        /// This method returns the number of different locations associated with an element in a given elements
        /// set.
        /// </summary>
        /// <param name="which">the element set of interest</param>
        /// <returns>the number of segments</returns>
        protected int GetExtentCount(int which)
        {
            int ret = 0;
            ExtentAd[] locarray;

            if (m_extent_location.TryGetValue(which, out locarray))
                ret = locarray.GetLength(0);

            return ret;
        }


        /// <summary>
        /// This method returns the location for this element in the given element set.
        /// </summary>
        /// <param name="which">the element set</param>
        /// <param name="index">the index of this location within the type</param>
        /// <returns>location of this element in the given element set</returns>
        protected ExtentAd GetExtent(int which, int index)
        {
            ExtentAd[] locarray;

            if (!m_extent_location.TryGetValue(which, out locarray))
            {
                Debug.Assert(false);
            }
            Debug.Assert(index < locarray.GetLength(0));

            return locarray[index];
        }

        #endregion

        #region public methods


        /// <summary>
        /// This method returns a human readable string to represent this element.
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return FileSystemLocation ;
        }

        /// <summary>
        /// This method returns the complete path name for this element
        /// </summary>
        /// <param name="type">the type of the name to return</param>
        /// <returns></returns>
        public string GetPath(NameType type)
        {
            string ret = "";

            if (Parent != null)
                ret = Parent.GetPath(type) + "/";
            ret += GetName(type);

            return ret;
        }

        /// <summary>
        /// Return the name of the element, only if it has been mapped sucessfully, otherwise return
        /// the empty string.
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public string GetNameIfSet(NameType type)
        {
            string name = string.Empty ;

            switch(type)
            {
                case NameType.Local:
                    name = GetLocalName() ;
                    break ;

                case NameType.ISO9660:
                    name = m_iso_name ;
                    break;

                case NameType.Joliet:
                    name = m_joliet_name ;
                    break ;

                case NameType.UDF:
                    name = m_udf_name ;
                    break ;
            }

            return name;
        }

        /// <summary>
        /// This method returns the name of this element.
        /// </summary>
        /// <param name="type">the type of name to return</param>
        /// <returns></returns>
        public string GetName(NameType type)
        {
            string name = string.Empty ;

            try
            {
                switch (type)
                {
                    case NameType.Local:
                        name = GetLocalName();
                        break;

                    case NameType.ISO9660:
                        if (m_iso_name == string.Empty)
                        {
                            m_iso_name = GetISO9660Name();
                            if (m_iso_name != GetLocalName())
                                Debug.WriteLine("GetName(ISO9660): mapped '" + GetLocalName() + "' -> '" + m_iso_name + "'");
                        }

                        name = m_iso_name;
                        break;

                    case NameType.Joliet:
                        if (m_joliet_name == string.Empty)
                        {
                            m_joliet_name = GetJolietName();
                            if (m_joliet_name != GetLocalName())
                                Debug.WriteLine("GetName(Joliet): mapped '" + GetLocalName() + "' -> '" + m_joliet_name + "'");
                        }
                        name = m_joliet_name;
                        break;

                    case NameType.UDF:
                        if (m_udf_name == string.Empty)
                        {
                            m_udf_name = GetLocalName();
                        }
                        name = m_udf_name;
                        break;
                }
            }
            catch (NameConversionException)
            {
                Debug.WriteLine("GetNameError '" + GetLocalName() + "' type = " + type.ToString());
                name = "BAD_NAME_MAP";
            }

            return name;

        }

        /// <summary>
        /// This method prints a single node to the logger.  This method is abstract and will be implemented by
        /// derived classes.
        /// </summary>
        /// <param name="logger">the logger where the node is printed</param>
        /// <param name="c">the category of message</param>
        /// <param name="indent">the level of the indent</param>
        /// <param name="recurse">if true, recurse if this element contains other elements</param>
        public abstract void PrintNode(Logger logger, UserMessage.Category c, int indent, bool recurse);

        /// <summary>
        /// Flush any cached names that have been mapped.
        /// </summary>
        public abstract void UpdateSettings(ImageSettings settings);

        #endregion

        #region private methods

        /// <summary>
        /// This method removes illegal characters from a name and replaces then with '_' chars
        /// </summary>
        /// <param name="settings">settings controlling the name mapping process</param>
        /// <param name="s">te string to convert</param>
        /// <param name="valid">the list of valid characters</param>
        /// <returns>an updated string with illegal chars replaced with '_' chars</returns>
        protected string RemoveIllegalCharacters(IsoJolietImageSettings settings, string s, string valid)
        {
            string result = "";
            string up;

            if (settings.ConvertCaseToUpper)
                up = s.ToUpper();
            else
                up = s;

            if (!settings.AllowNonISOChars)
            {
                foreach (char ch in up)
                {
                    char myc = ch;

                    if (valid.IndexOf(ch) == -1)
                        myc = '_';

                    result += myc;
                }
            }
            else
                result = up;

            return result;
        }

        /// <summary>
        /// This method mapes characters that are considered bad to a characters that is
        /// considered good
        /// </summary>
        /// <param name="name">the name to map</param>
        /// <param name="bad">a string containing all bad characters</param>
        /// <param name="good">the good character to use instead</param>
        /// <param name="dolast">if true, check all instances of bad chars, if not true, leave the last one alone</param>
        /// <returns>an updated and valid string</returns>
        protected string MapChars(string name, string bad, char good, bool dolast)
        {
            string result = "";

            for (int i = 0; i < name.Length; i++)
            {
                char ch = name[i];
                if (bad.IndexOf(ch) != -1 && (dolast || i != name.LastIndexOf(ch)))
                    result += good;
                else
                    result += ch;
            }
            return result;
        }

        /// <summary>
        /// Extract the version number from the file and return the filename without
        /// the version number.  If no version number exists on the file, return the filename
        /// as it was and set the version number to one.
        /// </summary>
        /// <param name="name">the original file name</param>
        /// <param name="version">the return value for version #</param>
        /// <returns>the filename without any (optional) version #</returns>
        protected string ExtractVersion(string name, out int version)
        {
            string result = name;
            int index = name.IndexOf(';');
            string ver = name.Substring(index + 1);

            if (!Int32.TryParse(ver, out version))
                version = 1;

            return result;
        }

        /// <summary>
        /// Shorten a given file name
        /// </summary>
        /// <param name="nametype"></param>
        /// <param name="name"></param>
        /// <param name="maxlen"></param>
        /// <param name="replace"></param>
        /// <param name="result"></param>
        /// <returns></returns>
        protected bool ShortenName(NameType nametype, string name, int maxlen, int replace, ref string result)
        {
            string basename;
            string fmt = "D" + replace.ToString();

            int endnum = 1;

            for (int i = 0; i < replace; i++)
                endnum *= 10;

            basename = name.Substring(0, maxlen - replace);

            for (int i = 0; i < endnum; i++)
            {
                result = basename + i.ToString(fmt);
                if (Parent.Find(nametype, result, true) == null)
                    return true;
            }

            return false;
        }

        /// <summary>
        /// Shorten a given file name, trying three different strategies
        /// </summary>
        /// <param name="nametype"></param>
        /// <param name="name"></param>
        /// <param name="maxlen"></param>
        /// <returns></returns>
        protected string ShortenName(NameType nametype, string name, int maxlen)
        {
            string result = "";

            if (!ShortenName(nametype, name, maxlen, 1, ref result))
            {
                if (!ShortenName(nametype, name, maxlen, 2, ref result))
                {
                    if (!ShortenName(nametype, name, maxlen, 3, ref result))
                    {
                        throw new Exception("internal error - cannot map a directory name to an ISO valid value");
                    }
                }
            }

            return result;
        }

        /// <summary>
        /// Insure that the given name is unique and not already used
        /// </summary>
        /// <param name="type">the name ty pe</param>
        /// <param name="name">the name</param>
        /// <returns>the (possibly) new name</returns>
        protected string CheckForDuplicateNames(NameType type, string name)
        {
            string result = string.Empty;

            if (Parent.Find(type, name, true) == null)
                return name;

            // This name already exists, so we need to map it

            if (!ShortenName(type, name, name.Length, 1, ref result))
            {
                if (!ShortenName(type, name, name.Length, 2, ref result))
                {
                    if (!ShortenName(type, name, name.Length, 3, ref result))
                    {
                        throw new Exception("internal error - cannot mape a directory name to an ISO valid value");
                    }
                }
            }

            return result;
        }

        /// <summary>
        /// Check the length of a file name against the settings.  If the name does not meet the lengths
        /// per the settings, update the name to meet the settings.
        /// </summary>
        /// <param name="settings">the settings to check against</param>
        /// <param name="type">the type of name</param>
        /// <param name="filename">the filename, may be updated</param>
        /// <param name="ext">the extension</param>
        /// <param name="ver">the version</param>
        /// <returns></returns>
        protected bool CheckNameLength(IsoJolietImageSettings settings, NameType type, ref string filename, ref string ext, int ver)
        {
            //
            // See if the name we have been handed is acceptable
            //
            if ((filename.Length <= settings.MaximumFileNameLength || settings.MaximumFileNameLength == 0) &&
                (ext.Length <= settings.MaximumFileExtensionLength || settings.MaximumFileExtensionLength == 0) &&
                (filename.Length + ext.Length <= settings.MaximumTotalNameLength || settings.MaximumTotalNameLength == 0))
                return true;

            if (ext.Length > settings.MaximumFileNameLength && settings.MaximumFileNameLength != 0)
                ext = ext.Substring(0, settings.MaximumFileNameLength);

            if (filename.Length > settings.MaximumFileNameLength && settings.MaximumFileNameLength != 0)
                filename = filename.Substring(0, settings.MaximumFileNameLength);
            else
            {
                if (settings.MaximumTotalNameLength != 0 && settings.MaximumTotalNameLength - ext.Length < filename.Length)
                    filename = filename.Substring(0, settings.MaximumTotalNameLength - ext.Length);
            }

            filename = filename.Substring(0, filename.Length - 3);

            string basename = filename;
            for (int i = 0; i < 999; i++)
            {
                filename = basename + i.ToString("D3");
                string thename = filename + "." + ext + ";" + ver.ToString() ;
                if (Parent.Find(type, thename, true) == null)
                    return true;
            }

            return false;
        }

        private string GetJolietName()
        {
            string name = GetLocalName();
            return name;
        }

        #endregion
    }
}
