//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;
using Bwg.Logging;

namespace Bwg.Isofs
{
    /// <summary>
    /// This class represents a file in the disk image.
    /// </summary>
    [Serializable()]
    public class DiskImageFile : DiskImageElement
    {
        #region private member variables
        /// <summary>
        /// The size of the file in bytes
        /// </summary>
        private long m_file_size_in_bytes;

        /// <summary>
        /// The characters that are valid in a file name.
        /// </summary>
        private const string ValidNameChars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.;";
        #endregion

        #region public types
        /// <summary>
        /// This type gives the location type for a file
        /// </summary>
        public enum FileLocationType
        {
            /// <summary>
            /// The location of the file data on the image
            /// </summary>
            DataLocation = 0,

            /// <summary>
            /// The location of the UDF file entry block for a file
            /// </summary>
            UDFFileEntryLocation = 1,
        } ;
        #endregion

        #region constructors
        /// <summary>
        /// Create a disk image file object from the local file system location
        /// and the name on the new disk image.
        /// </summary>
        /// <param name="settings">the settings assocaited with the image to be created</param>
        /// <param name="fsloc">location on the local disk</param>
        public DiskImageFile(ImageSettings settings, string fsloc) : base(settings, fsloc)
        {
            FileInfo info = new FileInfo(fsloc);
            m_file_size_in_bytes = info.Length;
        }
        #endregion

        #region public properties
        /// <summary>
        /// This property returns the file size in bytes
        /// </summary>
        public long FileSizeInBytes
        {
            get
            {
                return m_file_size_in_bytes;
            }
        }

        /// <summary>
        /// This property returns the file size in blocks
        /// </summary>
        public long FileSizeInBlocks
        {
            get
            {
                long blocks = m_file_size_in_bytes / 2048;
                if ((m_file_size_in_bytes % 2048) != 0)
                    blocks++;

                return blocks;
            }
        }
        #endregion

        #region public methods

        /// <summary>
        /// This method sets the location associated with a file
        /// </summary>
        /// <param name="type">the type of the location</param>
        /// <param name="index">the index of the location of this type</param>
        /// <param name="extent">the actual extent (size + location)</param>
        public void SetFileExtent(FileLocationType type, int index, ExtentAd extent)
        {
            SetExtent((int)type, index, extent);
        }

        /// <summary>
        /// This method returns the number of locations that are part of the given 
        /// data type assocaited with the file.
        /// </summary>
        /// <param name="type">the type of the location</param>
        /// <returns>the number of locations associated with this type</returns>
        public int GetFileExtentCount(FileLocationType type)
        {
            return GetExtentCount((int)type);
        }

        /// <summary>
        /// This method retreives the location of a given type
        /// </summary>
        /// <param name="type">the type of location</param>
        /// <param name="index">the index of the location within the type</param>
        /// <returns>the location value</returns>
        public ExtentAd GetFileExtent(FileLocationType type, int index)
        {
            return GetExtent((int)type, index);
        }

        /// <summary>
        /// flush any cached names associated with this element
        /// </summary>
        public override void UpdateSettings(ImageSettings settings)
        {
            UpdateMySettings(settings);
        }

        /// <summary>
        /// Print the file to the logger
        /// </summary>
        /// <param name="logger">the logger to print to</param>
        /// <param name="c">the message category</param>
        /// <param name="indent">the level of ident</param>
        /// <param name="recurse">not applicable for a file</param>
        public override void PrintNode(Bwg.Logging.Logger logger, Bwg.Logging.UserMessage.Category c, int indent, bool recurse)
        {
            string str = "";

            for (int i = 0; i < indent; i++)
                str += ' ';
            str += GetName(NameType.Local);
            logger.LogMessage(new UserMessage(c, 0, str));
        }
        #endregion

        #region protected methods
        /// <summary>
        /// Return an ISO 9660 valid name for this file
        /// </summary>
        /// <returns></returns>
        protected override string GetISO9660Name()
        {
            string name = GetLocalName();
            string filename, ext;
            int ver, index;

            // Remove illegal characters in the name and replace them with an '_' character
            name = RemoveIllegalCharacters(m_settings.ISO9660Settings, name, ValidNameChars);

            // If the name starts with a dot, rename the leading dot to an underscore
            if (name[0] == '.')
                name = '_' + name.Substring(1);

            // If we have multiple dots, the last dot is the one marking the 
            // extension
            name = MapChars(name, ";.", '_', false);

            index = name.IndexOf('.');
            if (index == -1)
            {
                filename = name;
                ext = "";

                // Extract the version from the filename if there was no
                // extension
                filename = ExtractVersion(filename, out ver);
            }
            else
            {
                filename = name.Substring(0, index);
                ext = name.Substring(index + 1);

                // Extract the version from the extension
                ext = ExtractVersion(ext, out ver);
            }

            //
            // Now, we have the pieces of the file name, try and find one that works
            //
            if (!CheckNameLength(m_settings.ISO9660Settings, NameType.ISO9660, ref filename, ref ext, ver))
            {
                name = filename + "." + ext + ";" + ver.ToString();
                throw new NameConversionException(name, NameType.ISO9660);
            }

            return filename + "." + ext + ";" + ver.ToString();
        }
        #endregion
    }
}
