using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Bwg.Database.Freedb ;

namespace Bwg.RipAudio
{
    public partial class SubmitFreedb : Form
    {
        #region private member variables
        /// <summary>
        /// The freedb server to submit the request to
        /// </summary>
        FreeDBServer m_server;

        /// <summary>
        /// The album to submit information about
        /// </summary>
        FreeDBAlbumInfo m_album;
        #endregion

        #region public member variables
        /// <summary>
        /// The category for the album
        /// </summary>
        public string Category;

        /// <summary>
        /// The year for the album
        /// </summary>
        public int Year;

        /// <summary>
        /// The genre for this album
        /// </summary>
        public string Genre;
        #endregion

        #region private types
        private delegate void CategoryResponseDelegate(FreeDBResponseArgs args);
        #endregion

        public SubmitFreedb(FreeDBServer server, FreeDBAlbumInfo info)
        {
            m_server = server ;
            m_album = info ;

            InitializeComponent();

            YearText.Validating += new CancelEventHandler(Year_Validating);

            Artist.Text = info.Artist;
            Title.Text = info.Title;

            Genre = string.Empty;

            for(int i = 0 ; i < info.TrackCount - 1; i++)
            {
                FreeDBTrackInfo t = info[i] ;
                ListViewItem item = new ListViewItem(t.Title);
                if (t.HasArtist)
                    item.SubItems.Add(t.Artist);
                Tracks.Items.Add(item);
            }

            m_server.FreeDBResponse += new EventHandler<FreeDBResponseArgs>(LookForCategories);

            OKButton.Enabled = false;
            Cancel.Enabled = false;
            CategoryList.Enabled = false;
            CategoryList.SelectedIndexChanged += new EventHandler(CategoryList_SelectedIndexChanged);

            m_server.CategoryRequest();
        }

        protected override void OnClosed(EventArgs e)
        {
            m_server.FreeDBResponse -= new EventHandler<FreeDBResponseArgs>(LookForCategories);
            base.OnClosed(e);
        }

        private void GotCategoryResponse(FreeDBResponseArgs resp)
        {
            if (resp.Status != FreeDBResponseArgs.ResponseType.Success)
            {
                MessageBox.Show("Error retreiving category list from server '" + m_server.Site.HostName + "'!" + 
                                " You cannot submit at this time",
                                "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error) ;

                DialogResult = DialogResult.Cancel ;
                Close() ;
            }

            OKButton.Enabled = true ;
            Cancel.Enabled = true ;
            CategoryList.Enabled = true ;

            FreeDBCategoryRequest req = (FreeDBCategoryRequest)resp.Result ;
            CategoryList.Items.Clear() ;
            foreach(string str in req.Categories)
                CategoryList.Items.Add(str) ;

            CategoryList.SelectedText = m_album.Category;
            if (CategoryList.SelectedIndex == -1)
                CategoryList.SelectedIndex = 0;
        }

        private void LookForCategories(object sender, FreeDBResponseArgs resp)
        {
            if (InvokeRequired)
            {
                object [] args = new object[1] ;
                args[0] = resp ;
                CategoryResponseDelegate proc = new CategoryResponseDelegate(GotCategoryResponse);
                Invoke(proc, args) ;
            }
            else
            {
                GotCategoryResponse(resp) ;
            }
        }

        private void CategoryList_SelectedIndexChanged(object sender, EventArgs e)
        {
            Category = (string)CategoryList.SelectedItem;
        }

        private void OKButton_Click(object sender, EventArgs e)
        {
            DialogResult = DialogResult.OK;
            Close();
        }

        private void Year_Validating(object sender, CancelEventArgs args)
        {
            if (YearText.Text.Length > 0 && !Int32.TryParse(YearText.Text, out Year))
            {
                MessageBox.Show("The text in the Year field is not a valid year!", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                args.Cancel = true;
            }
        }

        private void Year_TextChanged(object sender, EventArgs e)
        {
            if (!Int32.TryParse(YearText.Text, out Year))
                Year = 0;
        }

        private void GenreText_TextChanged(object sender, EventArgs e)
        {
            Genre = GenreText.Text;
        }
    }
}