//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices ;
using Microsoft.Win32.SafeHandles;

namespace Bwg.Utils
{
    /// <summary>
    /// This class represents a file at the Win32 level.  This class can read data from a file
    /// into an IntPtr buffer.  This is required to stream data from OS files to consumers that need
    /// data in IntPtr buffers (ACM manager and disk burner).
    /// </summary>
    public unsafe class OsFileReader : IDisposable
    {
        #region Win32 interface
        private const uint FILE_BEGIN = 0 ;
        private const uint FILE_CURRENT = 1;
        private const uint FILE_END = 2 ;

        [DllImport("Kernel32.dll", SetLastError = true, CharSet=CharSet.Unicode)]
        private static extern SafeFileHandle CreateFile(string h, uint acc, uint share, IntPtr sec, uint disp, uint flash, uint temp);

        [DllImport("Kernel32.dll", SetLastError = true)]
        private static extern bool ReadFile(SafeFileHandle h, IntPtr buffer, uint NumToRead, uint* NumberRead, IntPtr Overlapped);

        [DllImport("Kernel32.dll", SetLastError = true)]
        private static extern bool CloseHandle(IntPtr h);

        [DllImport("Kernel32.dll", SetLastError = true)]
        private static extern bool SetFilePointerEx(SafeFileHandle h, long lDistanceToMove, long *lpNewPos, uint MoveMethod);

        #endregion

        #region private member variables
        /// <summary>
        /// This is the OS handle for the file
        /// </summary>
        private SafeFileHandle m_handle;

        /// <summary>
        /// The name of the file assocated with this OsFileReader object.
        /// </summary>
        private string m_filename;

        /// <summary>
        /// If a WIN32 error occurs, this is the error that occurred
        /// </summary>
        private int m_last_error;

        #endregion

        #region public types
        /// <summary>
        /// The type of seek operation
        /// </summary>
        public enum SeekType
        {
            /// <summary>
            /// Seek from the beginning of the file
            /// </summary>
            Begin,

            /// <summary>
            /// Seek from the current position
            /// </summary>
            Current,

            /// <summary>
            /// Seek from the end of the file
            /// </summary>
            End
        } ;
        #endregion

        #region constructor
        /// <summary>
        /// This is the constructor for the file reader object.
        /// </summary>
        public OsFileReader()
        {
        }
        #endregion

        #region public properties

        /// <summary>
        /// Return the last Win32 error associated with this object
        /// </summary>
        public int LastError
        {
            get
            {
                return m_last_error;
            }
        }

        /// <summary>
        /// This property returns the name of the file
        /// </summary>
        public string FileName
        {
            get
            {
                return m_filename;
            }
        }

        /// <summary>
        /// This property returns the current position of the file pointer
        /// within the file
        /// </summary>
        public long Position
        {
            get
            {
                long result ;

                SetFilePointerEx(m_handle, 0, &result, FILE_CURRENT);
                return result;
            }
        }
        #endregion

        #region public member functions

        /// <summary>
        /// Close the file handle if it is open
        /// </summary>
        public void Dispose()
        {
            if (!m_handle.IsClosed && !m_handle.IsInvalid)
                Close();
        }

        /// <summary>
        /// Open a file to read data from
        /// </summary>
        /// <param name="filename">the name of the file</param>
        /// <returns>true if the file is opened, false otherwise</returns>
        public bool Open(string filename)
        {
            m_filename = filename;

            UInt32 acc = 0x80000000;                  // Generic Read
            m_handle = CreateFile(filename, acc, 0x01, IntPtr.Zero, 3, 0x00000080, (uint)0);
            if (m_handle.IsInvalid)
            {
                m_last_error = Marshal.GetLastWin32Error();
                return false;
            }

            return true;
        }

        /// <summary>
        /// Read data from the file into a memory buffer
        /// </summary>
        /// <param name="dest">the memory buffer for the data</param>
        /// <param name="count">the number of bytes to read</param>
        /// <param name="amountread">return value, the number of bytes read</param>
        /// <returns>true if the read suceeded, false otherwise</returns>
        public bool ReadData(IntPtr dest, int count, out int amountread)
        {
            uint NumberRead;

            amountread = 0;

            if (!ReadFile(m_handle, dest, (uint)count, &NumberRead, IntPtr.Zero))
                return false;

            amountread = (int)NumberRead;
            return true ;
        }

        /// <summary>
        /// Read data from the file into a byte array
        /// </summary>
        /// <param name="data">the byte array</param>
        /// <param name="amountread">the amount of data read</param>
        /// <returns></returns>
        public bool ReadData(ref byte[] data, out int amountread)
        {
            bool ret = false ;

            fixed (byte* dp = &data[0])
            {
                IntPtr ptr = new IntPtr(dp);
                ret = ReadData(ptr, data.GetLength(0), out amountread);
            }

            return ret;
        }

        /// <summary>
        /// Sets the file pointer to a specific position within the file
        /// </summary>
        /// <param name="pos">the position for the file pointer</param>
        /// <param name="type">the method for seeking</param>
        /// <returns>true if the seek suceeded, false otherwise</returns>
        public bool Seek(long pos, SeekType type)
        {
            uint sktype = FILE_BEGIN ;

            switch (type)
            {
                case SeekType.Begin:
                    sktype = FILE_BEGIN;
                    break;

                case SeekType.Current:
                    sktype = FILE_CURRENT;
                    break;

                case SeekType.End:
                    sktype = FILE_END;
                    break;

                default:
                    throw new Exception("invalid seek type");
            }

            return SetFilePointerEx(m_handle, pos, (long *)0, sktype);
        }

        /// <summary>
        /// Close the file
        /// </summary>
        /// <returns>returns true</returns>
        public bool Close()
        {
            m_handle.Close();
            return true;
        }
        #endregion
    }
}
